
Require Import Tactics.
Require Import Sequence.
Require Import Syntax.


Inductive rsub_head : nat -> head -> head -> Prop :=
| rsub_const {i c}
    : rsub_head i (const c) (const c)

| rsub_var_lt {i k}
    : k < i
      -> rsub_head i (var k) (var k)

| rsub_var_gt {i k}
    : k > i
      -> rsub_head i (var k) (var (k-1))
.


Inductive cutp : forall {z}, val -> pattern -> nat -> exp z -> exp z -> Prop :=
| cutp_down {z m i} {e e' : exp z}
    : rsub m i e e'
      -> cutp (thunk m) shiftm i e e'

| cutp_one {z i} {e : exp z}
    : cutp unitp unitpat i e e

| cutp_tens {z v v' p p' i} {e e' e'' : exp z}
    : cutp v p (bound p' + i) e e'
      -> cutp v' p' i e' e''
      -> cutp (pairp v v') (pairpat p p') i e e''


with cutm : term -> spine -> res -> Prop :=
| cutm_atom {r}
    : cutm (etam r) nl r

| cutm_up {r1 p r2 r}
    : lsub r1 p r2 r
      -> cutm (shiftp r1) (pm p r2) r

| cutm_pi {p m v s r m'}
    : cutp v p 0 m m'
      -> cutm m' s r
      -> cutm (lam p m) (app v s) r

| cutm_wth1 {m n s r}
    : cutm m s r
      -> cutm (pairm m n) (pi1 s) r

| cutm_wth2 {m n s r}
    : cutm n s r
      -> cutm (pairm m n) (pi2 s) r


with rsub : forall {z}, term -> nat -> exp z -> exp z -> Prop :=
| rsub_ret {f i v v'}
    : rsub f i v v'
      -> rsub f i (ret v) (ret v')

| rsub_inst {f i h s h' s'}
    : rsub_head i h h'
      -> rsub f i s s'
      -> rsub f i (use h s) (use h' s')

| rsub_reduce {f i s r s'}
    : rsub f i s s'
      -> cutm (sh 0 i f) s' r
      -> rsub f i (use (var i) s) r

| rsub_thunk {f i n n'}
    : rsub f i n n'
      -> rsub f i (thunk n) (thunk n')

| rsub_unitp {f i}
    : rsub f i unitp unitp

| rsub_pairp {f i v w v' w'}
    : rsub f i v v'
      -> rsub f i w w'
      -> rsub f i (pairp v w) (pairp v' w')

| rsub_nl {f i}
    : rsub f i nl nl

| rsub_pm {f i p r r'}
    : rsub f (bound p + i) r r'
      -> rsub f i (pm p r) (pm p r')

| rsub_app {f i v s v' s'}
    : rsub f i v v'
      -> rsub f i s s'
      -> rsub f i (app v s) (app v' s')

| rsub_pi1 {f i s s'}
    : rsub f i s s'
      -> rsub f i (pi1 s) (pi1 s')

| rsub_pi2 {f i s s'}
    : rsub f i s s'
      -> rsub f i (pi2 s) (pi2 s')

| rsub_etam {f i r r'}
    : rsub f i r r'
      -> rsub f i (etam r) (etam r')

| rsub_shiftp {f i r r'}
    : rsub f i r r'
      -> rsub f i (shiftp r) (shiftp r')

| rsub_lam {f i p n n'}
    : rsub f (bound p + i) n n'
      -> rsub f i (lam p n) (lam p n')

| rsub_pairm {f i n1 n2 n1' n2'}
    : rsub f i n1 n1'
      -> rsub f i n2 n2'
      -> rsub f i (pairm n1 n2) (pairm n1' n2')

| rsub_tuse {f i c s s'}
    : rsub f i s s'
      -> rsub f i (tuse c s) (tuse c s')

| rsub_up {f i A A'}
    : rsub f i A A'
      -> rsub f i (up A) (up A')

| rsub_pi {f i p A B A' B'}
    : rsub f i A A'
      -> rsub f (bound p + i) B B'
      -> rsub f i (pi p A B) (pi p A' B')

| rsub_wth {f i A B A' B'}
    : rsub f i A A'
      -> rsub f i B B'
      -> rsub f i (wth A B) (wth A' B')

| rsub_sigma {f i A B A' B'}
    : rsub f i A A'
      -> rsub f (1 + i) B B'
      -> rsub f i (sigma A B) (sigma A' B')

| rsub_down {f i A A'}
    : rsub f i A A'
      -> rsub f i (down A) (down A')

| rsub_one {f i}
    : rsub f i one one

| rsub_tens {f i A B A' B'}
    : rsub f i A A'
      -> rsub f i B B'
      -> rsub f i (tens A B) (tens A' B')

| rsub_ex {f i p A B A' B'}
    : rsub f i A A'
      -> rsub f (bound p + i) B B'
      -> rsub f i (ex p A B) (ex p A' B')

| rsub_tnl {f i}
    : rsub f i tnl tnl

| rsub_tapp {f i v s v' s'}
    : rsub f i v v'
      -> rsub f i s s'
      -> rsub f i (tapp v s) (tapp v' s')

| rsub_ty {f i}
    : rsub f i ty ty

| rsub_tpi {f i p A B A' B'}
    : rsub f i A A'
      -> rsub f (bound p + i) B B'
      -> rsub f i (tpi p A B) (tpi p A' B')

| rsub_trp {f i A A'}
    : rsub f i A A'
      -> rsub f i (trp A) (trp A')

| rsub_susp {f i A A'}
    : rsub f i A A'
      -> rsub f i (susp A) (susp A')

| rsub_judr {f i U U'}
    : rsub f i U U'
      -> rsub f i (judr U) (judr U')

| rsub_judv {f i A A'}
    : rsub f i A A'
      -> rsub f i (judv A) (judv A')

| rsub_juds {f i A A' U U'}
    : rsub f i A A'
      -> rsub f i U U'
      -> rsub f i (juds A U) (juds A' U')

| rsub_judt {f i A A'}
    : rsub f i A A'
      -> rsub f i (judt A) (judt A')

| rsub_judts {f i K K' L L'}
    : rsub f i K K'
      -> rsub f i L L'
      -> rsub f i (judts K L) (judts K' L')

| rsub_judnt {f i}
    : rsub f i judnt judnt

| rsub_judpt {f i}
    : rsub f i judpt judpt

| rsub_judk {f i}
    : rsub f i judk judk


with lsub : forall {z} (* re/sp *), exp z -> pattern -> res -> exp z -> Prop :=
| lsub_ret {v p r r'}
    : cutp v p 0 r r'
      -> lsub (ret v) p r r'

| lsub_use {h s p r s'}
    : lsub s p r s'
      -> lsub (use h s) p r (use h s')

| lsub_pm {p1 r1 p2 r2 r}
    : lsub r1 p2 (sh (bound p2) (bound p1) r2) r
      -> lsub (pm p1 r1) p2 r2 (pm p1 r)

| lsub_app {v s p r s'}
    : lsub s p r s'
      -> lsub (app v s) p r (app v s')

| lsub_pi1 {s p r s'}
    : lsub s p r s'
      -> lsub (pi1 s) p r (pi1 s')

| lsub_pi2 {s p r s'}
    : lsub s p r s'
      -> lsub (pi2 s) p r (pi2 s')
.


Scheme cutp_mut_ind := Minimality for cutp Sort Prop
with   cutm_mut_ind := Minimality for cutm Sort Prop
with   rsub_mut_ind := Minimality for rsub Sort Prop
with   lsub_mut_ind := Minimality for lsub Sort Prop.
Combined Scheme substitution_ind from cutp_mut_ind, cutm_mut_ind, rsub_mut_ind, lsub_mut_ind.


Inductive rsubctx : term -> nat -> context -> context -> Prop :=
| rsubctx_nil {f i}
    : rsubctx f i nil nil

| rsubctx_cons {f i G A G' A'}
    : rsubctx f i G G'
      -> rsub f (i + length G) A A'
      -> rsubctx f i (G; A) (G'; A')
.


Inductive cutpctx : val -> pattern -> nat -> context -> context -> Prop :=
| cutpctx_down {m i G G'}
    : rsubctx m i G G'
      -> cutpctx (thunk m) shiftm i G G'

| cutpctx_one {i G}
    : cutpctx unitp unitpat i G G

| cutpctx_tens {v v' p p' i G G' G''}
    : cutpctx v p (bound p' + i) G G'
      -> cutpctx v' p' i G' G''
      -> cutpctx (pairp v v') (pairpat p p') i G G''
.
