
Require Import Arith.
Require Import Compare.
Require Import Tactics.
Require Import Sequence.
Require Import Syntax.
Require Import Cut.
Require Import Self.
Require Import Static.
Require Import Substitution.
Require Import Weakening.
Require Import Size.
Require Import Compose.
Require Import Expand.
Require Import Similar.
Require Import StaticWeak.
Require Import IdentityWeak.
Require Import CutEliminationWeak.
Require Import Selfish.


Lemma sh_suspnormal :
  forall i j U,
    suspnormal U
    -> suspnormal (sh i j U).
Proof.
intros i j U H.
induct H; intros; simpl; auto using suspnormal.
Qed.


Lemma rsub_suspnormal :
   forall f i U U',
     rsub f i U U'
     -> suspnormal U
     -> suspnormal U'.
Proof.
intros f i U U' Hsub Hsusp.
revert f i U' Hsub.
#2 induct Hsusp.

(* trp *)
intros A f i U' Hsub.
invert Hsub; intros; subst; auto using suspnormal; done.

(* susp *)
intros c s f i U' Hsub.
invert Hsub; [].
intros A Hsub' <-.
invert Hsub'; [].
intros s' _ <-.
apply suspnormal_susp; done.
Qed.


Lemma rsub_pof :
  forall p A G f i A',
    pof p A G
    -> rsub f i A A'
    -> exists G',
         pof p A' G'
         /\ rsubctx f i G G'.
Proof.
intros p A G f i A' Hofp Hsub.
revert f i A' Hsub.
induct Hofp.

(* shiftm *)
intros A f i AA Hsub.
invert Hsub; [].
intros A' HsubA <-.
exists (nil; A').
split; eauto using pof; [].
apply rsubctx_cons; auto using rsubctx; [].
simpl.
replace (i+0) with i by omega.
assumption.

(* unitpat *)
intros f i AA Hsub.
invert Hsub; [].
intros <-.
exists nil.
split; auto using pof, rsubctx; done.

(* pairpat *)
intros p1 p2 A B G1 G2 _ IH1 _ IH2 f i AA Hsub.
invert Hsub; [].
intros A' B' HsubA HsubB <-.
so (IH1 _ _ _ HsubA) as (G1' & Hofp1 & HsubG1).
so (IH2 _ _ _ HsubB) as (G2' & Hofp2 & HsubG2).
exists (G1';; shctx 0 (length G1') G2').
split.
  apply pof_pairpat; auto; done.

  eapply rsubctx_append; eauto; [].
  replace (i + (length G1 + 0)) with (i + length G1) by omega.
  rewrite <- (rsubctx_length _#4 HsubG1); [].
  apply rsubctx_sh_commute_right_0; auto; done.

(* pairpat *)
intros p1 p2 A B G1 G2 _ IH1 _ IH2 f i AA Hsub.
invert Hsub; [].
intros A' B' HsubA HsubB <-.
so (IH1 _ _ _ HsubA) as (G1' & Hofp1 & HsubG1).
so (IH2 _ _ _ HsubB) as (G2' & Hofp2 & HsubG2).
exists (G1';; G2').
split.
  apply pof_pairpatx; auto; done.

  eapply rsubctx_append; eauto; [].
  replace (i + (length G1 + 0)) with (length G1 + i) by omega.
  rewrite -> (rsubctx_length _#4 HsubG1); [].
  rewrite -> (pof_bound_length _#3 Hofp1); [].
  exact HsubG2.
Qed.


Lemma cutp_pof :
  forall p A G v q i A',
    pof p A G
    -> cutp v q i A A'
    -> exists G',
         pof p A' G'
         /\ cutpctx v q i G G'.
Proof.
intros p A G v q i A' Hofp Hcut.
revert p A G v i A' Hofp Hcut.
induct q.

(* down *)
intros p A G v i A' Hofp Hcut.
invertc Hcut; [].
intros m HsubA <-.
so (rsub_pof _#6 Hofp HsubA) as (G' & Hofp' & HsubG).
eauto using cutpctx; done.

(* one *)
intros p A G v i A' Hofp Hcut.
invertc Hcut; [].
intros <- <-.
eauto using cutpctx; done.

(* tens *)
intros q1 IH1 q2 IH2 p A G vv i A'' Hofp Hcut.
invertc Hcut; [].
intros v w A' HcutA HcutA' <-.
so (IH1 _#6 Hofp HcutA) as (G' & Hofp' & HcutG).
so (IH2 _#6 Hofp' HcutA') as (G'' & Hofp'' & HcutG').
eauto using cutpctx; done.
Qed.


Lemma vof_pof_patmatch :
  forall v p A G G',
    vof G v A
    -> pof p A G'
    -> patmatch v p.
Proof.
intros v p A G G' Hofv Hofp.
revert A G G' v Hofv Hofp.
induct p.

(* down *)
intros AA G G' v Hofv Hofp.
invertc Hofp; [].
intros A <- <-.
invertc Hofv; [].
intros m _ <-.
apply patmatch_down; done.

(* one *)
intros A G G' v Hofv Hofp.
invertc Hofp; [].
intros <- <-.
invertc Hofv; [].
intros <-.
apply patmatch_one; done.

(* tens *)
intros p IH1 q IH2 AA G GG vv Hofvv Hofpp.
#2 invertc Hofpp.
  (* actual tens *)
  intros A B G1 G2 Hofp Hofq <- <-.
  invertc Hofvv; [].
  intros; subst; eauto using patmatch; done.

  (* ex *)
  intros A B G1 G2 Hofp Hofq <- <-.
  invertc Hofvv; [].
  intros v w B' Hofv HsubB Hofw <-.
  so (cutp_pof _#7 Hofq HsubB) as (G2' & Hofq' & _).
  apply patmatch_tens; eauto; done.
Qed.


Lemma lookup_invert :
  forall f i G1 A G2 G2' h B,
    lookup (G1; A;; G2) h B
    -> length G2 = i
    -> rsubctx f 0 G2 G2'
    -> (exists h' B',
          rsub_head i h h' 
          /\ lookup (G1;; G2') h' B'
          /\ rsub f i B B')
       \/
       (h = var i /\ B = sh 0 (S i) A).
Proof.
intros f i G1 A G2 G2' h B Hlookup HlenG2 HsubG2.
#2 invertc Hlookup.
  intros c D Hofc <- <-.
  left.
  exists (const c), (sh 0 (length (G1;; G2')) D).      
  do2 2 split; eauto using rsub_const, lookup_const; [].
  #3 exploit (rsub_sh_cancel f i 0 (length (G1; A;; G2)) D) as Hsub.
    omega.

    rewrite -> !app_length; [].
    simpl.
    omega.

    repl (length (G1; A;; G2) - 1) with (length (G1;; G2')) in Hsub; auto; [].
    rewrite -> !app_length; [].
    rewrite -> (rsubctx_length _#4 HsubG2); [].
    simpl.
    omega.

  intros j D Hindex <- <-.
  destruct (lt_dec j i) as [Hlt | Hnlt].
    left.
    sor (refine (index_app_lt _#5 _ Hindex)) as Hindex'; try omega; [].
    so (rsubctx_index _#6 HsubG2 Hindex') as (D' & HsubD' & Hindex'').          
    so (index_app_left _ _ G1 _ _ Hindex'') as Hindex'''.
    so (rsub_sh_commute_right_0 _ _ (S j) _ _ HsubD') as HsubD''.
    repl (0 + length G2 - j - 1 + S j) with i in HsubD'' by omega.
    exists (var j), (sh 0 (S j) D').
    do2 2 split; eauto using lookup_var, rsub_var_lt; done.

    sor (refine (index_app_ge _#5 _ Hindex)) as Hindex'; [omega |].
    clear Hindex.
    rewrite -> HlenG2 in Hindex'; [].
    destruct (dec_eq_nat j i) as [Heq | Hneq].
      subst j.
      right.
      repl (i - i) with 0 in Hindex' by omega.
      invertc Hindex'; [].
      intros <-.
      auto; done.

      left.
      repl (j - i) with (S (j - i - 1)) in Hindex' by omega.
      invertc Hindex'; [].
      intros Hindex'.
      so (index_app_right _ G2' _#3 Hindex') as Hindex''; [].
      rewrite <- (rsubctx_length _#4 HsubG2) in Hindex''; [].
      repl (j - i - 1 + length G2) with (j - 1) in Hindex'' by omega.
      exploit (rsub_sh_cancel f i 0 (S j) D) as Hsub; try omega; [].
      repl (S j - 1) with j in Hsub by omega.
      exists (var (j - 1)), (sh 0 j D).
      #2 do2 2 split; eauto.
        apply rsub_var_gt; omega.

        repl j with (S (j - 1)) at 2 by omega.
        apply lookup_var; auto; done.
Qed.


Lemma hof_invert :
  forall f i G1 A G2 G2' h B,
    tof G1 f A
    -> hof (G1; A;; G2) h B
    -> length G2 = i
    -> rsubctx f 0 G2 G2'
    -> (exists h' B',
          rsub_head i h h' 
          /\ hof (G1;; G2') h' B'
          /\ rsub f i B B')
       \/
       (h = var i /\ B = sh 0 (S i) A)
       \/
       (exists m, 
          h = var i 
          /\ expandm (sh 0 (S i) A) (use (var i) nl) m 
          /\ self m (sh 0 (S i) A) B).
Proof.
intros f i G1 A G2 G2' h B Hoff Hofh Hlen HsubG2.
#2 invert Hofh.
  (* id *)
  intros Hlookup.
  so (lookup_invert _#8 Hlookup Hlen HsubG2) as [(h' & B' & Hsubh & Hofh' & HsubB) | H].
    left.
    do 2 eexists; eauto using hof; done.

    right; left; assumption.

  (* self *)
  intros m C Hlookup Hexpand Hself.
  so (lookup_invert _#8 Hlookup Hlen HsubG2) as [(h' & C' & Hsubh & Hofh' & HsubC) | H].
    left.
    so (expandm_formation_weak _#4 (wrof_similar _#4 (rof_weak _#3 (rof_use Hofh sof_nl)) (similar_susp (similar_symm _ _ (self_similar _#3 Hself)))) Hexpand) as Hofm.
    so (can_rsub_t_weak _#7 Hlen (tof_weak _#3 Hoff) Hofm) as (m' & Hsubm & _).      
    so (inst_into_expandm _#7 (nres_use nspine_nl) Hexpand (inst_use Hsubh rsub_nl) Hsubm) as Hexpand'.
    so (sub_into_self _#7 Hself Hsubm HsubC) as (B' & Hself' & HsubB).
    exists h', B'.
    do2 2 split; eauto; [].
    eapply hof_self; eauto; [].
    eapply expandm_similar; eauto using rsub_similarm; done.

    destruct H as (-> & ->).
    right; right.
    exists m.
    do2 2 split; eauto; done.
Qed.



Definition can_cutm_prop (N : nat) :=
  forall G A m s U,
    size A N
    -> tof G m A
    -> sof G s A U
    -> suspnormal U
    -> exists r,
         cutm m s r
         /\ rof G r U.

Definition can_cutp_prop (N : nat) :=
  forall z G1 A G2 G3 G3' p v i (e : exp z) J J',
    size A N
    -> length G3 = i
    -> pof p A G2
    -> vof G1 v A
    -> gof (G1;; G2;; G3) e J
    -> cutpctx v p 0 G3 G3'
    -> cutp v p i J J'
    -> exists e',
         cutp v p i e e'
         /\ gof (G1;; G3') e' J'.

Definition can_rsub_prop (N : nat) :=
  forall z G1 A G2 G2' i f (e : exp z) J J',
    size A N
    -> length G2 = i
    -> tof G1 f A
    -> gof (G1; A;; G2) e J
    -> rsubctx f 0 G2 G2'
    -> rsub f i J J'
    -> exists e',
         rsub f i e e'
         /\ gof (G1;; G2') e' J'.

Definition can_lsub_r_prop (N : nat) :=
  forall G A r1 p G' r2 B,
    size A N
    -> rof G r1 (trp A)
    -> pof p A G'
    -> rof (G;; G') r2 (trp (sh 0 (length G') B))
    -> exists r,
         lsub r1 p r2 r
         /\ rof G r (trp B).

Definition can_lsub_s_prop (N : nat) :=
  forall G C A B s p G' r,
    size A N
    -> sof G s C (trp A)
    -> pof p A G'
    -> rof (G;; G') r (trp (sh 0 (length G') B))
    -> exists s',
         lsub s p r s'
         /\ sof G s' C (trp B).


Definition can_rsub_r_prop N G r U :=
  forall G1 A G2 G2' i f U',
    size A N
    -> G = G1; A;; G2
    -> length G2 = i
    -> tof G1 f A
    -> rsubctx f 0 G2 G2'
    -> rsub f i U U'
    -> suspnormal U
    -> exists r',
         rsub f i r r'
         /\ rof (G1;; G2') r' U'.

Definition can_rsub_v_prop N G v B :=
  forall G1 A G2 G2' i f B',
    size A N
    -> G = G1; A;; G2
    -> length G2 = i
    -> tof G1 f A
    -> rsubctx f 0 G2 G2'
    -> rsub f i B B'
    -> exists v',
         rsub f i v v'
         /\ vof (G1;; G2') v' B'.

Definition can_rsub_s_prop N G s B U :=
  forall G1 A G2 G2' i f B' U',
    size A N
    -> G = G1; A;; G2
    -> length G2 = i
    -> tof G1 f A
    -> rsubctx f 0 G2 G2'
    -> rsub f i B B'
    -> rsub f i U U'
    -> suspnormal U
    -> exists s',
         rsub f i s s'
         /\ sof (G1;; G2') s' B' U'.

Definition can_rsub_t_prop N G m B :=
  forall G1 A G2 G2' i f B',
    size A N
    -> G = G1; A;; G2
    -> length G2 = i
    -> tof G1 f A
    -> rsubctx f 0 G2 G2'
    -> rsub f i B B'
    -> exists m',
         rsub f i m m'
         /\ tof (G1;; G2') m' B'.

Definition can_cutpz_prop (N : nat) :=
  forall z G A p v G' (e : exp z) J J',
    size A N
    -> pof p A G'
    -> vof G v A
    -> gof (G;; G') e J
    -> cutp v p 0 J J'
    -> exists e',
         cutp v p 0 e e'
         /\ gof G e' J'.


Lemma can_cutpz :
  forall N, can_cutp_prop N -> can_cutpz_prop N.
Proof.
unfold can_cutp_prop, can_cutpz_prop.
intros N H z G A p v G'.
so (H z G A G' nil nil) as H'.
simpl in H'.
intros; eapply H'; eauto using cutpctx_nil, cutp_patmatch; done.
Qed.


Theorem cut_elimination :
  forall N,
    can_cutm_prop N
    /\ can_cutp_prop N
    /\ can_rsub_prop N
    /\ can_lsub_r_prop N
    /\ can_lsub_s_prop N.
Proof.
intro N.
wfinduct N using lt_wf.
intros N IH.
so (fun N' Hlt => IH N' Hlt andel) as IHcn.
so (fun N' Hlt => IH N' Hlt anderl) as IHcp.
so (fun N' Hlt => IH N' Hlt anderrl) as IHr.
so (fun N' Hlt => IH N' Hlt anderrrl) as IHlr.
so (fun N' Hlt => IH N' Hlt anderrrr) as IHls.
clear IH.
assert (can_cutm_prop N) as Hc.
  clear IHr IHls.
  intros G A m s U Hsize Hofm Hofs Hsusp.
  revert m Hsize Hofm Hsusp.
  #5 cases Hofs.
    (* nl *)
    intros G A m Hsize Hofm Hsusp.
    invert Hsusp; [].
    intros c s <-.
    invert Hofm; [].
    intros r Hofr <-.
    exists r.
    split; auto; [].
    apply cutm_atom; done.

    (* pm *)
    intros G p r2 A U G' Hofp Hofr2 m Hsize Hofm Hsusp.
    invert Hofm; clear Hofm; [].
    intros r1 Hofr1 <-.
    invert Hsize; [].
    intros N' HsizeA Heqsize.
    assert (N' < N) as HltN'_N by omega.
    unfold can_lsub_r_prop in IHlr.
    exploit (IHlr _ HltN'_N G A r1 p G' r2 U) as H; clear IHlr; eauto; [].
    destruct H as (r & Hsub & Hofr).
    exists r; split; eauto using cutm; done.

    (* app *)
    intros G v s p A B U B' Hofv HsubB Hofs m Hsize Hofm Hsusp.
    invert Hofm; clear Hofm; [].
    intros n G' Hofp Hofn <-.
    invert Hsize; [].
    intros N1 N2 HsizeA HsizeB Heqsize.
    unfold can_cutp_prop in IHcp.
    assert (N1 < N) as Hlt_N1_N by omega.
    so (cutp_compat1 _#6 HsubB (@rsub_judt)) as Hsubj.
    so (can_cutpz _ (IHcp N1 Hlt_N1_N) _#9 HsizeA Hofp Hofv (gof_t Hofn) Hsubj) as (n' & Hsubn & Hgof).
    invertc Hgof; [].
    intros Hofn'.
    so (cutp_size _#6 HsubB HsizeB) as HsizeB'.
    assert (N2 < N) as HsizeN2_N by omega.
    unfold can_cutm_prop in IHcn.
    exploit (IHcn _ HsizeN2_N G B' n' s U) as H; clear IHcn; eauto.
    destruct H as (r & Hcutm & Hofr).
    exists r; split; eauto using cutm; done.

    (* pi1 *)
    intros G s A B U Hofs l Hsize Hofm Hsusp.
    invert Hofm; clear Hofm; [].
    intros m n Hofm Hofn <-.
    invert Hsize; [].
    intros N1 N2 HsizeA HsizeB Heqsize.
    assert (N1 < N) as HltN1_N by omega.
    unfold can_cutm_prop in IHcn.
    exploit (IHcn _ HltN1_N G A m s U) as H; clear IHcn; eauto; [].
    destruct H as (r & Hcut & Hofr).
    exists r; split; eauto using cutm; done.

    (* pi2 *)
    intros G s A B U Hofs l Hsize Hofm Hsusp.
    invert Hofm; clear Hofm; [].
    intros m n Hofm Hofn <-.
    invert Hsize; [].
    intros N1 N2 HsizeA HsizeB Heqsize.
    assert (N2 < N) as HltN2_N by omega.
    unfold can_cutm_prop in IHcn.
    exploit (IHcn _ HltN2_N G B n s U) as H; clear IHcn; eauto; [].
    destruct H as (r & Hcut & Hofr).
    exists r; split; eauto using cutm; done.
split; [exact Hc |].
assert (forall N', N' <= N -> can_cutm_prop N') as H.
  intros N' Hleq.
  so (le_decide _ _ Hleq) as [Hlt | Heq].
  apply IHcn; auto; done.

  subst N'; auto; done.
renameover H into IHcn.  
clear IHlr IHls Hc.
assert (can_cutp_prop N) as Hc.
  clear IHcn.
  unfold can_cutp_prop.
  intros z G1 A G2 G3 G3' p v i e J J' Hsize HlenG3 Hofp Hofv Hgof HsubG3 HsubJ.
  revert z G2 G3 G3' p i e J J' Hsize HlenG3 Hofp Hgof HsubG3 HsubJ.
  #4 cases Hofv.
    (* down *)
    intros G1 m A Hofm z G2 G3 G3' p i e J J' Hsize HlenG3 Hofp Hof HcutG3 HcutJ.
    clear IHcp.
    invertc HcutJ; [].
    intros HsubJ <-.
    invertc Hsize; [].
    intros N' HsizeA <-.
    invertc Hofp; [].
    intros <-.
    simpl in Hof.
    invertc HcutG3; [].
    intros HsubG3.
    unfold can_rsub_prop in IHr.
    so (IHr N' (lt_n_Sn _) _#10 HsizeA HlenG3 Hofm Hof HsubG3 HsubJ) as (e' & Hsub & Hof').
    exists e'; eauto using @cutp; done.

    (* one *)
    intros G1 z G2 G3 G3' p i e J J' Hsize HlenG3 Hofp Hgof HsubG3 HsubJ.
    invertc HsubG3; [].
    intros <- <-.
    invertc HsubJ; [].
    intros <-.
    invertc Hofp; [].
    intros <-.
    simpl in Hgof.
    exists e.
    split; eauto using @cutp; done.

    (* tens *)
    intros G1 v w A B Hofv Hofw z G2 G3 G3'' pp i e J J'' Hsize HlenG3 Hofpp Hgof HsubG3 HsubJ.
    clear IHr.
    invertc HsubG3; [].
    intros p q G3' HsubG3 HsubG3' <-.
    invertc HsubJ; [].
    intros J' HsubJ HsubJ'.
    invertc Hofpp; [].
    intros G2a G2b Hofp Hofq <-.
    invertc Hsize; [].
    intros N1 N2 HsizeA HsizeB <-.
    autorewrite with canonlist in Hgof.
    rewrite -> (pof_bound_length _#3 Hofp) in Hgof.
    repl (bound q + 0) with (bound q) in HsubG3 by omega.
    unfold can_cutp_prop in IHcp.
    assert (N1 < 1 + N1 + N2) as Hlt_N1_N by omega.
    #4 exploit (IHcp _ Hlt_N1_N z G1 A G2a (shctx 0 (bound p) G2b;; G3) (G2b;; G3') p v (bound q + i) e J J') as H; auto.
      rewrite -> app_length; [].
      rewrite -> shctx_length; [].
      so (pof_bound_length _#3 Hofq).
      omega.

      autorewrite with canonlist.
      exact Hgof.

      #2 apply cutpctx_append.
        apply cutpctx_shctx_cancel_all; eauto using cutp_patmatch; done.

        rewrite -> shctx_length; [].
        simpl.
        rewrite -> (pof_bound_length _#3 Hofq); [].
        exact HsubG3.
    destruct H as (e' & Hsube & Hofe').
    clear Hlt_N1_N.
    autorewrite with canonlist in Hofe'.
    assert (N2 < 1 + N1 + N2) as Hlt_N2_N by omega.
    #2 exploit (IHcp _ Hlt_N2_N z G1 B G2b G3' G3'' q w i e' J' J'') as H; auto.
      rewrite <- (cutpctx_length _#5 HsubG3); [].
      assumption.
    destruct H as (e'' & Hsube' & Hofe'').
    exists e''.
    split; eauto using @cutp; done.

    (* ex *)
    intros G1 v w p A B B' Hofv HsubB Hofw z G2 G3 G3'' pp i e J J'' Hsize HlenG3 Hofpp Hgof HsubG3 HsubJ.
    clear IHr.
    invertc HsubG3; [].
    intros p' q G3' HsubG3 HsubG3' <-.
    invertc HsubJ; [].
    intros J' HsubJ HsubJ'.
    invertc Hofpp; [].
    intros G2a G2b Hofp Hofq -> <-.
    invertc Hsize; [].
    intros N1 N2 HsizeA HsizeB <-.
    autorewrite with canonlist in Hgof.
    repl (bound q + 0) with (bound q) in HsubG3 by omega.
    so (cutp_pof _#7 Hofq HsubB) as (G2b' & Hofq' & HsubG2b).
    unfold can_cutp_prop in IHcp.
    assert (N1 < 1 + N1 + N2) as Hlt_N1_N by omega.
    #4 exploit (IHcp _ Hlt_N1_N z G1 A G2a (G2b;; G3) (G2b';; G3') p v (bound q + i) e J J') as H; auto.
      rewrite -> app_length; [].
      so (pof_bound_length _#3 Hofq).
      omega.

      autorewrite with canonlist.
      exact Hgof.

      apply cutpctx_append; auto; [].
      simpl.
      rewrite -> (pof_bound_length _#3 Hofq); [].
      exact HsubG3.
    destruct H as (e' & Hsube & Hofe').
    clear Hlt_N1_N.
    autorewrite with canonlist in Hofe'.
    assert (N2 < 1 + N1 + N2) as Hlt_N2_N by omega.
    #3 exploit (IHcp _ Hlt_N2_N z G1 B' G2b' G3' G3'' q w i e' J' J'') as H; auto.
      eapply cutp_size; eauto; done.

      rewrite <- (cutpctx_length _#5 HsubG3); [].
      assumption.
    destruct H as (e'' & Hsube' & Hofe'').
    exists e''.
    split; eauto using @cutp; done.
split; [exact Hc |].
assert (forall N', N' <= N -> can_cutp_prop N') as H.
  intros N' Hleq.
  destruct (le_decide _ _ Hleq) as [Hlt | Heq].
    apply IHcp; omega.

    subst N'.
    exact Hc.
renameover H into IHcp.
clear IHr Hc.
#17 exploit
  (static_term_ind
     (can_rsub_r_prop N)
     (can_rsub_v_prop N)
     (can_rsub_s_prop N)
     (can_rsub_t_prop N)) as Hrvst;
  [unfold can_rsub_r_prop, can_rsub_v_prop, can_rsub_s_prop, can_rsub_t_prop .. |].
  (* ret *)
  intros G v A Hofv IH G1 G2 G3 G3' i f U' Hsize -> HlenG3 Hoff HsubG3 Hsublax Hsusp.
  invert Hsublax; [].
  intros A' HsubA <-.
  exploit (IH G1 G2 G3 G3' i f A') as H; auto; [].
  destruct H as (v' & Hsubv & Hofv').
  exists (ret v').
  split; eauto using @rsub, rof; done.

  (* use *)
  intros G h s U B Hofh Hofs IH G1 A G2 G2' i f U' HsizeA -> HlenG2 Hoff HsubG2 HsubU Hsusp.
  so (hof_invert _#8 Hoff Hofh HlenG2 HsubG2) as [H | [H | H]].
    (* inst *)
    destruct H as (h' & B' & Hsubh & Hofh' & HsubB).
    so (IH _#8 HsizeA (eq_refl _) HlenG2 Hoff HsubG2 HsubB HsubU Hsusp) as (s' & Hsubs & Hofs').
    exists (use h' s').
    split; eauto using @rsub, rof; done.

    (* reduce id *)
    destruct H as (-> & ->).
    exploit (rsub_sh_cancel f i 0 (S i) A) as HsubA; try omega; [].
    repl (S i - 1) with i in HsubA by omega.
    exploit (IH G1 A G2 G2' i f (sh 0 i A) U') as (s' & Hsubs & Hofs'); auto; [].
    clear IH.
    so (rsubctx_length _#4 HsubG2) as HlenG2'.
    symmetry in HlenG2'.
    rewrite -> HlenG2 in HlenG2'; [].
    so (weaken_tof _ _ nil _#4 HlenG2' (eq_refl _) Hoff) as Hoff'.
    simpl in Hoff'.
    unfold can_cutm_prop in IHcn.
    exploit (IHcn _ (le_refl N) (G1;; G2') (sh 0 i A) (sh 0 i f) s' U') as H; eauto using shift_size, rsub_suspnormal; [].
    destruct H as (r & Hcutm & Hofr).
    exists r.
    split.
      eapply rsub_reduce; eauto; omega.

      exact Hofr.

    (* reduce self *)
    destruct H as (m & -> & Hexpand & Hself).
    exploit (rsub_sh_cancel f i 0 (S i) A) as HsubA; try omega; [].
    repl (S i - 1) with i in HsubA by omega.
    so (expandm_formation_weak _#4 (wrof_similar _#4 (rof_weak _#3 (rof_use Hofh sof_nl)) (similar_susp (similar_symm _ _ (self_similar _#3 Hself)))) Hexpand) as Hofm.
    so (can_rsub_t_weak _#7 HlenG2 (tof_weak _#3 Hoff) Hofm) as (m' & Hsubm & _).      
    clear Hofm.
    so (sub_into_expandm _#7 nspine_nl Hexpand Hsubm rsub_nl) as Hpcut.
    invertc Hpcut; [].
    intros <-.
    so (sub_into_self _#7 Hself Hsubm HsubA) as (B' & Hself' & HsubB).
    so (IH _#8 HsizeA (eq_refl _) HlenG2 Hoff HsubG2 HsubB HsubU Hsusp) as (s' & Hsubs & Hofs').
    so (rsubctx_length _#4 HsubG2) as HlenG2'.
    symmetry in HlenG2'.
    rewrite -> HlenG2 in HlenG2'; [].
    so (weaken_tof _ _ nil _#4 HlenG2' (eq_refl _) Hoff) as Hoffw.
    simpl in Hoffw.
    so (self_member _#4 Hself' Hoffw) as Hoff'.
    #2 exploit (IHcn _ (le_refl N) (G1;; G2') B' (sh 0 i f) s' U') as H; eauto using rsub_suspnormal.
      eapply rsub_size; eauto using self_size, shift_size; done.
    destruct H as (r & Hcutm & Hofr).
    exists r.
    split.
      eapply rsub_reduce; eauto; omega.

      exact Hofr.

  (* thunk *)
  intros G m A Hofm IH G1 G2 G3 G3' i f AA Hsize -> HlenG3 Hoff HsubG3 Hsubty.
  invert Hsubty; [].
  intros A' HsubA <-.
  so (IH _#7 Hsize (eq_refl _) HlenG3 Hoff HsubG3 HsubA) as (m' & Hsubm & Hofm').
  exists (thunk m').
  split; eauto using @rsub, vof; done.

  (* unitp *)
  intros G G1 G2 G3 G3' i f AA Hsize -> HlenG3 Hoff HsubG3 Hsubty.
  invert Hsubty; [].
  intros <-.
  exists unitp.
  split; eauto using @rsub, vof; done.

  (* pairp *)
  intros G v w A B Hofv IH1 Hofw IH2 G1 G2 G3 G3' i f AA Hsize -> HlenG3 Hoff HsubG3 Hsubty.
  invert Hsubty; [].
  intros A' B' HsubA HsubB <-.
  so (IH1 _#7 Hsize (eq_refl _) HlenG3 Hoff HsubG3 HsubA) as (v' & Hsubv & Hofv').
  so (IH2 _#7 Hsize (eq_refl _) HlenG3 Hoff HsubG3 HsubB) as (w' & Hsubw & Hofw').
  exists (pairp v' w').
  split; eauto using @rsub, vof; done.

  (* pairpx *)
  intros G v w p A B B' Hofv IH1 HcutB Hofw IH2 G1 C G2 G2' i f BB Hsize -> HlenG2 Hoff HsubG2 Hsubex.
  invertc Hsubex; [].
  intros A' B'' HsubA HsubB <-.
  so (IH1 _#7 Hsize (eq_refl _) HlenG2 Hoff HsubG2 HsubA) as (v' & Hsubv & Hofv').
  so Hsubv as Hsubv'.
  repl i with (bound p + i - 0 - bound p) in Hsubv' by omega.
  sor (refine (commute_cutp_rsub _#10 _ HsubB HcutB Hsubv')) as H; [omega |].
  destruct H as (Bc & HcutB' & HsubB').
  repl (bound p + i - bound p) with i in HsubB' by omega.
  so (IH2 _#7 Hsize (eq_refl _) HlenG2 Hoff HsubG2 HsubB') as (w' & Hsubw & Hofw').
  exists (pairp v' w').
  split; eauto using @rsub, vof; done.

  (* nl *)
  intros G A G1 G2 G3 G3' i f A' U Hsize -> HlenG3 Hoff HsubG3 HsubA HsubU Hsusp.
  invert HsubU; [].
  intros A'' HsubA' <-.
  so (rsub_fun _#5 HsubA HsubA'); subst A''.
  exists nl.
  split; eauto using @rsub, sof; done.

  (* pm *)
  intros GG p r A B G Hofp Hofr IH G1 G2 G3 G3' i f AA U Hsize -> HlenG3 Hoff HsubG3 Hsubty HsubU Hsusp.
  invertc Hsubty; [].
  intros A' HsubA <-.
  invertc HsubU; [].
  intros B' HsubB <-.
  so (pof_bound_length _#3 Hofp) as HlenG.
  so (rsub_pof _#6 Hofp HsubA) as (G' & Hofp' & HsubG).
  #6 exploit (IH G1 G2 (G3;; G) (G3';; G') (i+bound p) f (trp (sh 0 (bound p) B'))) as H; clear IH; auto.
    autorewrite with canonlist; reflexivity.

    rewrite -> app_length; omega.

    eapply rsubctx_append; eauto; done.

    apply rsub_trp; [].
    so (rsub_sh_commute_right_0 _ _ (bound p) _ _ HsubB) as Hsub.
    rewrite -> HlenG; auto; done.

    exact suspnormal_trp.
  destruct H as (r' & Hsubr & Hofr').
  exists (pm p r').
  split.
    apply rsub_pm; [].
    replace (bound p + i) with (i + bound p) by omega.
    exact Hsubr.

    eapply sof_pm; eauto; [].
    autorewrite with canonlist in Hofr'.
    so (rsubctx_length _#4 HsubG) as HlenG'.
    rewrite <- HlenG'; rewrite -> HlenG; auto; done.

  (* app *)
  intros G v s p A B U Bg Hofv IH1 HsubBg Hofs IH2 G1 G2 G3 G3' i f AA U' Hsize -> HlenG3 Hoff HsubG3 Hsubpi HsubU Hsusp.
  invert Hsubpi; clear Hsubpi; [].
  intros A' Bf HsubA HsubBf <-.
  so (IH1 _#7 Hsize (eq_refl _) HlenG3 Hoff HsubG3 HsubA) as H; clear IH1.
  destruct H as (v' & Hsubv & Hofv').
  so Hsubv as Hsubv'.
  repl i with (bound p + i - 0 - bound p) in Hsubv' by omega.
  sor (refine (commute_cutp_rsub _#10 _ HsubBf HsubBg Hsubv')) as H; [omega |].
  destruct H as (B' & HsubBf' & HsubBg').
  clear Hsubv'.
  repl (bound p + i - bound p) with i in HsubBg' by omega.
  so (IH2 _#8 Hsize (eq_refl _) HlenG3 Hoff HsubG3 HsubBg' HsubU Hsusp) as (s' & Hsubs & Hofs').
  exists (app v' s').
  split; eauto using @rsub; [].
  eapply sof_app; eauto; done.

  (* pi1 *)
  intros G s A B U Hofs IH G1 G2 G3 G3' i f AA U' Hsize -> HlenG3 Hoff HsubG3 Hsubty HsubU Hsusp.
  invert Hsubty; [].
  intros A' B' HsubA HsubB <-.
  so (IH _#8 Hsize (eq_refl _) HlenG3 Hoff HsubG3 HsubA HsubU Hsusp) as H; clear IH.
  destruct H as (s' & Hsubs & Hofs').
  exists (pi1 s').
  split; eauto using @rsub, sof; done.

  (* pi2 *)
  intros G s A B U Hofs IH G1 G2 G3 G3' i f AA U' Hsize -> HlenG3 Hoff HsubG3 Hsubty HsubU Hsusp.
  invert Hsubty; [].
  intros A' B' HsubA HsubB <-.
  so (IH _#8 Hsize (eq_refl _) HlenG3 Hoff HsubG3 HsubB HsubU Hsusp) as H; clear IH.
  destruct H as (s' & Hsubs & Hofs').
  exists (pi2 s').
  split; eauto using @rsub, sof; done.

  (* etam *)
  intros G r c s Hofr IH G1 G2 G3 G3' i f A' Hsize -> HlenG3 Hoff HsubG3 Hsubty.
  invert Hsubty; [].
  intros s' Hsubs <-.
  so (IH _#7 Hsize (eq_refl _) HlenG3 Hoff HsubG3 (rsub_susp Hsubty) suspnormal_susp) as H; clear IH.
  destruct H as (r' & Hsubr & Hofr').
  exists (etam r').
  split; eauto using @rsub, tof; done.

  (* shiftp *)
  intros G r A Hof IH G1 G2 G3 G3' i f AA Hsize -> HlenG3 Hoff HsubG3 Hsubty.
  invert Hsubty; [].
  intros A' HsubA <-.
  so (IH _#7 Hsize (eq_refl _) HlenG3 Hoff HsubG3 (rsub_trp HsubA) suspnormal_trp) as H; clear IH.
  destruct H as (r' & Hsubr & Hofr).
  exists (shiftp r').
  split; eauto using @rsub, tof; done.

  (* lam *)
  intros GG p m A B G Hofp Hofm IH G1 G2 G3 G3' i f AA Hsize -> HlenG3 Hoff HsubG3 Hsubty.
  invert Hsubty; [].
  intros A' B' HsubA HsubB <-.
  so (rsub_pof _#6 Hofp HsubA) as (G' & Hofp' & HsubG).
  #4 exploit (IH G1 G2 (G3;; G) (G3';; G') (bound p + i) f B') as H; clear IH; auto.
    autorewrite with canonlist; reflexivity.

    rewrite -> app_length; [].
    so (pof_bound_length _#3 Hofp) as HlenG.
    omega.

    eapply rsubctx_append; eauto; done.
  destruct H as (m' & Hsubm & Hofm').
  autorewrite with canonlist in Hofm'.
  exists (lam p m').
  split; eauto using @rsub, tof; done.

  (* pairm *)
  intros G m n A B Hofm IH1 Hofn IH2 G1 G2 G3 G3' i f AA Hsize -> HlenG3 Hoff HsubG3 Hsubty.
  invert Hsubty; [].
  intros A' B' HsubA HsubB <-.
  so (IH1 _#7 Hsize (eq_refl _) HlenG3 Hoff HsubG3 HsubA) as H; clear IH1.
  destruct H as (m' & Hsubm & Hofm').
  so (IH2 _#7 Hsize (eq_refl _) HlenG3 Hoff HsubG3 HsubB) as H; clear IH2.
  destruct H as (n' & Hsubn & Hofn').
  exists (pairm m' n').
  split; eauto using @rsub, tof; done.

  (* pairms *)
  intros G m n A B B' Hofm IH1 HsubB Hofn IH2 G1 C G2 G2' i f AA Hsize -> HlenG2 Hoff HsubG2 Hsubty.
  invertc Hsubty; [].
  intros A' D HsubA HsubD <-.
  so (IH1 _#7 Hsize (eq_refl _) HlenG2 Hoff HsubG2 HsubA) as H; clear IH1.
  destruct H as (m' & Hsubm & Hofm').
  so Hsubm as Hsubm'.
  repl i with (S i - 0 - 1) in Hsubm' by omega.
  so (commute_rsub _#8 (lt_0_Sn _) HsubD HsubB Hsubm') as (D' & HsubD' & HsubB').
  clear Hsubm'.
  repl (S i - 1) with i in HsubB' by omega.
  so (IH2 _#7 Hsize (eq_refl _) HlenG2 Hoff HsubG2 HsubB') as H; clear IH2.
  destruct H as (n' & Hsubn & Hofn').
  exists (pairm m' n').
  split; eauto using @rsub, tof; done.
assert (forall G1 A G2 G2' i f s K K' L L',
          size A N
          -> length G2 = i
          -> tof G1 f A
          -> tsof (G1; A;; G2) s K L
          -> rsubctx f 0 G2 G2'
          -> rsub f i K K'
          -> rsub f i L L'
          -> exists s',
               rsub f i s s'
               /\ tsof (G1;; G2') s' K' L') as Hts.
  intros G1 G2 G3 G3' i f s K K' L L' Hsize HlenG3 Hoff Hofs HsubG3 HsubK HsubL.
  remember (G1; G2;; G3) as G eqn:HeqG.
  revert G1 G2 G3 G3' i f K' L' Hsize HeqG HlenG3 Hoff HsubG3 HsubK HsubL.
  induct Hofs.
  
  (* tnl *)
  intros G K G1 G2 G3 G3' i f K' L' Hsize -> HlenG3 Hoff HsubG3 HsubK HsubL.
  so (rsub_fun _#5 HsubK HsubL); subst L'.
  exists tnl.
  split; eauto using @rsub, tsof; done.
  
  (* app *)
  intros G v s p A K U Kg Hofv HsubKg Hofs IH G1 G2 G3 G3' i f KK L' Hsize -> HlenG3 Hoff HsubG3 Hsubpi HsubL.
  invert Hsubpi; clear Hsubpi; [].
  intros A' Kf HsubA HsubKf <-.
  so (Hrvst anderl _ _ _ Hofv _#7 Hsize (eq_refl _) HlenG3 Hoff HsubG3 HsubA) as (v' & Hsubv & Hofv').
  so Hsubv as Hsubv'.
  repl i with (bound p + i - 0 - bound p) in Hsubv' by omega.
  sor (refine (commute_cutp_rsub _#10 _ HsubKf HsubKg Hsubv')) as H; [omega |].
  destruct H as (K' & HsubKfg & HsubKgf).
  replace (bound p + i - bound p) with i in HsubKgf by omega.
  so (IH _#8 Hsize (eq_refl _) HlenG3 Hoff HsubG3 HsubKgf HsubL) as H; clear IH.
  destruct H as (s' & Hsubs & Hofs').
  exists (tapp v' s').
  split; eauto using @rsub; [].
  eapply tsof_app; eauto; done.
exploit
  (static_tp_ind
     (fun G A =>
        forall G1 B G2 G2' i f,
          size B N
          -> G = G1; B;; G2
          -> length G2 = i
          -> tof G1 f B
          -> rsubctx f 0 G2 G2'
          -> exists A',
               rsub f i A A'
               /\ ntof (G1;; G2') A')
     (fun G A =>
        forall G1 B G2 G2' i f,
          size B N
          -> G = G1; B;; G2
          -> length G2 = i
          -> tof G1 f B
          -> rsubctx f 0 G2 G2'
          -> exists A',
               rsub f i A A'
               /\ ptof (G1;; G2') A'))
  as Htp;
  [clear IHcn IHcp Hrvst .. |];
  [| clear Hts .. |].
  (* use *)
  intros G c s K Hofc Hofs G1 G2 G3 G3' i f Hsize -> HlenG3 Hoff HsubG3.
  #2 exploit (Hts G1 G2 G3 G3' i f s (sh 0 (length (G1; G2;; G3)) K) (sh 0 (length (G1;; G3')) K) ty ty) as H; eauto using @rsub.
    #3 exploit (rsub_sh_cancel f i 0 (length (G1; G2;; G3)) K) as Hsub.
      omega.
  
      rewrite -> !app_length; [].
      simpl.
      omega.
    force_exact Hsub.
    f_equal.
    rewrite -> !app_length; [].
    erewrite <- (rsubctx_length _ _ _ G3'); eauto; [].
    f_equal; [].
    simpl.
    omega.
  destruct H as (s' & Hsubs & Hofs').
  exists (tuse c s').
  split; eauto using @rsub, rsub_head; [].
  eapply ntof_tuse; eauto; done.
  
  (* up *)
  intros G A HofA IH G1 G2 G3 G3' i f Hsize -> HlenG3 Hoff HsubG3.
  so (IH _#6 Hsize (eq_refl _) HlenG3 Hoff HsubG3) as (A' & HsubA & HofA').
  exists (up A').
  split; eauto using @rsub, ntof; done.
  
  (* pi *)
  intros G p A B G4 HofA IH1 Hofp HofB IH2 G1 G2 G3 G3' i f Hsize -> HlenG3 Hoff HsubG3.
  so (IH1 _#6 Hsize (eq_refl _) HlenG3 Hoff HsubG3) as (A' & HsubA & HofA').
  so (rsub_pof _#6 Hofp HsubA) as (G4' & Hofp' & HsubG4).
  #4 exploit (IH2 G1 G2 (G3;; G4) (G3';; G4') (bound p + i) f); eauto.
    autorewrite with canonlist; reflexivity.
  
    rewrite -> app_length; [].
    so (pof_bound_length _#3 Hofp) as HlenG.
    omega.
  
    eapply rsubctx_append; eauto; done.
  destruct H as (B' & HsubB & HofB').
  autorewrite with canonlist in HofB'.
  exists (pi p A' B').
  split; eauto using @rsub, ntof; done.
  
  (* with *)
  intros G A B HofA IH1 HofB IH2 G1 G2 G3 G3' i f Hsize -> HlenG3 Hoff HsubG3.
  so (IH1 _#6 Hsize (eq_refl _) HlenG3 Hoff HsubG3) as H; clear IH1.
  destruct H as (A' & HsubA & HofA').
  so (IH2 _#6 Hsize (eq_refl _) HlenG3 Hoff HsubG3) as H; clear IH2.
  destruct H as (B' & HsubB & HofB').
  exists (wth A' B').
  split; eauto using @rsub, ntof; done.

  (* sigma *)
  intros G A B HofA IH1 HofB IH2 G1 C G2 G2' i f HsizeC -> Hlen Hoff HsubG2.
  so (IH1 _#6 HsizeC (eq_refl _) Hlen Hoff HsubG2) as (A' & HsubA & HofA').
  #3 exploit (IH2 G1 C (G2; A) (G2'; A') (1 + i) f) as H; eauto.
    simpl; omega.

    apply rsubctx_cons; auto; [].
    repl (0 + length G2) with i by omega.
    exact HsubA.
  destruct H as (B' & HsubB & HofB').
  exists (sigma A' B').
  split; eauto using @rsub, ntof; done.

  (* up *)
  intros G A HofA IH G1 G2 G3 G3' i f Hsize -> HlenG3 Hoff HsubG3.
  so (IH _#6 Hsize (eq_refl _) HlenG3 Hoff HsubG3) as (A' & HsubA & HofA').
  exists (down A').
  split; eauto using @rsub, ptof; done.
  
  (* one *)
  intros G G1 G2 G3 G3' i f Hsize -> HlenG3 Hoff HsubG3.
  exists one.
  split; eauto using @rsub, ptof; done.
  
  (* tens *)
  intros G A B HofA IH1 HofB IH2 G1 G2 G3 G3' i f Hsize -> HlenG3 Hoff HsubG3.
  so (IH1 _#6 Hsize (eq_refl _) HlenG3 Hoff HsubG3) as H; clear IH1.
  destruct H as (A' & HsubA & HofA').
  so (IH2 _#6 Hsize (eq_refl _) HlenG3 Hoff HsubG3) as H; clear IH2.
  destruct H as (B' & HsubB & HofB').
  exists (tens A' B').
  split; eauto using @rsub, ptof; done.
  
  (* ex *)
  intros ? p A B G HofA IH1 Hofp HofB IH2 G1 G2 G3 G3' i f Hsize -> HlenG3 Hoff HsubG3.
  so (IH1 _#6 Hsize (eq_refl _) HlenG3 Hoff HsubG3) as H; clear IH1.
  destruct H as (A' & HsubA & HofA').
  so (rsub_pof _#6 Hofp HsubA) as (G' & Hofp' & HsubG).
  so (rsubctx_append _#7 HlenG3 HsubG3 HsubG) as HsubG3G.
  so (app_length G G3) as Hlen.
  #2 sor (refine (IH2 _#6 Hsize _ Hlen Hoff HsubG3G)) as H; clear IH2.
    autorewrite with canonlist; reflexivity.
  destruct H as (B' & HsubB & HofB').
  rewrite -> HlenG3 in HsubB; [].
  rewrite -> (pof_bound_length _#3 Hofp) in HsubB; [].
  autorewrite with canonlist in HofB'.
  exists (ex p A' B').
  split; eauto using @rsub, ptof; done.
assert (forall G1 A G2 G2' i f K,
          size A N
          -> length G2 = i
          -> tof G1 f A
          -> kof (G1; A;; G2) K
          -> rsubctx f 0 G2 G2'
          -> exists K',
               rsub f i K K'
               /\ kof (G1;; G2') K') as Hk;
  [clear IHcn IHcp Hrvst Hts .. |].
  intros G1 G2 G3 G3' i f K Hsize HlenG3 Hoff HofK HsubG3.
  remember (G1; G2;; G3) as G eqn:HeqG.
  revert G1 G2 G3 G3' i f Hsize HeqG HlenG3 Hoff HsubG3.
  induct HofK.
  
  (* ty *)
  intros G G1 G2 G3 G3' i f _ HeqG HlenG3 Hoff HsubG3.
  exists ty.
  split; eauto using @rsub, kof; done.
  
  (* tpi *)
  intros G p A K G4 HofA Hofp HofK IH G1 G2 G3 G3' i f Hsize -> HlenG3 Hoff HsubG3.
  so (Htp ander _ _ HofA _#6 Hsize (eq_refl _) HlenG3 Hoff HsubG3) as (A' & HsubA & HofA').
  so (rsub_pof _#6 Hofp HsubA) as (G4' & Hofp' & HsubG4).
  #4 exploit (IH G1 G2 (G3;; G4) (G3';; G4') (bound p + i) f); eauto.
    autorewrite with canonlist; reflexivity.
  
    unfold ntp in HlenG3.
    rewrite -> app_length; [].
    so (pof_bound_length _#3 Hofp) as HlenG.
    omega.
  
    eapply rsubctx_append; eauto; done.
  destruct H as (K' & HsubK & HofK').
  autorewrite with canonlist in HofK'.
  exists (tpi p A' K').
  split; eauto using @rsub, kof; done.
split.
  unfold can_rsub_prop.
  intros z G1 A G2 G2' i f e J J' Hsize Hlen Hoff Hof HsubG2 HsubJ.
  remember (G1; A;; G2) as G eqn:HeqG.
  revert HsubJ HeqG.
  #8 cases Hof.
    clear IHcn IHcp Hts Htp Hk.
    intros ? r U Hof Hsusp HsubJ ->.
    invertc HsubJ; [].
    intros U' HsubU <-.
    so (Hrvst andel _#3 Hof _#7 Hsize (eq_refl _) Hlen Hoff HsubG2 HsubU Hsusp) as (e & Hsub & Hof').
    so (rsub_suspnormal _#4 HsubU Hsusp) as Hsusp'.
    exists e; split; eauto using @cutp, @gof; done.

    clear IHcn IHcp Hts Htp Hk.
    intros ? r U Hof HsubJ ->.
    invertc HsubJ; [].
    intros U' HsubU <-.
    so (Hrvst anderl _#3 Hof _#7 Hsize (eq_refl _) Hlen Hoff HsubG2 HsubU) as (e & Hsub & Hof').
    exists e; split; eauto using @cutp, @gof; done.

    clear IHcn IHcp Hts Htp Hk.
    intros ? r B U Hof Hsusp HsubJ ->.
    invertc HsubJ; [].
    intros A' U' HsubA HsubU <-.
    so (Hrvst anderrl _#4 Hof _#8 Hsize (eq_refl _) Hlen Hoff HsubG2 HsubA HsubU Hsusp) as (e & Hsub & Hof').
    so (rsub_suspnormal _#4 HsubU Hsusp) as Hsusp'.
    exists e; split; eauto using @cutp, @gof; done.

    clear IHcn IHcp Hts Htp Hk.
    intros ? r U Hof HsubJ ->.
    invertc HsubJ; [].
    intros U' HsubU <-.
    so (Hrvst anderrr _#3 Hof _#7 Hsize (eq_refl _) Hlen Hoff HsubG2 HsubU) as (e & Hsub & Hof').
    exists e; split; eauto using @cutp, @gof; done.

    clear IHcn IHcp Hrvst Htp Hk.
    intros G s K L Hof HsubJ ->.
    invertc HsubJ; [].
    intros K' L' HsubK HsubL <-.
    so (Hts _#11 Hsize Hlen Hoff Hof HsubG2 HsubK HsubL) as (e & Hsub & Hof').
    exists e; split; eauto using @cutp, @gof; done.

    clear IHcn IHcp Hrvst Hts Hk.
    intros ? B Hof HsubJ ->.
    invertc HsubJ; [].
    intros <-.
    so (Htp andel _ _ Hof _#6 Hsize (eq_refl _) Hlen Hoff HsubG2) as (e & Hsub & Hof').
    exists e; split; eauto using @cutp, @gof; done.

    clear IHcn IHcp Hrvst Hts Hk.
    intros ? B Hof HsubJ ->.
    invertc HsubJ; [].
    intros <-.
    so (Htp ander _ _ Hof _#6 Hsize (eq_refl _) Hlen Hoff HsubG2) as (e & Hsub & Hof').
    exists e; split; eauto using @cutp, @gof; done.

    clear IHcn IHcp Hrvst Hts Htp.
    intros ? K Hof HsubJ ->.
    invertc HsubJ; [].
    intros <-.
    so (Hk _#7 Hsize Hlen Hoff Hof HsubG2) as (e & Hsub & Hof').
    exists e; split; eauto using @cutp, @gof; done.
clear Hrvst Hts Htp Hk.
#7 exploit
  (static_term_ind
     (fun G r1 U =>
        forall A p G' r2 B,
          size A N
          -> U = trp A
          -> pof p A G'
          -> rof (G;; G') r2 (trp (sh 0 (length G') B))
          -> exists r,
               lsub r1 p r2 r
               /\ rof G r (trp B))
     (fun _ _ _ => True)
     (fun G s C U =>
        forall A B p G' r,
          size A N
          -> U = trp A
          -> pof p A G'
          -> rof (G;; G') r (trp (sh 0 (length G') B))
          -> exists s',
               lsub s p r s'
               /\ sof G s' C (trp B))
     (fun _ _ _ => True)) as H; try (intros; trivial; done); try (intros; discriminate).
  (* ret *)
  intros G v A Hofv _ A' p G' r B HsizeA Heq Hofp Hofr.
  inversion Heq; clear Heq; [].
  subst A'.
  unfold can_cutp_prop in IHcp.
  #3 exploit (can_cutpz _ (IHcp N (le_refl _)) re G A p v G' r (judr (trp (sh 0 (length G') B))) (judr (trp B))) as H; auto.
    apply gof_r; auto using suspnormal_trp; done.

    rewrite -> (pof_bound_length _#3 Hofp); [].
    #2 apply (cutp_compat1 _ _ _ (fun x => judr (trp x))).
      apply cutp_sh_cancel_all; eauto using vof_pof_patmatch; done.

      intros; eauto using @rsub; done.
  destruct H as (r' & Hsubr & Hgof).
  invert Hgof; [].
  intros Hofr'.
  exists r'; split; auto; [].
  apply lsub_ret; auto; done.

  (* use *)
  intros G h s ? A Hofh Hofs IH B p G' r2 U Hsize -> Hofp Hofr2.
  so (IH _#5 Hsize (eq_refl _) Hofp Hofr2) as H.
  destruct H as (s' & Hsubs & Hofs').
  exists (use h s').
  split; eauto using @lsub, rof; done.

  (* pm *)
  intros G1 p1 r1 A ? G2 Hofp1 Hofr1 IH B C p2 G3 r2 Hsize Heq Hofp2 Hofr2.  
  injection Heq; clear Heq; intros ->.
  remember (length G2) as i eqn:HlenG2.
  remember (length G3) as j eqn:HlenG3.
  symmetry in HlenG2; symmetry in HlenG3.
  #4 exploit (IH (sh 0 i B) p2 (shctx 0 i G3) (sh j i r2) (sh 0 i C)); auto.
    apply shift_size; auto; done.

    apply weaken_pof; auto; done.

    rewrite -> shctx_length; [].
    rewrite -> HlenG3; [].
    so (weaken_rof _#7 HlenG2 HlenG3 Hofr2) as Hofr2w.
    simpl in Hofr2w.
    rewrite -> sh_sum; [].
    replace (j+i) with (i+j) by omega.
    rewrite <- sh_sum_left; [].
    exact Hofr2w.
  destruct H as (r & Hsub & Hofr).
  exists (pm p1 r).
  split.
    apply lsub_pm; [].
    rewrite <- (pof_bound_length _#3 Hofp1); [].
    rewrite <- (pof_bound_length _#3 Hofp2); [].
    rewrite -> HlenG3; rewrite -> HlenG2; [].
    exact Hsub.

    eapply sof_pm; eauto; [].
    rewrite -> HlenG2; auto; done.

  (* app *)
  intros G v s p1 A B ? B' Hofv _ HsubB Hofs IH C U p2 G' r HsizeC -> Hofp2 Hofr.
  so (IH _#5 HsizeC (eq_refl _) Hofp2 Hofr) as (s' & Hsub & Hofs').
  exists (app v s').
  split; eauto using @lsub, sof; done.

  (* pi1 *)
  intros G s A B ? Hof IH C U p G' r Hsize -> Hofp Hofr.
  so (IH _#5 Hsize (eq_refl _) Hofp Hofr) as (s' & Hsub & Hofs').
  exists (pi1 s').
  split; eauto using @lsub, sof; done.

  (* pi2 *)
  intros G s A B ? Hof IH C U p G' r Hsize -> Hofp Hofr.
  so (IH _#5 Hsize (eq_refl _) Hofp Hofr) as (s' & Hsub & Hofs').
  exists (pi2 s').
  split; eauto using @lsub, sof; done.
destruct H as (Hr & _ & Hs & _).
split; intro; intros; eauto; done.
Qed.


Lemma can_cutm :
  forall G A m s U,
    tof G m A
    -> sof G s A U
    -> suspnormal U
    -> exists r,
         cutm m s r
         /\ rof G r U.
Proof.
intros G A m s U.
so (ntp_size A) as (N & Hsize).
eapply (cut_elimination N andel); eauto.
Qed.


Lemma can_cutp :
  forall z G1 A G2 G3 G3' p v i (e : exp z) J J',
    length G3 = i
    -> pof p A G2
    -> vof G1 v A
    -> gof (G1;; G2;; G3) e J
    -> cutpctx v p 0 G3 G3'
    -> cutp v p i J J'
    -> exists e',
         cutp v p i e e'
         /\ gof (G1;; G3') e' J'.
Proof.
intros z G1 A.
so (ptp_size A) as (N & Hsize).
so (cut_elimination N anderl); eauto.
Qed.


Lemma can_cutp_nt :
  forall G1 A G2 G3 G3' i v p B,
    length G3 = i
    -> pof p A G2
    -> vof G1 v A
    -> ntof (G1;; G2;; G3) B
    -> cutpctx v p 0 G3 G3'
    -> exists B',
         cutp v p i B B'
         /\ ntof (G1;; G3') B'.
Proof.
intros G1 A G2 G3 G3' i v p B Hlen Hofp Hofv HofB HsubG3.
so (gof_nt HofB) as Hgof.
exploit (cutp_compat0 v p i judnt) as HsubJ; eauto using cutpctx_patmatch, @rsub; [].
so (can_cutp _#12 Hlen Hofp Hofv Hgof HsubG3 HsubJ) as (B' & HsubB & Hgof').
invert Hgof'; eauto.
Qed.


Lemma can_cutp_pt :
  forall G1 A G2 G3 G3' i v p B,
    length G3 = i
    -> pof p A G2
    -> vof G1 v A
    -> ptof (G1;; G2;; G3) B
    -> cutpctx v p 0 G3 G3'
    -> exists B',
         cutp v p i B B'
         /\ ptof (G1;; G3') B'.
Proof.
intros G1 A G2 G3 G3' i v p B Hlen Hofp Hofv HofB HsubG3.
so (gof_pt HofB) as Hgof.
exploit (cutp_compat0 v p i judpt) as HsubJ; eauto using cutpctx_patmatch, @rsub; [].
so (can_cutp _#12 Hlen Hofp Hofv Hgof HsubG3 HsubJ) as (B' & HsubB & Hgof').
invert Hgof'; eauto.
Qed.


Lemma can_rsub :
  forall {z} G1 A G2 G2' i f (e : exp z) J J',
    length G2 = i
    -> tof G1 f A
    -> gof (G1; A;; G2) e J
    -> rsubctx f 0 G2 G2'
    -> rsub f i J J'
    -> exists e',
         rsub f i e e'
         /\ gof (G1;; G2') e' J'.
Proof.
intros z G1 A G2 G2' i f e J J'.
so (ntp_size A) as (N & Hsize).
eapply (cut_elimination N anderrl); eauto.
Qed.


Lemma can_rsub_r :
  forall G1 A G2 G2' i f r U U',
    length G2 = i
    -> tof G1 f A
    -> rof (G1; A;; G2) r U
    -> rsubctx f 0 G2 G2'
    -> rsub f i U U'
    -> suspnormal U
    -> exists r',
         rsub f i r r'
         /\ rof (G1;; G2') r' U'.
Proof.
intros G1 A G2 G2' i f r B B' Hlen Hoff Hofr HsubG2 HsubU Hsusp.
so (gof_r Hofr Hsusp) as Hgof.
so (can_rsub _#9 Hlen Hoff Hgof HsubG2 (rsub_judr HsubU)) as (r' & Hsubr & Hgof').
invert Hgof'; eauto.
Qed.


Lemma can_rsub_v :
  forall G1 A G2 G2' i f v B B',
    length G2 = i
    -> tof G1 f A
    -> vof (G1; A;; G2) v B
    -> rsubctx f 0 G2 G2'
    -> rsub f i B B'
    -> exists v',
         rsub f i v v'
         /\ vof (G1;; G2') v' B'.
Proof.
intros G1 A G2 G2' i f v B B' Hlen Hoff Hofv HsubG2 HsubB.
so (gof_v Hofv) as Hgof.
so (can_rsub _#9 Hlen Hoff Hgof HsubG2 (rsub_judv HsubB)) as (v' & Hsubv & Hgof').
invert Hgof'; eauto.
Qed.


Lemma can_rsub_s :
  forall G1 A G2 G2' i f s B B' U U',
    length G2 = i
    -> tof G1 f A
    -> sof (G1; A;; G2) s B U
    -> rsubctx f 0 G2 G2'
    -> rsub f i B B'
    -> rsub f i U U'
    -> suspnormal U
    -> exists s',
         rsub f i s s'
         /\ sof (G1;; G2') s' B' U'.
Proof.
intros G1 A G2 G2' i f r B B' U U' Hlen Hoff Hofr HsubG2 HsubB HsubU Hsusp.
so (gof_s Hofr Hsusp) as Hgof.
so (can_rsub _#9 Hlen Hoff Hgof HsubG2 (rsub_juds HsubB HsubU)) as (s' & Hsubs & Hgof').
invert Hgof'; eauto.
Qed.


Lemma can_rsub_t :
  forall G1 A G2 G2' i f m B B',
    length G2 = i
    -> tof G1 f A
    -> tof (G1; A;; G2) m B
    -> rsubctx f 0 G2 G2'
    -> rsub f i B B'
    -> exists m',
         rsub f i m m'
         /\ tof (G1;; G2') m' B'.
Proof.
intros G1 A G2 G2' i f v B B' Hlen Hoff Hofm HsubG2 HsubB.
so (gof_t Hofm) as Hgof.
so (can_rsub _#9 Hlen Hoff Hgof HsubG2 (rsub_judt HsubB)) as (m' & Hsubm & Hgof').
invert Hgof'; eauto.
Qed.


Lemma can_rsub_ts :
  forall G1 A G2 G2' i f s K K' L L',
    length G2 = i
    -> tof G1 f A
    -> tsof (G1; A;; G2) s K L
    -> rsubctx f 0 G2 G2'
    -> rsub f i K K'
    -> rsub f i L L'
    -> exists s',
         rsub f i s s'
         /\ tsof (G1;; G2') s' K' L'.
Proof.
intros G1 A G2 G2' i f r B B' U U' Hlen Hoff Hofs HsubG2 HsubK HsubL.
so (gof_ts Hofs) as Hgof.
so (can_rsub _#9 Hlen Hoff Hgof HsubG2 (rsub_judts HsubK HsubL)) as (s' & Hsubs & Hgof').
invert Hgof'; eauto.
Qed.


Lemma can_rsub_nt :
  forall G1 A G2 G2' i f B,
    length G2 = i
    -> tof G1 f A
    -> ntof (G1; A;; G2) B
    -> rsubctx f 0 G2 G2'
    -> exists B',
         rsub f i B B'
         /\ ntof (G1;; G2') B'.
Proof.
intros G1 A G2 G2' i f B Hlen Hoff Hof HsubG2.
so (gof_nt Hof) as Hgof.
so (can_rsub _#9 Hlen Hoff Hgof HsubG2 rsub_judnt) as (m' & Hsubm & Hgof').
invert Hgof'; eauto.
Qed.


Lemma can_rsub_pt :
  forall G1 A G2 G2' i f B,
    length G2 = i
    -> tof G1 f A
    -> ptof (G1; A;; G2) B
    -> rsubctx f 0 G2 G2'
    -> exists B',
         rsub f i B B'
         /\ ptof (G1;; G2') B'.
Proof.
intros G1 A G2 G2' i f B Hlen Hoff Hof HsubG2.
so (gof_pt Hof) as Hgof.
so (can_rsub _#9 Hlen Hoff Hgof HsubG2 rsub_judpt) as (m' & Hsubm & Hgof').
invert Hgof'; eauto.
Qed.


Lemma can_rsub_k :
  forall G1 A G2 G2' i f K,
    length G2 = i
    -> tof G1 f A
    -> kof (G1; A;; G2) K
    -> rsubctx f 0 G2 G2'
    -> exists K',
         rsub f i K K'
         /\ kof (G1;; G2') K'.
Proof.
intros G1 A G2 G2' i f K Hlen Hoff Hof HsubG2.
so (gof_k Hof) as Hgof.
so (can_rsub _#9 Hlen Hoff Hgof HsubG2 rsub_judk) as (k' & Hsubk & Hgof').
invert Hgof'; eauto.
Qed.


Lemma can_lsub_r :
  forall G A r1 p G' r2 B,
    rof G r1 (trp A)
    -> pof p A G'
    -> rof (G;; G') r2 (trp (sh 0 (length G') B))
    -> exists r,
         lsub r1 p r2 r
         /\ rof G r (trp B).
Proof.
intros G A r1 p G' r2 U.
so (ptp_size A) as (N & Hsize).
eapply (cut_elimination N anderrrl); eauto.
Qed.


Lemma can_lsub_s :
  forall G C A B s p G' r,
    sof G s C (trp A)
    -> pof p A G'
    -> rof (G;; G') r (trp (sh 0 (length G') B))
    -> exists s',
         lsub s p r s'
         /\ sof G s' C (trp B).
Proof.
intros G C A U s p G' r HofA.
so (ptp_size A) as (N & Hsize).
eapply (cut_elimination N anderrrr); eauto.
Qed.
