
Require Import Arith.
Require Import Compare.
Require Import Tactics.
Require Import Sequence.
Require Import Syntax.
Require Import Cut.
Require Import Self.
Require Import Static.
Require Import StaticWeak.
Require Import Substitution.
Require Import Weakening.
Require Import Size.
Require Import Similar.


Lemma weak_lookup_invert :
  forall i G1 A G2 h B,
    lookup (G1; A;; G2) h B
    -> length G2 = i
    -> (exists h' B',
          rsub_head i h h' 
          /\ lookup (G1;; G2) h' B'
          /\ similar B B')
       \/
       (h = var i /\ B = sh 0 (S i) A).
Proof.
intros i G1 A G2 h B Hlookup HlenG2.
#2 invertc Hlookup.
  intros c D Hofc <- <-.
  left.
  exists (const c), (sh 0 (length (G1;; G2)) D).      
  do2 2 split; eauto using lookup_const, rsub_const, sh_similarm with similar; done.

  intros j D Hindex <- <-.
  destruct (lt_dec j i) as [Hlt | Hnlt].
    left.
    sor (refine (index_app_lt _#5 _ Hindex)) as Hindex'; try omega; [].
    so (index_app_left _ _ G1 _ _ Hindex') as Hindex''.
    exists (var j), (sh 0 (S j) D).
    do2 2 split; eauto using lookup_var, rsub_var_lt with similar; done.

    sor (refine (index_app_ge _#5 _ Hindex)) as Hindex'; [omega |].
    clear Hindex.
    rewrite -> HlenG2 in Hindex'; [].
    destruct (dec_eq_nat j i) as [Heq | Hneq].
      subst j.
      right.
      repl (i - i) with 0 in Hindex' by omega.
      invertc Hindex'; [].
      intros <-.
      auto; done.

      left.
      repl (j - i) with (S (j - i - 1)) in Hindex' by omega.
      invertc Hindex'; [].
      intros Hindex'.
      so (index_app_right _ G2 _#3 Hindex') as Hindex''; [].
      repl (j - i - 1 + length G2) with (j - 1) in Hindex'' by omega.
      exists (var (j - 1)), (sh 0 j D).
      do2 2 split.
        apply rsub_var_gt; omega.

        repl j with (S (j - 1)) at 2 by omega.
        apply lookup_var; auto; done.

        eauto using sh_similarm with similar; done.
Qed.


Definition wcan_cutm_prop (N : nat) :=
  forall G A m s U,
    size A N
    -> wtof G m A
    -> wsof G s A U
    -> suspnormal U
    -> exists r,
         cutm m s r
         /\ wrof G r U.

Definition wcan_cutp_prop (N : nat) :=
  forall z G1 A G2 G3 p v i (e : exp z) J,
    size A N
    -> length G3 = i
    -> wpof p A G2
    -> wvof G1 v A
    -> wgof (G1;; G2;; G3) e J
    -> exists e',
         cutp v p i e e'
         /\ wgof (G1;; G3) e' J.

Definition wcan_rsub_prop (N : nat) :=
  forall z G1 A G2 i f (e : exp z) J,
    size A N
    -> length G2 = i
    -> wtof G1 f A
    -> wgof (G1; A;; G2) e J
    -> exists e',
         rsub f i e e'
         /\ wgof (G1;; G2) e' J.

Definition wcan_lsub_r_prop (N : nat) :=
  forall G A r1 p G' r2 B,
    size A N
    -> wrof G r1 (trp A)
    -> wpof p A G'
    -> wrof (G;; G') r2 (trp B)
    -> exists r,
         lsub r1 p r2 r
         /\ wrof G r (trp B).

Definition wcan_lsub_s_prop (N : nat) :=
  forall G C A B s p G' r,
    size A N
    -> wsof G s C (trp A)
    -> wpof p A G'
    -> wrof (G;; G') r (trp B)
    -> exists s',
         lsub s p r s'
         /\ wsof G s' C (trp B).


Definition wcan_rsub_r_prop N G r U :=
  forall G1 A G2 i f,
    size A N
    -> G = G1; A;; G2
    -> length G2 = i
    -> wtof G1 f A
    -> suspnormal U
    -> exists r',
         rsub f i r r'
         /\ wrof (G1;; G2) r' U.

Definition wcan_rsub_v_prop N G v B :=
  forall G1 A G2 i f,
    size A N
    -> G = G1; A;; G2
    -> length G2 = i
    -> wtof G1 f A
    -> exists v',
         rsub f i v v'
         /\ wvof (G1;; G2) v' B.

Definition wcan_rsub_s_prop N G s B U :=
  forall G1 A G2 i f,
    size A N
    -> G = G1; A;; G2
    -> length G2 = i
    -> wtof G1 f A
    -> suspnormal U
    -> exists s',
         rsub f i s s'
         /\ wsof (G1;; G2) s' B U.

Definition wcan_rsub_t_prop N G m B :=
  forall G1 A G2 i f,
    size A N
    -> G = G1; A;; G2
    -> length G2 = i
    -> wtof G1 f A
    -> exists m',
         rsub f i m m'
         /\ wtof (G1;; G2) m' B.

Definition wcan_cutpz_prop (N : nat) :=
  forall z G A p v G' (e : exp z) J,
    size A N
    -> wpof p A G'
    -> wvof G v A
    -> wgof (G;; G') e J
    -> exists e',
         cutp v p 0 e e'
         /\ wgof G e' J.


Lemma wcan_cutpz :
  forall N, wcan_cutp_prop N -> wcan_cutpz_prop N.
Proof.
unfold wcan_cutp_prop, wcan_cutpz_prop.
intros N H z G A p v G'.
so (H z G A G' nil) as H'.
simpl in H'.
intros; eapply H'; eauto using cutpctx_nil, cutp_patmatch; done.
Qed.


Lemma cut_elimination_weak :
  forall N,
    wcan_cutm_prop N
    /\ wcan_cutp_prop N
    /\ wcan_rsub_prop N
    /\ wcan_lsub_r_prop N
    /\ wcan_lsub_s_prop N.
Proof.
intro N.
wfinduct N using lt_wf.
intros N IH.
so (fun N' Hlt => IH N' Hlt andel) as IHcn.
so (fun N' Hlt => IH N' Hlt anderl) as IHcp.
so (fun N' Hlt => IH N' Hlt anderrl) as IHr.
so (fun N' Hlt => IH N' Hlt anderrrl) as IHlr.
so (fun N' Hlt => IH N' Hlt anderrrr) as IHls.
clear IH.
assert (wcan_cutm_prop N) as Hc.
  clear IHr IHls.
  intros G A m s U Hsize Hofm Hofs Hsusp.
  revert m Hsize Hofm Hsusp.
  #5 cases Hofs.
    (* nl *)
    intros G A m Hsize Hofm Hsusp.
    invert Hsusp; [].
    intros c s <-.
    invert Hofm; [].
    intros r Hofr <-.
    exists r.
    split; auto; [].
    apply cutm_atom; done.

    (* pm *)
    intros G p r2 A B G' Hofp Hofr2 m Hsize Hofm Hsusp.
    invert Hofm; clear Hofm; [].
    intros r1 Hofr1 <-.
    invert Hsize; [].
    intros N' HsizeA Heqsize.
    assert (N' < N) as HltN'_N by omega.
    unfold wcan_lsub_r_prop in IHlr.
    exploit (IHlr _ HltN'_N G A r1 p G' r2 B) as H; clear IHlr; eauto; [].
    destruct H as (r & Hsub & Hofr).
    exists r; split; eauto using cutm; done.

    (* app *)
    intros G v s p A B U Hofv Hofs m Hsize Hofm Hsusp.
    invert Hofm; clear Hofm; [].
    intros p' n G' Hofp Hofn <-.
    invert Hsize; [].
    intros N1 N2 HsizeA HsizeB Heqsize.
    unfold wcan_cutp_prop in IHcp.
    assert (N1 < N) as Hlt_N1_N by omega.
    so (wcan_cutpz _ (IHcp N1 Hlt_N1_N) _#8 HsizeA Hofp Hofv (wgof_t Hofn)) as (n' & Hsubn & Hgof).
    invertc Hgof; [].
    intros Hofn'.
    assert (N2 < N) as HsizeN2_N by omega.
    unfold wcan_cutm_prop in IHcn.
    exploit (IHcn _ HsizeN2_N G B n' s U) as H; clear IHcn; eauto.
    destruct H as (r & Hcutm & Hofr).
    exists r; split; eauto using cutm; done.

    (* pi1 *)
    intros G s A B U Hofs l Hsize Hofm Hsusp.
    invert Hofm; clear Hofm; [].
    intros m n Hofm Hofn <-.
    invert Hsize; [].
    intros N1 N2 HsizeA HsizeB Heqsize.
    assert (N1 < N) as HltN1_N by omega.
    unfold wcan_cutm_prop in IHcn.
    exploit (IHcn _ HltN1_N G A m s U) as H; clear IHcn; eauto; [].
    destruct H as (r & Hcut & Hofr).
    exists r; split; eauto using cutm; done.

    (* pi2 *)
    intros G s A B U Hofs l Hsize Hofm Hsusp.
    invert Hofm; clear Hofm; [].
    intros m n Hofm Hofn <-.
    invert Hsize; [].
    intros N1 N2 HsizeA HsizeB Heqsize.
    assert (N2 < N) as HltN2_N by omega.
    unfold wcan_cutm_prop in IHcn.
    exploit (IHcn _ HltN2_N G B n s U) as H; clear IHcn; eauto; [].
    destruct H as (r & Hcut & Hofr).
    exists r; split; eauto using cutm; done.
split; [exact Hc |].
assert (forall N', N' <= N -> wcan_cutm_prop N') as H.
  intros N' Hleq.
  so (le_decide _ _ Hleq) as [Hlt | Heq].
  apply IHcn; auto; done.

  subst N'; auto; done.
renameover H into IHcn.  
clear IHlr IHls Hc.
assert (wcan_cutp_prop N) as Hc.
  clear IHcn.
  unfold wcan_cutp_prop.
  intros z G1 A G2 G3 p v i e J Hsize HlenG3 Hofp Hofv Hgof.
  revert z G2 G3 p i e J Hsize HlenG3 Hofp Hgof.
  #4 cases Hofv.
    (* down *)
    intros G1 m A Hofm z G2 G3 p i e J Hsize HlenG3 Hofp Hof.
    clear IHcp.
    invertc Hsize; [].
    intros N' HsizeA <-.
    invertc Hofp; [].
    intros <- <-.
    simpl in Hof.
    unfold wcan_rsub_prop in IHr.
    so (IHr N' (lt_n_Sn _) _#8 HsizeA HlenG3 Hofm Hof) as (e' & Hsub & Hof').
    exists e'; eauto using @cutp; done.

    (* one *)
    intros G1 z G2 G3 p i e J Hsize HlenG3 Hofp Hgof.
    invertc Hofp; [].
    intros <- <-.
    simpl in Hgof.
    exists e.
    split; eauto using @cutp; done.

    (* tens *)
    intros G1 v w A B Hofv Hofw z G2 G3 pp i e J Hsize HlenG3 Hofpp Hgof.
    clear IHr.
    invertc Hofpp; [].
    intros p q G2a G2b Hofp Hofq <- <-.
    invertc Hsize; [].
    intros N1 N2 HsizeA HsizeB <-.
    autorewrite with canonlist in Hgof.
    rewrite -> (wpof_bound_length _#3 Hofp) in Hgof.
    unfold wcan_cutp_prop in IHcp.
    assert (N1 < 1 + N1 + N2) as Hlt_N1_N by omega.
    #3 exploit (IHcp _ Hlt_N1_N z G1 A G2a (G2b;; G3) p v (bound q + i) e J) as H; auto.
      rewrite -> app_length; [].
      so (wpof_bound_length _#3 Hofq).
      omega.

      autorewrite with canonlist.
      eapply wgof_similarctx; eauto using shctx_similarctx with similar; done.
    destruct H as (e' & Hsube & Hofe').
    clear Hlt_N1_N.
    autorewrite with canonlist in Hofe'.
    assert (N2 < 1 + N1 + N2) as Hlt_N2_N by omega.
    exploit (IHcp _ Hlt_N2_N z G1 B G2b G3 q w i e' J) as H; auto; [].
    destruct H as (e'' & Hsube' & Hofe'').
    exists e''.
    split; eauto using @cutp; done.

    (* ex *)
    intros G1 v w p' A B Hofv Hofw z G2 G3 pp i e J Hsize HlenG3 Hofpp Hgof.
    clear IHr.
    invertc Hofpp; [].
    intros p q G2a G2b Hofp Hofq <- <-.
    invertc Hsize; [].
    intros N1 N2 HsizeA HsizeB <-.
    autorewrite with canonlist in Hgof.
    unfold wcan_cutp_prop in IHcp.
    assert (N1 < 1 + N1 + N2) as Hlt_N1_N by omega.
    #3 exploit (IHcp _ Hlt_N1_N z G1 A G2a (G2b;; G3) p v (bound q + i) e J) as H; auto.
      rewrite -> app_length; [].
      so (wpof_bound_length _#3 Hofq).
      omega.

      autorewrite with canonlist.
      exact Hgof.
    destruct H as (e' & Hsube & Hofe').
    clear Hlt_N1_N.
    autorewrite with canonlist in Hofe'.
    assert (N2 < 1 + N1 + N2) as Hlt_N2_N by omega.
    exploit (IHcp _ Hlt_N2_N z G1 B G2b G3 q w i e' J) as H; auto; [].
    destruct H as (e'' & Hsube' & Hofe'').
    exists e''.
    split; eauto using @cutp; done.
split; [exact Hc |].
assert (forall N', N' <= N -> wcan_cutp_prop N') as H.
  intros N' Hleq.
  destruct (le_decide _ _ Hleq) as [Hlt | Heq].
    apply IHcp; omega.

    subst N'.
    exact Hc.
renameover H into IHcp.
clear IHr Hc.
#17 exploit
  (weak_static_term_ind
     (wcan_rsub_r_prop N)
     (wcan_rsub_v_prop N)
     (wcan_rsub_s_prop N)
     (wcan_rsub_t_prop N)) as Hrvst;
  [unfold wcan_rsub_r_prop, wcan_rsub_v_prop, wcan_rsub_s_prop, wcan_rsub_t_prop .. |].
  (* ret *)
  intros G v A Hofv IH G1 G2 G3 i f Hsize -> HlenG3 Hoff Hsusp.
  exploit (IH G1 G2 G3 i f) as H; auto; [].
  destruct H as (v' & Hsubv & Hofv').
  exists (ret v').
  split; eauto using @rsub, wrof; done.

  (* use *)
  intros G h s B B' U Hlookup Hsim Hofs IH G1 A G2 i f HsizeA -> HlenG2 Hoff Hsusp.
  so (weak_lookup_invert _#6 Hlookup HlenG2) as [H | H].
    (* inst *)
    destruct H as (h' & B'' & Hofh' & Hlookup' & Hsim').
    so (IH _#5 HsizeA (eq_refl _) HlenG2 Hoff Hsusp) as (s' & Hsubs & Hofs').
    exists (use h' s').
    split; eauto using @rsub; [].
    eapply wrof_use; eauto; [].
    eapply similar_trans; eauto with similar; done.

    (* reduce *)
    destruct H as (-> & ->).
    exploit (IH G1 A G2 i f) as (s' & Hsubs & Hofs'); auto; [].
    clear IH.
    so (weaken_wtof _ _ nil _#4 HlenG2 (eq_refl _) Hoff) as Hoff'.
    simpl in Hoff'.
    unfold wcan_cutm_prop in IHcn.
    assert (similar A B') as Hsim'.
      eapply similar_trans; eauto using sh_similarm; done.
    #2 exploit (IHcn _ (le_refl N) (G1;; G2) B' (sh 0 i f) s' U) as H; eauto using similar_size.
      eapply wtof_similar; eauto; done.
    destruct H as (r & Hcutm & Hofr).
    exists r.
    split.
      eapply rsub_reduce; eauto; omega.

      exact Hofr.

  (* thunk *)
  intros G m A Hofm IH G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  so (IH _#5 Hsize (eq_refl _) HlenG3 Hoff) as (m' & Hsubm & Hofm').
  exists (thunk m').
  split; eauto using @rsub, wvof; done.

  (* unitp *)
  intros G G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  exists unitp.
  split; eauto using @rsub, wvof; done.

  (* pairp *)
  intros G v w A B Hofv IH1 Hofw IH2 G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  so (IH1 _#5 Hsize (eq_refl _) HlenG3 Hoff) as (v' & Hsubv & Hofv').
  so (IH2 _#5 Hsize (eq_refl _) HlenG3 Hoff) as (w' & Hsubw & Hofw').
  exists (pairp v' w').
  split; eauto using @rsub, wvof; done.

  (* pairpx *)
  intros G v w p A B Hofv IH1 Hofw IH2 G1 C G2 i f Hsize -> HlenG2 Hoff.
  so (IH1 _#5 Hsize (eq_refl _) HlenG2 Hoff) as (v' & Hsubv & Hofv').
  so (IH2 _#5 Hsize (eq_refl _) HlenG2 Hoff) as (w' & Hsubw & Hofw').
  exists (pairp v' w').
  split; eauto using @rsub, wvof; done.

  (* nl *)
  intros G A G1 B G2 i f Hsize -> HlenG3 Hoff Hsusp.
  exists nl.
  split; eauto using @rsub, wsof; done.

  (* pm *)
  intros GG p r A U G Hofp Hofr IH G1 G2 G3 i f Hsize -> HlenG3 Hoff Hsusp.
  so (wpof_bound_length _#3 Hofp) as HlenG.
  #3 exploit (IH G1 G2 (G3;; G) (i+bound p) f) as H; clear IH; auto.
    autorewrite with canonlist; reflexivity.

    rewrite -> app_length; omega.
  destruct H as (r' & Hsubr & Hofr').
  exists (pm p r').
  split.
    apply rsub_pm; [].
    replace (bound p + i) with (i + bound p) by omega.
    exact Hsubr.

    eapply wsof_pm; eauto; [].
    autorewrite with canonlist in Hofr'.
    assumption.

  (* app *)
  intros G v s p A B U Hofv IH1 Hofs IH2 G1 G2 G3 i f Hsize -> HlenG3 Hoff Hsusp.
  so (IH1 _#5 Hsize (eq_refl _) HlenG3 Hoff) as H; clear IH1.
  destruct H as (v' & Hsubv & Hofv').
  so (IH2 _#5 Hsize (eq_refl _) HlenG3 Hoff Hsusp) as (s' & Hsubs & Hofs').
  exists (app v' s').
  split; eauto using @rsub; [].
  eapply wsof_app; eauto; done.

  (* pi1 *)
  intros G s A B U Hofs IH G1 G2 G3 i f Hsize -> HlenG3 Hoff Hsusp.
  so (IH _#5 Hsize (eq_refl _) HlenG3 Hoff Hsusp) as H; clear IH.
  destruct H as (s' & Hsubs & Hofs').
  exists (pi1 s').
  split; eauto using @rsub, wsof; done.

  (* pi2 *)
  intros G s A B U Hofs IH G1 G2 G3 i f Hsize -> HlenG3 Hoff Hsusp.
  so (IH _#5 Hsize (eq_refl _) HlenG3 Hoff Hsusp) as H; clear IH.
  destruct H as (s' & Hsubs & Hofs').
  exists (pi2 s').
  split; eauto using @rsub, wsof; done.

  (* etam *)
  intros G r c s Hofr IH G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  so (IH _#5 Hsize (eq_refl _) HlenG3 Hoff suspnormal_susp) as H; clear IH.
  destruct H as (r' & Hsubr & Hofr').
  exists (etam r').
  split; eauto using @rsub, wtof; done.

  (* shiftp *)
  intros G r A Hof IH G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  so (IH _#5 Hsize (eq_refl _) HlenG3 Hoff suspnormal_trp) as H; clear IH.
  destruct H as (r' & Hsubr & Hofr).
  exists (shiftp r').
  split; eauto using @rsub, wtof; done.

  (* lam *)
  intros GG p m p' A B G Hofp Hofm IH G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  #3 exploit (IH G1 G2 (G3;; G) (bound p + i) f) as H; clear IH; auto.
    autorewrite with canonlist; reflexivity.

    rewrite -> app_length; [].
    so (wpof_bound_length _#3 Hofp) as HlenG.
    omega.
  destruct H as (m' & Hsubm & Hofm').
  autorewrite with canonlist in Hofm'.
  exists (lam p m').
  split; eauto using @rsub, wtof; done.

  (* pairm *)
  intros G m n A B Hofm IH1 Hofn IH2 G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  so (IH1 _#5 Hsize (eq_refl _) HlenG3 Hoff) as H; clear IH1.
  destruct H as (m' & Hsubm & Hofm').
  so (IH2 _#5 Hsize (eq_refl _) HlenG3 Hoff) as H; clear IH2.
  destruct H as (n' & Hsubn & Hofn').
  exists (pairm m' n').
  split; eauto using @rsub, wtof; done.

  (* pairms *)
  intros G m n A B Hofm IH1 Hofn IH2 G1 C G2 i f Hsize -> HlenG2 Hoff.
  so (IH1 _#5 Hsize (eq_refl _) HlenG2 Hoff) as H; clear IH1.
  destruct H as (m' & Hsubm & Hofm').
  so (IH2 _#5 Hsize (eq_refl _) HlenG2 Hoff) as H; clear IH2.
  destruct H as (n' & Hsubn & Hofn').
  exists (pairm m' n').
  split; eauto using @rsub, wtof; done.
assert (forall G1 A G2 i f s K L,
          size A N
          -> length G2 = i
          -> wtof G1 f A
          -> wtsof (G1; A;; G2) s K L
          -> exists s',
               rsub f i s s'
               /\ wtsof (G1;; G2) s' K L) as Hts.
  intros G1 G2 G3 i f s K L Hsize HlenG3 Hoff Hofs.
  remember (G1; G2;; G3) as G eqn:HeqG.
  revert G1 G2 G3 i f Hsize HeqG HlenG3 Hoff.
  induct Hofs.
  
  (* tnl *)
  intros G K G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  exists tnl.
  split; eauto using @rsub, wtsof; done.
  
  (* app *)
  intros G v s p A K U Hofv Hofs IH G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  so (Hrvst anderl _ _ _ Hofv _#5 Hsize (eq_refl _) HlenG3 Hoff) as (v' & Hsubv & Hofv').
  so (IH _#5 Hsize (eq_refl _) HlenG3 Hoff) as H; clear IH.
  destruct H as (s' & Hsubs & Hofs').
  exists (tapp v' s').
  split; eauto using @rsub; [].
  eapply wtsof_app; eauto; done.
exploit
  (weak_static_tp_ind
     (fun G A =>
        forall G1 B G2 i f,
          size B N
          -> G = G1; B;; G2
          -> length G2 = i
          -> wtof G1 f B
          -> exists A',
               rsub f i A A'
               /\ wntof (G1;; G2) A')
     (fun G A =>
        forall G1 B G2 i f,
          size B N
          -> G = G1; B;; G2
          -> length G2 = i
          -> wtof G1 f B
          -> exists A',
               rsub f i A A'
               /\ wptof (G1;; G2) A'))
  as Htp;
  [clear IHcn IHcp Hrvst .. |];
  [| clear Hts .. |].
  (* use *)
  intros G c s K K' Hofc Hsim Hofs G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  exploit (Hts G1 G2 G3 i f s K' ty) as (s' & Hsubs & Hofs'); eauto using @rsub; [].
  exists (tuse c s').
  split; eauto using @rsub, rsub_head; [].
  eapply wntof_tuse; eauto; done.
  
  (* up *)
  intros G A HofA IH G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  so (IH _#5 Hsize (eq_refl _) HlenG3 Hoff) as (A' & HsubA & HofA').
  exists (up A').
  split; eauto using @rsub, wntof; done.
  
  (* pi *)
  intros G p A B G4 HofA IH1 Hofp HofB IH2 G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  so (IH1 _#5 Hsize (eq_refl _) HlenG3 Hoff) as (A' & HsubA & HofA').
  so (wpof_similar _#4 Hofp (rsub_similarp _#4 HsubA)) as (G4' & Hofp' & HsimG4).
  #3 exploit (IH2 G1 G2 (G3;; G4) (bound p + i) f); eauto.
    autorewrite with canonlist; reflexivity.
  
    rewrite -> app_length; [].
    so (wpof_bound_length _#3 Hofp) as HlenG.
    omega.
  destruct H as (B' & HsubB & HofB').
  autorewrite with canonlist in HofB'.
  exists (pi p A' B').
  split; eauto using @rsub; [].
  eapply wntof_pi; eauto; [].
  eapply wntof_similarctx; eauto with similar; done.
  
  (* with *)
  intros G A B HofA IH1 HofB IH2 G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  so (IH1 _#5 Hsize (eq_refl _) HlenG3 Hoff) as H; clear IH1.
  destruct H as (A' & HsubA & HofA').
  so (IH2 _#5 Hsize (eq_refl _) HlenG3 Hoff) as H; clear IH2.
  destruct H as (B' & HsubB & HofB').
  exists (wth A' B').
  split; eauto using @rsub, wntof; done.

  (* sigma *)
  intros G A B HofA IH1 HofB IH2 G1 C G2 i f HsizeC -> Hlen Hoff.
  so (IH1 _#5 HsizeC (eq_refl _) Hlen Hoff) as (A' & HsubA & HofA').
  #2 exploit (IH2 G1 C (G2; A) (1 + i) f) as H; eauto.
    simpl; omega.
  destruct H as (B' & HsubB & HofB').
  autorewrite with canonlist in HofB'.
  exists (sigma A' B').
  split; eauto using @rsub; [].
  apply wntof_sigma; auto; [].
  eapply wntof_similarctx; eauto using rsub_similarm with similar; done.

  (* up *)
  intros G A HofA IH G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  so (IH _#5 Hsize (eq_refl _) HlenG3 Hoff) as (A' & HsubA & HofA').
  exists (down A').
  split; eauto using @rsub, wptof; done.
  
  (* one *)
  intros G G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  exists one.
  split; eauto using @rsub, wptof; done.
  
  (* tens *)
  intros G A B HofA IH1 HofB IH2 G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  so (IH1 _#5 Hsize (eq_refl _) HlenG3 Hoff) as H; clear IH1.
  destruct H as (A' & HsubA & HofA').
  so (IH2 _#5 Hsize (eq_refl _) HlenG3 Hoff) as H; clear IH2.
  destruct H as (B' & HsubB & HofB').
  exists (tens A' B').
  split; eauto using @rsub, wptof; done.

  (* ex *)
  intros ? p A B G HofA IH1 Hofp HofB IH2 G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  so (IH1 _#5 Hsize (eq_refl _) HlenG3 Hoff) as H; clear IH1.
  destruct H as (A' & HsubA & HofA').
  so (wpof_similar _#4 Hofp (rsub_similarp _#4 HsubA)) as (G' & Hofp' & HsimG).
  so (app_length G G3) as Hlen.
  #2 sor (refine (IH2 _#5 Hsize _ Hlen Hoff)) as H; clear IH2.
    autorewrite with canonlist; reflexivity.
  destruct H as (B' & HsubB & HofB').
  rewrite -> HlenG3 in HsubB; [].
  rewrite -> (wpof_bound_length _#3 Hofp) in HsubB; [].
  autorewrite with canonlist in HofB'.
  exists (ex p A' B').
  split; eauto using @rsub; [].
  eapply wptof_ex; eauto; [].
  eapply wptof_similarctx; eauto with similar; done.
assert (forall G1 A G2 i f K,
          size A N
          -> length G2 = i
          -> wtof G1 f A
          -> wkof (G1; A;; G2) K
          -> exists K',
               rsub f i K K'
               /\ wkof (G1;; G2) K') as Hk;
  [clear IHcn IHcp Hrvst Hts .. |].
  intros G1 G2 G3 i f K Hsize HlenG3 Hoff HofK.
  remember (G1; G2;; G3) as G eqn:HeqG.
  revert G1 G2 G3 i f Hsize HeqG HlenG3 Hoff.
  induct HofK.
  
  (* ty *)
  intros G G1 G2 G3 i f _ HeqG HlenG3 Hoff.
  exists ty.
  split; eauto using @rsub, wkof; done.

  (* tpi *)
  intros G p A K G4 HofA Hofp HofK IH G1 G2 G3 i f Hsize -> HlenG3 Hoff.
  so (Htp ander _ _ HofA _#5 Hsize (eq_refl _) HlenG3 Hoff) as (A' & HsubA & HofA').
  so (wpof_similar _#4 Hofp (rsub_similarp _#4 HsubA)) as (G4' & Hofp' & HsimG4).
  #3 exploit (IH G1 G2 (G3;; G4) (bound p + i) f); eauto.
    autorewrite with canonlist; reflexivity.
  
    unfold ntp in HlenG3.
    rewrite -> app_length; [].
    so (wpof_bound_length _#3 Hofp) as HlenG.
    omega.
  destruct H as (K' & HsubK & HofK').
  autorewrite with canonlist in HofK'.
  exists (tpi p A' K').
  split; eauto using @rsub; [].
  eapply wkof_tpi; eauto; [].
  eapply wkof_similarctx; eauto with similar; done.
split.
  unfold wcan_rsub_prop.
  intros z G1 A G2 i f e J Hsize Hlen Hoff Hof.
  remember (G1; A;; G2) as G eqn:HeqG.
  revert HeqG.
  #8 cases Hof.
    clear IHcn IHcp Hts Htp Hk.
    intros ? r U Hof Hsusp ->.
    so (Hrvst andel _#3 Hof _#5 Hsize (eq_refl _) Hlen Hoff Hsusp) as (e & Hsub & Hof').
    exists e; split; eauto using @cutp, @wgof; done.

    clear IHcn IHcp Hts Htp Hk.
    intros ? r U Hof ->.
    so (Hrvst anderl _#3 Hof _#5 Hsize (eq_refl _) Hlen Hoff) as (e & Hsub & Hof').
    exists e; split; eauto using @cutp, @wgof; done.

    clear IHcn IHcp Hts Htp Hk.
    intros ? r B U Hof Hsusp ->.
    so (Hrvst anderrl _#4 Hof _#5 Hsize (eq_refl _) Hlen Hoff Hsusp) as (e & Hsub & Hof').
    exists e; split; eauto using @cutp, @wgof; done.

    clear IHcn IHcp Hts Htp Hk.
    intros ? r U Hof ->.
    so (Hrvst anderrr _#3 Hof _#5 Hsize (eq_refl _) Hlen Hoff) as (e & Hsub & Hof').
    exists e; split; eauto using @cutp, @wgof; done.

    clear IHcn IHcp Hrvst Htp Hk.
    intros G s K L Hof ->.
    so (Hts _#8 Hsize Hlen Hoff Hof) as (e & Hsub & Hof').
    exists e; split; eauto using @cutp, @wgof; done.

    clear IHcn IHcp Hrvst Hts Hk.
    intros ? B Hof ->.
    so (Htp andel _ _ Hof _#5 Hsize (eq_refl _) Hlen Hoff) as (e & Hsub & Hof').
    exists e; split; eauto using @cutp, @wgof; done.

    clear IHcn IHcp Hrvst Hts Hk.
    intros ? B Hof ->.
    so (Htp ander _ _ Hof _#5 Hsize (eq_refl _) Hlen Hoff) as (e & Hsub & Hof').
    exists e; split; eauto using @cutp, @wgof; done.

    clear IHcn IHcp Hrvst Hts Htp.
    intros ? K Hof ->.
    so (Hk _#6 Hsize Hlen Hoff Hof) as (e & Hsub & Hof').
    exists e; split; eauto using @cutp, @wgof; done.
clear Hrvst Hts Htp Hk.
#7 exploit
  (weak_static_term_ind
     (fun G r1 U =>
        forall A p G' r2 B,
          size A N
          -> U = trp A
          -> wpof p A G'
          -> wrof (G;; G') r2 (trp B)
          -> exists r,
               lsub r1 p r2 r
               /\ wrof G r (trp B))
     (fun _ _ _ => True)
     (fun G s C U =>
        forall A p G' r B,
          size A N
          -> U = trp A
          -> wpof p A G'
          -> wrof (G;; G') r (trp B)
          -> exists s',
               lsub s p r s'
               /\ wsof G s' C (trp B))
     (fun _ _ _ => True)) as H; try (intros; trivial; done); try (intros; discriminate).
  (* ret *)
  intros G v A Hofv _ A' p G' r U HsizeA Heq Hofp Hofr.
  inversion Heq; clear Heq; [].
  subst A'.
  #2 exploit (wcan_cutpz _ (IHcp N (le_refl _)) re G A p v G' r (judr (trp U))) as H; auto.
    apply wgof_r; auto using suspnormal; done.
  destruct H as (r' & Hsubr & Hgof).
  invert Hgof; [].
  intros Hofr' _.
  exists r'; split; auto; [].
  apply lsub_ret; auto; done.

  (* use *)
  intros G h s A A' ? Hlookup Hsim Hofs IH B p G' r2 U Hsize -> Hofp Hofr2.
  so (IH _#5 Hsize (eq_refl _) Hofp Hofr2) as H.
  destruct H as (s' & Hsubs & Hofs').
  exists (use h s').
  split; eauto using @lsub, wrof; done.

  (* pm *)
  intros G1 p1 r1 A C G2 Hofp1 Hofr1 IH B p2 G3 r2 D Hsize Heq Hofp2 Hofr2.
  injection Heq; clear Heq; intros ->.
  remember (length G2) as i eqn:HlenG2.
  remember (length G3) as j eqn:HlenG3.
  symmetry in HlenG2; symmetry in HlenG3.
  #2 exploit (IH B p2 G3 (sh j i r2) D); auto.
    so (weaken_wrof G1 G2 G3 j i r2 (trp D) HlenG2 HlenG3 Hofr2) as Hofr2w.
    eapply wrof_similarctx; eauto using shctx_similarctx with similar; done.
  destruct H as (r & Hsub & Hofr).
  exists (pm p1 r).
  split.
    apply lsub_pm; [].
    rewrite <- (wpof_bound_length _#3 Hofp1); [].
    rewrite <- (wpof_bound_length _#3 Hofp2); [].
    rewrite -> HlenG3; rewrite -> HlenG2; [].
    exact Hsub.

    eapply wsof_pm; eauto; done.

  (* app *)
  intros G v s p1 A B ? Hofv _ Hofs IH C p2 G' r U HsizeC -> Hofp2 Hofr.
  so (IH _#5 HsizeC (eq_refl _) Hofp2 Hofr) as (s' & Hsub & Hofs').
  exists (app v s').
  split; eauto using @lsub, wsof; done.

  (* pi1 *)
  intros G s A B ? Hof IH C p G' r U Hsize -> Hofp Hofr.
  so (IH _#5 Hsize (eq_refl _) Hofp Hofr) as (s' & Hsub & Hofs').
  exists (pi1 s').
  split; eauto using @lsub, wsof; done.

  (* pi2 *)
  intros G s A B ? Hof IH C p G' r U Hsize -> Hofp Hofr.
  so (IH _#5 Hsize (eq_refl _) Hofp Hofr) as (s' & Hsub & Hofs').
  exists (pi2 s').
  split; eauto using @lsub, wsof; done.
destruct H as (Hr & _ & Hs & _).
split; intro; intros; eauto; done.
Qed.


Lemma can_cutp_v_weak :
  forall G1 A G2 G3 i v p w B,
    length G3 = i
    -> wpof p A G2
    -> wvof G1 v A
    -> wvof (G1;; G2;; G3) w B
    -> exists w',
         cutp v p i w w'
         /\ wvof (G1;; G3) w' B.
Proof.
intros G1 A G2 G3 i v p w B Hlen Hofp Hofv Hofw.
so (ptp_size A) as (N & Hsize).
exploit (cut_elimination_weak N anderl _ G1 A G2 G3 p v i w (judv B)) as (w' & Hcut & Hgof); eauto using @wgof.
invert Hgof; eauto; done.
Qed.


Lemma can_cutp_t_weak :
  forall G1 A G2 G3 i v p m B,
    length G3 = i
    -> wpof p A G2
    -> wvof G1 v A
    -> wtof (G1;; G2;; G3) m B
    -> exists m',
         cutp v p i m m'
         /\ wtof (G1;; G3) m' B.
Proof.
intros G1 A G2 G3 i v p m B Hlen Hofp Hofv Hofm.
so (ptp_size A) as (N & Hsize).
exploit (cut_elimination_weak N anderl _ G1 A G2 G3 p v i m (judt B)) as (w' & Hcut & Hgof); eauto using @wgof.
invert Hgof; eauto; done.
Qed.


Lemma can_cutp_nt_weak :
  forall G1 A G2 G3 i v p B,
    length G3 = i
    -> wpof p A G2
    -> wvof G1 v A
    -> wntof (G1;; G2;; G3) B
    -> exists B',
         cutp v p i B B'
         /\ wntof (G1;; G3) B'.
Proof.
intros G1 A G2 G3 i v p B Hlen Hofp Hofv HofB.
so (ptp_size A) as (N & Hsize).
exploit (cut_elimination_weak N anderl _ G1 A G2 G3 p v i B judnt) as (w' & Hcut & Hgof); eauto using @wgof.
invert Hgof; eauto; done.
Qed.


Lemma can_cutp_pt_weak :
  forall G1 A G2 G3 i v p B,
    length G3 = i
    -> wpof p A G2
    -> wvof G1 v A
    -> wptof (G1;; G2;; G3) B
    -> exists B',
         cutp v p i B B'
         /\ wptof (G1;; G3) B'.
Proof.
intros G1 A G2 G3 i v p B Hlen Hofp Hofv HofB.
so (ptp_size A) as (N & Hsize).
exploit (cut_elimination_weak N anderl _ G1 A G2 G3 p v i B judpt) as (w' & Hcut & Hgof); eauto using @wgof.
invert Hgof; eauto; done.
Qed.


Lemma can_rsub_t_weak :
  forall G1 A G2 i f m B,
    length G2 = i
    -> wtof G1 f A
    -> wtof (G1; A;; G2) m B
    -> exists m',
         rsub f i m m'
         /\ wtof (G1;; G2) m' B.
Proof.
intros G1 A G2 i f m B Hlen HofA Hof.
so (ntp_size A) as (N & Hsize).
exploit (cut_elimination_weak N anderrl _ G1 A G2 i f m (judt B)) as (w' & Hcut & Hgof); eauto using @wgof.
invert Hgof; eauto; done.
Qed.


Lemma can_rsub_nt_weak :
  forall G1 A G2 i f B,
    length G2 = i
    -> wtof G1 f A
    -> wntof (G1; A;; G2) B
    -> exists B',
         rsub f i B B'
         /\ wntof (G1;; G2) B'.
Proof.
intros G1 A G2 i f B Hlen HofA Hof.
so (ntp_size A) as (N & Hsize).
exploit (cut_elimination_weak N anderrl _ G1 A G2 i f B judnt) as (w' & Hcut & Hgof); eauto using @wgof.
invert Hgof; eauto; done.
Qed.
