
Require Import Arith.
Require Import Tactics.
Require Import Sequence.
Require Import Syntax.
Require Import Cut.


Inductive cutpf : forall {z}, nat -> val -> pattern -> nat -> exp z -> exp z -> Prop :=
| cutpf_down {z a m i} {e e' : exp z}
    : rsubf a m i e e'
      -> cutpf a (thunk m) shiftm i e e'

| cutpf_one {z a i} {e : exp z}
    : cutpf a unitp unitpat i e e

| cutpf_tens {z a v v' p p' i} {e e' e'' : exp z}
    : cutpf a v p (bound p' + i) e e'
      -> cutpf a v' p' i e' e''
      -> cutpf a (pairp v v') (pairpat p p') i e e''


with cutmf : nat -> term -> spine -> res -> Prop :=
| cutmf_atom {a r}
    : cutmf a (etam r) nl r

| cutmf_up {a r1 p r2 r}
    : lsubf a r1 p r2 r
      -> cutmf a (shiftp r1) (pm p r2) r

| cutmf_pi {a p m v s r m'}
    : cutpf a v p 0 m m'
      -> cutmf a m' s r
      -> cutmf a (lam p m) (app v s) r

| cutmf_wth1 {a m n s r}
    : cutmf a m s r
      -> cutmf a (pairm m n) (pi1 s) r

| cutmf_wth2 {a m n s r}
    : cutmf a n s r
      -> cutmf a (pairm m n) (pi2 s) r


with rsubf : forall {z}, nat -> term -> nat -> exp z -> exp z -> Prop :=
| rsubf_ret {a f i v v'}
    : rsubf a f i v v'
      -> rsubf a f i (ret v) (ret v')

| rsubf_inst {a f i h s h' s'}
    : rsub_head i h h'
      -> rsubf a f i s s'
      -> rsubf a f i (use h s) (use h' s')

| rsubf_reduce {a f i s r s'}
    : rsubf (S a) f i s s'
      -> cutmf a (sh 0 i f) s' r
      -> rsubf (S a) f i (use (var i) s) r

| rsubf_thunk {a f i n n'}
    : rsubf a f i n n'
      -> rsubf a f i (thunk n) (thunk n')

| rsubf_unitp {a f i}
    : rsubf a f i unitp unitp

| rsubf_pairp {a f i v w v' w'}
    : rsubf a f i v v'
      -> rsubf a f i w w'
      -> rsubf a f i (pairp v w) (pairp v' w')

| rsubf_nl {a f i}
    : rsubf a f i nl nl

| rsubf_pm {a f i p r r'}
    : rsubf a f (bound p + i) r r'
      -> rsubf a f i (pm p r) (pm p r')

| rsubf_app {a f i v s v' s'}
    : rsubf a f i v v'
      -> rsubf a f i s s'
      -> rsubf a f i (app v s) (app v' s')

| rsubf_pi1 {a f i s s'}
    : rsubf a f i s s'
      -> rsubf a f i (pi1 s) (pi1 s')

| rsubf_pi2 {a f i s s'}
    : rsubf a f i s s'
      -> rsubf a f i (pi2 s) (pi2 s')

| rsubf_etam {a f i r r'}
    : rsubf a f i r r'
      -> rsubf a f i (etam r) (etam r')

| rsubf_shiftp {a f i r r'}
    : rsubf a f i r r'
      -> rsubf a f i (shiftp r) (shiftp r')

| rsubf_lam {a f i p n n'}
    : rsubf a f (bound p + i) n n'
      -> rsubf a f i (lam p n) (lam p n')

| rsubf_pairm {a f i n1 n2 n1' n2'}
    : rsubf a f i n1 n1'
      -> rsubf a f i n2 n2'
      -> rsubf a f i (pairm n1 n2) (pairm n1' n2')

| rsubf_tuse {a f i c s s'}
    : rsubf a f i s s'
      -> rsubf a f i (tuse c s) (tuse c s')

| rsubf_up {a f i A A'}
    : rsubf a f i A A'
      -> rsubf a f i (up A) (up A')

| rsubf_pi {a f i p A B A' B'}
    : rsubf a f i A A'
      -> rsubf a f (bound p + i) B B'
      -> rsubf a f i (pi p A B) (pi p A' B')

| rsubf_wth {a f i A B A' B'}
    : rsubf a f i A A'
      -> rsubf a f i B B'
      -> rsubf a f i (wth A B) (wth A' B')

| rsubf_sigma {a f i A B A' B'}
    : rsubf a f i A A'
      -> rsubf a f (1 + i) B B'
      -> rsubf a f i (sigma A B) (sigma A' B')

| rsubf_down {a f i A A'}
    : rsubf a f i A A'
      -> rsubf a f i (down A) (down A')

| rsubf_one {a f i}
    : rsubf a f i one one

| rsubf_tens {a f i A B A' B'}
    : rsubf a f i A A'
      -> rsubf a f i B B'
      -> rsubf a f i (tens A B) (tens A' B')

| rsubf_ex {a f i p A B A' B'}
    : rsubf a f i A A'
      -> rsubf a f (bound p + i) B B'
      -> rsubf a f i (ex p A B) (ex p A' B')

| rsubf_tnl {a f i}
    : rsubf a f i tnl tnl

| rsubf_tapp {a f i v s v' s'}
    : rsubf a f i v v'
      -> rsubf a f i s s'
      -> rsubf a f i (tapp v s) (tapp v' s')

| rsubf_ty {a f i}
    : rsubf a f i ty ty

| rsubf_tpi {a f i p A B A' B'}
    : rsubf a f i A A'
      -> rsubf a f (bound p + i) B B'
      -> rsubf a f i (tpi p A B) (tpi p A' B')

| rsubf_trp {a f i A A'}
    : rsubf a f i A A'
      -> rsubf a f i (trp A) (trp A')

| rsubf_susp {a f i A A'}
    : rsubf a f i A A'
      -> rsubf a f i (susp A) (susp A')

| rsubf_judr {a f i U U'}
    : rsubf a f i U U'
      -> rsubf a f i (judr U) (judr U')

| rsubf_judv {a f i A A'}
    : rsubf a f i A A'
      -> rsubf a f i (judv A) (judv A')

| rsubf_juds {a f i A A' U U'}
    : rsubf a f i A A'
      -> rsubf a f i U U'
      -> rsubf a f i (juds A U) (juds A' U')

| rsubf_judt {a f i A A'}
    : rsubf a f i A A'
      -> rsubf a f i (judt A) (judt A')

| rsubf_judts {a f i K K' L L'}
    : rsubf a f i K K'
      -> rsubf a f i L L'
      -> rsubf a f i (judts K L) (judts K' L')

| rsubf_judnt {a f i}
    : rsubf a f i judnt judnt

| rsubf_judpt {a f i}
    : rsubf a f i judpt judpt

| rsubf_judk {a f i}
    : rsubf a f i judk judk


with lsubf : forall {z} (* re/sp *), nat -> exp z -> pattern -> res -> exp z -> Prop :=
| lsubf_ret {a v p r r'}
    : cutpf a v p 0 r r'
      -> lsubf a (ret v) p r r'

| lsubf_use {a h s p r s'}
    : lsubf a s p r s'
      -> lsubf a (use h s) p r (use h s')

| lsubf_pm {a p1 r1 p2 r2 r}
    : lsubf a r1 p2 (sh (bound p2) (bound p1) r2) r
      -> lsubf a (pm p1 r1) p2 r2 (pm p1 r)

| lsubf_app {a v s p r s'}
    : lsubf a s p r s'
      -> lsubf a (app v s) p r (app v s')

| lsubf_pi1 {a s p r s'}
    : lsubf a s p r s'
      -> lsubf a (pi1 s) p r (pi1 s')

| lsubf_pi2 {a s p r s'}
    : lsubf a s p r s'
      -> lsubf a (pi2 s) p r (pi2 s')
.


Scheme cutpf_mut_ind := Minimality for cutpf Sort Prop
with   cutmf_mut_ind := Minimality for cutmf Sort Prop
with   rsubf_mut_ind := Minimality for rsubf Sort Prop
with   lsubf_mut_ind := Minimality for lsubf Sort Prop.
Combined Scheme substitutionf_ind from cutpf_mut_ind, cutmf_mut_ind, rsubf_mut_ind, lsubf_mut_ind.


Lemma increase :
  (forall z a b p v i (E E' : exp z), a <= b -> cutpf a p v i E E' -> cutpf b p v i E E')
  /\
  (forall a b m s r, a <= b -> cutmf a m s r -> cutmf b m s r)
  /\
  (forall z a b f i (E E' : exp z), a <= b -> rsubf a f i E E' -> rsubf b f i E E')
  /\
  (forall z a b (E : exp z) p r E', a <= b -> lsubf a E p r E' -> lsubf b E p r E').
Proof.
#2 exploit
  (substitutionf_ind
     (fun z a p v i E E' => forall b, a <= b -> cutpf b p v i E E')
     (fun a m s r => forall b, a <= b -> cutmf b m s r)
     (fun z a f i E E' => forall b, a <= b -> @rsubf z b f i E E')
     (fun z a E p r E' => forall b, a <= b -> @lsubf z b E p r E'))
  as Hind;
try (intros; eauto using @cutpf, cutmf, @rsubf, @lsubf; done).

(* use/same *)
intros a f i s r s' _ IH1 _ IH2 b Hinc.
replace b with (S (b-1)) by omega.
#2 eapply rsubf_reduce; eauto.
  apply IH1; eauto; omega.

  apply IH2; eauto; omega.

(* wrapup *)
destruct_all Hind; do2 3 split; intros; eauto; done.
Qed.


Lemma increase_cutpf :
  forall z a b p v i (E E' : exp z), a <= b -> cutpf a p v i E E' -> cutpf b p v i E E'.
Proof.
exact (increase andel).
Qed.


Lemma increase_cutmf :
  forall a b m s r, a <= b -> cutmf a m s r -> cutmf b m s r.
Proof.
exact (increase anderl).
Qed.


Lemma increase_rsubf :
  forall {z} a b f i (E E' : exp z), a <= b -> rsubf a f i E E' -> rsubf b f i E E'.
Proof.
exact (increase anderrl).
Qed.


Lemma increase_lsubf :
  forall {z} a b (E : exp z) p r E', a <= b -> lsubf a E p r E' -> lsubf b E p r E'.
Proof.
exact (increase anderrr).
Qed.


Lemma unfuel :
  (forall z a p v i (E E' : exp z), cutpf a p v i E E' -> cutp p v i E E')
  /\
  (forall a m s r, cutmf a m s r -> cutm m s r)
  /\
  (forall z a f i (E E' : exp z), rsubf a f i E E' -> rsub f i E E')
  /\
  (forall z a (E : exp z) p r E', lsubf a E p r E' -> lsub E p r E').
Proof.
exploit
  (substitutionf_ind
     (fun z a p v i E E' => cutp p v i E E')
     (fun a m s r => cutm m s r)
     (fun z a f i E E' => @rsub z f i E E')
     (fun z a E p r E' => @lsub z E p r E'))
  as Hind;
intros; eauto using @cutp, cutm, @rsub, @lsub.
Qed.


Lemma unfuel_cutp :
  forall z a p v i (E E' : exp z), cutpf a p v i E E' -> cutp p v i E E'.
Proof.
exact (unfuel andel).
Qed.


Lemma unfuel_cutm :
  forall a m s r, cutmf a m s r -> cutm m s r.
Proof.
exact (unfuel anderl).
Qed.


Lemma unfuel_rsub :
  forall {z} a f i (E E' : exp z), rsubf a f i E E' -> rsub f i E E'.
Proof.
exact (unfuel anderrl).
Qed.


Lemma unfuel_lsub :
  forall {z} a (E : exp z) p r E', lsubf a E p r E' -> lsub E p r E'.
Proof.
exact (unfuel anderrr).
Qed.


Lemma fuel :
  (forall z v p i (E E' : exp z), cutp p v i E E' -> exists a, cutpf a p v i E E')
  /\
  (forall m s r, cutm m s r -> exists a, cutmf a m s r)
  /\
  (forall z f i (E E' : exp z), rsub f i E E' -> exists a, rsubf a f i E E')
  /\
  (forall z (E : exp z) p r E', lsub E p r E' -> exists a, lsubf a E p r E').
Proof.
exploit
  (substitution_ind
     (fun z p v i E E' => exists a, cutpf a p v i E E')
     (fun m s r => exists a, cutmf a m s r)
     (fun z f i E E' => exists a, @rsubf z a f i E E')
     (fun z E p r E' => exists a, @lsubf z a E p r E'))
  as Hind;
try (intros;
     repeat
       (match goal with
        | H : exists _, _ |- _ => destruct_all H
        end);
     eexists;
     eauto using @cutpf, cutmf, @rsubf, @lsubf, increase_cutpf, increase_cutmf, increase_rsubf, increase_lsubf, le_plus_l, le_plus_r;
     done); [].
destruct_all Hind.
do2 3 split; intros; eauto; done.

Grab Existential Variables.
exact 0.
exact 0.
exact 0.
exact 0.
exact 0.
exact 0.
exact 0.
exact 0.
exact 0.
exact 0.
exact 0.
exact 0.
Qed.


Lemma fuel_cutp :
  forall z v p i (E E' : exp z), cutp p v i E E' -> exists a, cutpf a p v i E E'.
Proof.
exact (fuel andel).
Qed.


Lemma fuel_cutm :
  forall m s r, cutm m s r -> exists a, cutmf a m s r.
Proof.
exact (fuel anderl).
Qed.


Lemma fuel_rsub :
  forall {z} f i (E E' : exp z), rsub f i E E' -> exists a, rsubf a f i E E'.
Proof.
exact (fuel anderrl).
Qed.


Lemma fuel_lsub :
  forall {z} (E : exp z) p r E', lsub E p r E' -> exists a, lsubf a E p r E'.
Proof.
exact (fuel anderrr).
Qed.
