
Require Import Arith.
Require Import Compare.
Require Import Tactics.
Require Import Sequence.
Require Import Syntax.
Require Import Cut.
Require Import Self.
Require Import Static.
Require Import Substitution.
Require Import Weakening.
Require Import Size.
Require Import Compose.
Require Import Expand.
Require Import Similar.
Require Import StaticWeak.
Require Import IdentityWeak.
Require Import CutEliminationWeak.
Require Import Regularity.


Lemma compose_formation :
  forall G A B U s1 s2 s,
    compose s1 s2 s
    -> sof G s1 A (susp B)
    -> sof G s2 B U
    -> sof G s A U.
Proof.
cut (forall z (h : z = sp) G A B U (s1 : exp z) s2 s,
       sof G (eq_rect _ _ s1 _ h) A (susp B)
       -> sof G s2 B U
       -> compose s1 s2 s
       -> sof G (eq_rect _ _ s _ h) A U).
  intro Hind.
  intros; eapply (Hind sp (eq_refl _)); eauto; done.
intros z h G A B U s1 s2 s Hofs1 Hofs2 Hcompose.     
revert h G A B U Hofs1 Hofs2.
#5 induct Hcompose; revertnew (intros; substrefls; try discriminate).

(* nl *)
intros s G A B U Hofs2 Hofs1.
invert Hofs1; [].
intros <-.
exact Hofs2.

(* pm *)
intros p r1 s2 r _ _ G A B U Hofs2 Hofs1.
invert Hofs1; done.

(* app *)
intros v s1 s2 s _ IH G AA C U Hofs2 Hofsp.
invertc Hofsp; [].
intros p A B B' Hofv HsubB Hofs1 <-.
so (IH (eq_refl _) _#4 Hofs1 Hofs2) as Hofs.
simpl in Hofs.
eapply sof_app; eauto; done.

(* pi1 *)
intros s1 s2 s _ IH G AA C U Hofs2 Hofsp.
invertc Hofsp; [].
intros A B Hofs1 <-.
so (IH (eq_refl _) _#4 Hofs1 Hofs2) as Hofs.
simpl in Hofs.
apply sof_pi1; auto; done.

(* pi2 *)
intros s1 s2 s _ IH G AA C U Hofs2 Hofsp.
invertc Hofsp; [].
intros A B Hofs1 <-.
so (IH (eq_refl _) _#4 Hofs1 Hofs2) as Hofs.
simpl in Hofs.
apply sof_pi2; auto; done.
Qed.


Lemma rcompose_formation :
  forall G A U r s r',
    compose r s r'
    -> rof G r (susp A)
    -> sof G s A U
    -> rof G r' U.
Proof.
intros G A U r s2 r' Hcompose Hofr Hofs2.
invertc Hofr; [].
intros h s1 B Hofh Hofs1 <-.
invert Hcompose; [].
intros s Hcompose' <-.
so (compose_formation _#7 Hcompose' Hofs1 Hofs2) as Hofs.
eapply rof_use; eauto; done.
Qed.


Lemma self_fun :
  forall m A B B',
    self m A B
    -> self m A B'
    -> B = B'.
Proof.
intros m A B C HselfB HselfC.
revert C HselfC.
induct HselfB; try (intros; invertc HselfC; intros; subst; f_equal; eauto; done); [].
(* sigma *)
intros m n A B C B' D HselfA IH1 HsubB HselfB IH2 AA Hselfty.
invertc Hselfty; [].
intros C' B'' D' HselfA' HsubB' HselfB' <-.
so (IH1 _ HselfA'); subst C'.
so (rsub_fun _#5 HsubB HsubB'); subst B''.
so (IH2 _ HselfB'); subst D'.
reflexivity.
Qed.


Lemma self_idem :
  forall m A B C,
    self m A B
    -> self m B C
    -> B = C.
Proof.
intros m A B C HselfA HselfB.
revert C HselfB.
induct HselfA; intros; invertc HselfB; intros; subst; f_equal; eauto.
Qed.


Lemma self_member :
  forall G m A A',
    self m A A'
    -> tof G m A
    -> tof G m A'.
Proof.
intros G m A A' Hself Hofm.
revert A' Hself.
induct Hofm; try (intros; invertc Hself; intros; subst; eauto using tof; done).

(* sigma *)
intros G m n A B B' Hofm IH1 HsubB Hofn IH2 AA Hself.
invertc Hself; [].
intros A' B'' C HselfA HsubB' HselfB <-.
so (rsub_fun _#5 HsubB HsubB'); subst B''.
clear HsubB'.
apply tof_pairm; auto; done.
Qed.


Lemma self_member_rev :
  forall G m A A',
    self m A A'
    -> tof G m A'
    -> tof G m A.
Proof.
intros G m A A' Hself Hofm.
revert A Hself.
induct Hofm; try (intros; invertc Hself; intros; subst; eauto using tof; done).
Qed.    


Lemma sub_into_self :
  forall f i m m' A A' B,
    self m A B
    -> rsub f i m m'
    -> rsub f i A A'
    -> exists B',
         self m' A' B'
         /\ rsub f i B B'.
Proof.
intros f i m m' A A' B Hself Hsubm HsubA.
revert i m' A' Hsubm HsubA.
induct Hself.

(* atom *)
intros m c s i m' AA Hsubm Hsubty.
invertc Hsubty; [].
intros s' Hsubs <-.
exists (tuse c s').
split; eauto using self, @rsub; done.

(* up *)
intros m A i m' AA Hsubm Hsubty.
invertc Hsubty; [].
intros A' HsubA <-.
exists (up A').
split; eauto using self, @rsub; done.

(* pi *)
intros p m A B C HselfB IH i mm AA Hsub Hsubty.
invertc Hsubty; [].
intros A' B' HsubA HsubB <-.
invertc Hsub; [].
intros m' Hsubm <-.
so (IH _#3 Hsubm HsubB) as (C' & HselfB' & HsubC).
exists (pi p A' C').
split; eauto using self, @rsub; done.

(* wth *)
intros m n A B C D HselfA IH1 HselfB IH2 i mm AA Hsub Hsubty.
invertc Hsubty; [].
intros A' B' HsubA HsubB <-.
invertc Hsub; [].
intros m' n' Hsubm Hsubn <-.
so (IH1 _#3 Hsubm HsubA) as (C' & HselfA' & HsubC).
so (IH2 _#3 Hsubn HsubB) as (D' & HselfB' & HsubD).
eauto using self, @rsub; done.

(* sigma *)
intros m n A B C B1 D HselfA IH1 HsubB1 HselfB IH2 i mm AA Hsub Hsubty.
invertc Hsubty; [].
intros A' B2 HsubA HsubB2 <-.
invertc Hsub; [].
intros m' n' Hsubm Hsubn <-.
so Hsubm as Hsubm'.
repl i with (S i - 0 - 1) in Hsubm' by omega.
so (commute_rsub _#8 (lt_0_Sn _) HsubB2 HsubB1 Hsubm') as (B' & HsubB2' & HsubB1').
clear Hsubm'.
repl (S i - 1) with i in HsubB1' by omega.
so (IH1 _#3 Hsubm HsubA) as (C' & HselfA' & HsubC).
so (IH2 _#3 Hsubn HsubB1') as (D' & HselfB' & HsubD).
eauto using self, @rsub; done.
Qed.


Lemma cutp_into_self :
  forall v p i m m' A A' B,
    self m A B
    -> cutp v p i m m'
    -> cutp v p i A A'
    -> exists B',
         self m' A' B'
         /\ cutp v p i B B'.
Proof.
intros v p i m m' A A' B Hself Hsubm HsubA.
revert v i m m' A A' B Hself Hsubm HsubA.
induct p.

(* down *)
intros v i m m' A A' B Hself Hsubm HsubA.
invertc Hsubm; [].
intros f Hsubm' <-.
invertc HsubA; [].
intros HsubA'.
so (sub_into_self _#7 Hself Hsubm' HsubA') as (B' & Hself' & HsubB).
eauto using @cutp; done.

(* one *)
intros v i m m' A A' B Hself Hsubm HsubA.
invertc Hsubm; [].
intros <- <-.
invertc HsubA; [].
intros <-.
eauto using @cutp; done.

(* tens *)
intros p IH1 q IH2 vv i m m'' A A'' B Hself Hsubm HsubA.
invertc Hsubm; [].
intros v w m' Hsubm Hsubm' <-.
invertc HsubA; [].
intros A' HsubA HsubA'.
so (IH1 _#7 Hself Hsubm HsubA) as (B' & Hself' & HsubB).
so (IH2 _#7 Hself' Hsubm' HsubA') as (B'' & Hself'' & HsubB').
eauto using @cutp; done.
Qed.


Lemma self_formation :
  forall G m A,
    wtof G m A
    -> wntof G A
    -> exists B,
         self m A B
         /\ wntof G B.
Proof.
intros G m A Hofm HofA.
so (ntp_size A) as (N & Hsize).
revert G m A Hsize Hofm HofA.
wfinduct N using lt_wf.
intros N IH G m A Hsize Hofm HofA.
revert m Hsize Hofm.
cases HofA.

(* atom *)
intros G c s K K' Hofc Hsim Hofs m _ _.
exists (tuse c s).
eauto using self, wntof; done.

(* up *)
intros G A HofA m Hsize Hofm.
exists (up A).
eauto using self, wntof; done.

(* pi *)
intros G p A B G' HofA Hofp HofB mm Hsize Hoftm.
invertc Hoftm; [].
intros p' m G'' Hofp' Hofm <-.
so (wpof_fun _#5 Hofp Hofp') as (<- & <-).
invertc Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
sor (refine (IH N2 _ _ _ _ HsizeB Hofm HofB)) as (C & Hself & HofC); [omega |].
exists (pi p A C).
eauto using self, wntof; done.

(* wth *)
intros G A B HofA HofB mm Hsize Hoftm.
invertc Hoftm; [].
intros m n Hofm Hofn <-.
invertc Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
sor (refine (IH N1 _ _ _ _ HsizeA Hofm HofA)) as (C & HselfA & HofC); [omega |].
sor (refine (IH N2 _ _ _ _ HsizeB Hofn HofB)) as (D & HselfB & HofD); [omega |].
exists (wth C D).
eauto using self, wntof; done.

(* sigma *)
intros G A B HofA HofB mm Hsize Hoftm.
invertc Hoftm; [].
intros m n Hofm Hofn <-.
invertc Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
exploit (can_rsub_nt_weak G A nil 0 m B) as (B' & HsubB & HofB'); eauto using rsubctx; [].
simpl in HofB'.
so (rsub_size _#5 HsubB HsizeB) as HsizeB'.
so (wtof_similar _#4 Hofn (rsub_similarm _#4 HsubB)) as Hofn'.
sor (refine (IH N1 _ _ _ _ HsizeA Hofm HofA)) as (C & HselfA & HofC); [omega |].
sor (refine (IH N2 _ _ _ _ HsizeB' Hofn' HofB')) as (D & HselfB & HofD); [omega |].
exists (wth C D).
eauto using self, wntof; done.
Qed.


Lemma self_spine :
  forall G A B r s m A',
    expandm A r m
    -> self m A A'
    -> wwf nil G  (* this could be weakened, but we have it and it's convenient *)
    -> rof G r (susp A')
    -> sof G s A (susp B)
    -> exists r' n B',
         compose r s r'
         /\ expandm B r' n
         /\ self n B B'
         /\ sof G s A' (susp B').
Proof.
intros G A B r s m A' HexpandA Hself Hwf Hofr Hsof.
remember (susp B) as U eqn:HeqU.
revert B r m A' HexpandA Hself Hwf Hofr HeqU.
induct Hsof.

(* nl *)
intros G A B r m A' HexpandA Hself _ Hofr HeqU.
injection HeqU.
intros <-.
so (rof_nres _#3 Hofr) as His.
so (nres_compose_nl _ His) as Hcompose.
exists r, m, A'.
eauto using sof; done.

(* pm *)
intros G p r A B G' _ _ C r' m A' HexpandA Hself Hwf Hofr HeqU.
discriminate HeqU.

(* pi *)
intros G v s p A B U B' Hofv HsubB Hofs IH D r mm AA Hexpand Hself Hwf Hofr ->.
invertc Hexpand; [].
intros p' w rw m HexpandA Hcomposerw HexpandB <-.
invertc Hself; [].
intros C HselfB -> <-.
so (rof_nres _#3 Hofr) as Hisr.
invert (rof_reg_weak _#3 Hwf Hofr); [].
intros G' _ Hofp _.
so (expandp_formation_weak G _#4 HexpandA Hofp) as Hofw.
exploit (can_cutp_v_weak G A G' nil 0 v p w A) as (w' & Hsubw & _); eauto using vof_weak; [].
so (sub_into_expandp _#5 HexpandA Hsubw); subst w'.
so (cutp_sh_cancel_all _ _ 0 r (cutp_patmatch _#5 HsubB)) as Hsubr.
#2 sor (refine (cutp_compat1 _ _ 0 (fun x => app x nl) _ _ Hsubw _)) as Hsubwnl.
  eauto using @rsub; done.
simpl in Hsubwnl.
#2 sor (refine (cutp_into_rcompose _#8 (sh_nres _#3 Hisr) Hcomposerw _ Hsubr Hsubwnl)) as (rv & Hsubrw & Hcomposerv).
  apply sh_implies_rigid; auto; omega.
so (compose_nres_nspine _#3 Hcomposerw (sh_nres _#3 Hisr) (nspine_app nspine_nl)) as Hisrw.
so (weakenn_rof _#5 Hofr (wpof_bound_length _#3 Hofp)) as Hofr_w.
simpl in Hofr_w.
so (rcompose_formation_weak _#6 Hcomposerw (rof_weak _#3 Hofr_w) (wsof_app (wvof_similar _#4 Hofw (sh_similarp 0 (bound p) _)) wsof_nl)) as Hofrw.
clear Hofr_w.
#2 sor (refine (wrof_similar _#3 (susp B) Hofrw _)) as Hofrw'.
  apply similar_susp; [].
  apply similar_symm; [].
  #2 eapply similar_trans.
    eapply self_similar; eauto; done.

    apply sh_similarm; eauto; done.
clear Hofrw.
so (expandm_formation_weak _#4 Hofrw' HexpandB) as Hofm.
clear Hofrw'.
exploit (can_cutp_t_weak G A G' nil 0 v p m B) as (m' & Hsubm & _); eauto using vof_weak; [].
#2 sor (refine (cutp_into_expandm _#8 Hisrw _ HexpandB Hsubrw Hsubm)) as HexpandB'.
  eapply rigid_compose; eauto using sh_implies_rigid; done.
so (cutp_into_self _#8 HselfB Hsubm HsubB) as (C' & HselfB' & HsubC).
so (rcompose_formation _#6 Hcomposerv Hofr (sof_app Hofv HsubC sof_nl)) as Hofrv.
so (expandm_similar _#4 (cutp_similarm _#5 HsubB) HexpandB') as HexpandB''.
so (IH _#4 HexpandB'' HselfB' Hwf Hofrv (eq_refl _)) as (rvs & n & D' & Hcomposervs & HexpandD & HselfD & Hofs').
exists rvs, n, D'.
do2 3 split; eauto using sof; [].
so (compose_assoc _#5 Hcomposerv Hcomposervs) as (vs & Hcomposevs & Hcomposer_vs).
invertc Hcomposevs; [].
intros s' Hcomposenls <-.
invertc Hcomposenls; [].
intros <-.
exact Hcomposer_vs.

(* pi1 *)
intros G s A B ? Hofs IH E r mm AA Hexpand Hself Hwf Hofr ->.
invertc Hexpand; [].
intros r1 r2 m1 m2 Hcompose1 Hcompose2 HexpandA HexpandB <-.
invertc Hself; [].
intros C D HselfA HselfB <-.
so (rcompose_formation _#6 Hcompose1 Hofr (sof_pi1 sof_nl)) as Hofr1.
so (IH _#4 HexpandA HselfA Hwf Hofr1 (eq_refl _)) as (r' & n & E' & Hcompose1s & HexpandE & HselfE & Hofs').
exists r', n, E'.
do2 3 split; eauto using sof; [].
so (compose_assoc _#5 Hcompose1 Hcompose1s) as (s1 & Hcomposes1 & Hcomposer_1s).
invertc Hcomposes1; [].
intros s' Hcomposenls <-.
invertc Hcomposenls; [].
intros <-.
exact Hcomposer_1s.

(* pi2 *)
intros G s A B ? Hofs IH E r mm AA Hexpand Hself Hwf Hofr ->.
invertc Hexpand; [].
intros r1 r2 m1 m2 Hcompose1 Hcompose2 HexpandA HexpandB <-.
invertc Hself; [].
intros C D HselfA HselfB <-.
so (rcompose_formation _#6 Hcompose2 Hofr (sof_pi2 sof_nl)) as Hofr2.
so (IH _#4 HexpandB HselfB Hwf Hofr2 (eq_refl _)) as (r' & n & E' & Hcompose2s & HexpandE & HselfE & Hofs').
exists r', n, E'.
do2 3 split; eauto using sof; [].
so (compose_assoc _#5 Hcompose2 Hcompose2s) as (s2 & Hcomposes2 & Hcomposer_2s).
invertc Hcomposes2; [].
intros s' Hcomposenls <-.
invertc Hcomposenls; [].
intros <-.
exact Hcomposer_2s.
Qed.


Lemma self_head :
  forall G h A m A',
    hof G h A
    -> expandm A (use h nl) m
    -> self m A A'
    -> hof G h A'.
Proof.
intros G h A m A' Hofh Hexpand Hself.
revert m A' Hexpand Hself.
cases Hofh.

(* id *)
intros G h A Hlookup m A' Hexpand Hself.
eapply hof_self; eauto; done.

(* self *)
intros G h m A B Hlookup Hexpand HselfA m' C Hexpand' HselfB.
so (expandm_similar_fun _#5 (self_similar _#3 HselfA) Hexpand Hexpand'); subst m'.
so (self_idem _#4 HselfA HselfB); subst C.
eapply hof_self; eauto; done.
Qed.


Lemma selfify :
  forall G r A m,
    wwf nil G
    -> rof G r (susp A)
    -> expandm A r m
    -> exists A',
         self m A A'
         /\ wntof G A'
         /\ rof G r (susp A').
Proof.
intros G r A m Hwf Hofr HexpandA.
so (rof_nres _#3 Hofr) as Hisr.
so (expandm_formation_weak _#4 (rof_weak _#3 Hofr) HexpandA) as Hofm.
so (rof_reg_weak _#3 Hwf Hofr) as HofA.
so (self_formation _#3 Hofm HofA) as (A' & HselfA & HofA').
exists A'.
do2 2 split; auto; [].
invert Hofr; [].
intros h s B Hofh Hofs <-.
so (can_expandm B (use h nl) (nres_use nspine_nl)) as (n & HexpandB).
so (expandm_formation_weak _#4 (rof_weak _#3 (rof_use Hofh sof_nl)) HexpandB) as Hofn.
so (hof_reg_weak _#3 Hwf Hofh) as HofB.
so (self_formation _#3 Hofn HofB) as (B' & HselfB & _).
so (self_head _#5 Hofh HexpandB HselfB) as Hofh'.
so (self_spine _#7 HexpandB HselfB Hwf (rof_use Hofh' sof_nl) Hofs) as (r' & m' & A'' & Hcompose & HexpandA' & HselfA' & Hofs').
invertc Hcompose; [].
intros s' Hcompose' <-.
invertc Hcompose'; [].
intros <-.
so (expandm_fun _#4 HexpandA HexpandA'); subst m'.
so (self_fun _#4 HselfA HselfA'); subst A''.
eapply rof_use; eauto; done.
Qed.
