
Require Import Arith.
Require Import Compare.
Require Import Tactics.
Require Import Sequence.
Require Import Syntax.
Require Import Cut.
Require Import Self.
Require Import Static.
Require Import Substitution.
Require Import Compose.


Inductive similar : forall {z}, exp z -> exp z -> Prop :=
| similar_atom {c c' s s'}
    : similar (tuse c s) (tuse c' s')

| similar_up {A A'}
    : similar A A'
      -> similar (up A) (up A')

| similar_pi {p p' A A' B B'}
    : similar A A'
      -> similar B B'
      -> similar (pi p A B) (pi p' A' B')

| similar_wth {A A' B B'}
    : similar A A'
      -> similar B B'
      -> similar (wth A B) (wth A' B')

| similar_sigma {A A' B B'}
    : similar A A'
      -> similar B B'
      -> similar (sigma A B) (sigma A' B')

| similar_wth_sigma {A A' B B'}
    : similar A A'
      -> similar B B'
      -> similar (wth A B) (sigma A' B')

| similar_sigma_wth {A A' B B'}
    : similar A A'
      -> similar B B'
      -> similar (sigma A B) (wth A' B')

| similar_down {A A'}
    : similar A A'
      -> similar (down A) (down A')

| similar_one
    : similar one one

| similar_tens {A A' B B'}
    : similar A A'
      -> similar B B'
      -> similar (tens A B) (tens A' B')

| similar_ex {p p' A A' B B'}
    : similar A A'
      -> similar B B'
      -> similar (ex p A B) (ex p' A' B')

| similar_tens_ex {p A A' B B'}
    : similar A A'
      -> similar B B'
      -> similar (tens A B) (ex p A' B')

| similar_ex_tens {p A A' B B'}
    : similar A A'
      -> similar B B'
      -> similar (ex p A B) (tens A' B') 

| similar_ty
    : similar ty ty

| similar_tpi {p p' A A' K K'}
    : similar A A'
      -> similar K K'
      -> similar (tpi p A K) (tpi p' A' K')

| similar_susp {A A'}
    : similar A A'
      -> similar (susp A) (susp A')

| similar_trp {A A'}
    : similar A A'
      -> similar (trp A) (trp A')
.


Lemma sh_similar :
  forall z (A : exp z),
    (z = nt -> forall i j, similar A (sh i j A))
    /\ (z = pt -> forall i j, similar A (sh i j A))
    /\ (z = cn -> forall i j, similar A (sh i j A))
    /\ (z = kd -> forall i j, similar A (sh i j A)).
Proof.
intros z A.
induct A;
intros;
do2 3 split;
intros; 
try discriminate;
repeat
  (match goal with
   | H : _ /\ _ |- _ => destruct H as (? & ? & ? & ?)
   end);
eauto using @similar.
Qed.


Lemma sh_similarm :
  forall i j (A : ntp), similar A (sh i j A).
Proof.
intros i j A.
exact (sh_similar _ A andel (eq_refl _) i j).
Qed.
    

Lemma sh_similarp :
  forall i j (A : ptp), similar A (sh i j A).
Proof.
intros i j A.
exact (sh_similar _ A anderl (eq_refl _) i j).
Qed.


Lemma sh_similarcn :
  forall i j (U : con), similar U (sh i j U).
Proof.
intros i j A.
exact (sh_similar _ A anderrl (eq_refl _) i j).
Qed.


Lemma sh_similarkd :
  forall i j (K : kind), similar K (sh i j K).
Proof.
intros i j A.
exact (sh_similar _ A anderrr (eq_refl _) i j).
Qed.


Lemma similarm_refl :
  forall (A : ntp), similar A A.
Proof.
intro A.
so (sh_similarm 0 0 A) as H.
rewrite -> sh_zero in H.
exact H.
Qed.


Lemma similarp_refl :
  forall (A : ptp), similar A A.
Proof.
intro A.
so (sh_similarp 0 0 A) as H.
rewrite -> sh_zero in H.
exact H.
Qed.


Lemma similarcn_refl :
  forall (U : con), similar U U.
Proof.
intro U.
so (sh_similarcn 0 0 U) as H.
rewrite -> sh_zero in H.
exact H.
Qed.


Lemma similarkd_refl :
  forall (K : kind), similar K K.
Proof.
intro U.
so (sh_similarkd 0 0 U) as H.
rewrite -> sh_zero in H.
exact H.
Qed.


Lemma similar_symm :
  forall {z} (A B : exp z), similar A B -> similar B A.
Proof.
intros z A B H.
induct H; eauto using @similar.
Qed.


Lemma similar_trans :
  forall {z} (A B C : exp z), similar A B -> similar B C -> similar A C.
Proof.
intros z A B C H Hsimilar.
revert C Hsimilar.
induct H;
intros;
invertc Hsimilar;
intros;
subst;
eauto using @similar.
Qed.


Ltac destruct_everywhere :=
  repeat
    (match goal with
     | H : exists _,_ |- _ => destruct H
     | H : _ /\ _ |- _ => destruct H
     | H : prod _ _ |- _ => destruct H
     end).


Lemma rsub_similar :
  forall z f i (A A' : exp z),
    rsub f i A A'
    -> (z = nt -> similar A A')
       /\ (z = pt -> similar A A')
       /\ (z = cn -> similar A A')
       /\ (z = kd -> similar A A').
Proof.
intros z f i A A' H.
induct H;
intros;
do2 3 split;
intro;
try discriminate;
destruct_everywhere;
eauto using @similar.
Qed.


Lemma rsub_similarm :
  forall f i (A A' : ntp),
    rsub f i A A'
    -> similar A A'.
Proof.
intros f i A A' H.
exact (rsub_similar _#5 H andel (eq_refl _)).
Qed.


Lemma rsub_similarp :
  forall f i (A A' : ptp),
    rsub f i A A'
    -> similar A A'.
Proof.
intros f i A A' H.
exact (rsub_similar _#5 H anderl (eq_refl _)).
Qed.


Lemma rsub_similarcn :
  forall f i (U U' : con),
    rsub f i U U'
    -> similar U U'.
Proof.
intros f i A A' H.
exact (rsub_similar _#5 H anderrl (eq_refl _)).
Qed.


Lemma rsub_similarkd :
  forall f i (K K' : kind),
    rsub f i K K'
    -> similar K K'.
Proof.
intros f i A A' H.
exact (rsub_similar _#5 H anderrr (eq_refl _)).
Qed.


Lemma cutp_similarm :
  forall v p i (A A' : ntp),
    cutp v p i A A'
    -> similar A A'.
Proof.
intros v p i A A' H.
unfold ntp in A, A'.
remember nt as z eqn:h.
revert h.
induct H; intros; subst; eauto using rsub_similarm, similarm_refl, similar_trans.
Qed.


Lemma cutp_similarp :
  forall v p i (A A' : ptp),
    cutp v p i A A'
    -> similar A A'.
Proof.
intros v p i A A' H.
unfold ptp in A, A'.
remember pt as z eqn:h.
revert h.
induct H; intros; subst; eauto using rsub_similarp, similarp_refl, similar_trans.
Qed.


Lemma cutp_similarkd :
  forall v p i (K K' : kind),
    cutp v p i K K'
    -> similar K K'.
Proof.
intros v p i A A' H.
unfold kind in A, A'.
remember kd as z eqn:h.
revert h.
induct H; intros; subst; eauto using rsub_similarkd, similarkd_refl, similar_trans.
Qed.


Lemma self_similar :
  forall m A A',
    self m A A'
    -> similar A A'.
Proof.
intros m A A' H.
induct H; eauto using @similar, similarp_refl, similarm_refl, rsub_similarm, similar_trans.
Qed.


Definition similarctx (G G' : context) :=
  Forall2 (@similar nt) G G'.


Lemma similarctx_cons :
  forall G G' A A',
    similarctx G G'
    -> similar A A'
    -> similarctx (G; A) (G'; A').
Proof.
intros G G' A A' HsimG HsimA.
apply Forall2_cons; auto.
Qed.


Lemma similarctx_append :
  forall G1 G1' G2 G2',
    similarctx G1 G1'
    -> similarctx G2 G2'
    -> similarctx (G1;; G2) (G1';; G2').
Proof.
intros G1 G1' G2 G2' Hsim1 Hsim2.
induct Hsim2.

(* nil *)
exact Hsim1.

(* cons *)
intros A A' G2 G2' Hsim _ IH.
simpl.
apply Forall2_cons; auto.
Qed.


Lemma similarctx_length :
  forall G G',
    similarctx G G'
    -> length G = length G'.
Proof.
intros G G' H.
induct H; intros; simpl; eauto; f_equal; eauto.
Qed.


Lemma similarctx_refl :
  forall G, similarctx G G.
Proof.
intros H.
unfold similarctx.
induct H; intros; eauto using similarm_refl, Forall2.
Qed.


Lemma similarctx_symm :
  forall G G', similarctx G G' -> similarctx G' G.
Proof.
unfold similarctx.
intros G G' H.
induct H; eauto using similar_symm, Forall2.
Qed.


Lemma similarctx_trans :
  forall G1 G2 G3, similarctx G1 G2 -> similarctx G2 G3 -> similarctx G1 G3.
Proof.
unfold similarctx.
intros G1 G2 G3 H Hsim.
revert G3 Hsim.
induct H; intros; invertc Hsim; intros; subst; eauto using similar_trans, Forall2.
Qed.


Lemma shctx_similarctx :
  forall i j G,
    similarctx G (shctx i j G).
Proof.
intros i j G.
unfold similarctx.
induct G; intros; simpl; eauto using sh_similarm, Forall2.
Qed.


Lemma rsubctx_similarctx :
  forall f i G G',
    rsubctx f i G G'
    -> similarctx G G'.
Proof.
intros f i G G' H.
unfold similarctx.
induct H; intros; eauto using rsub_similarm, Forall2.
Qed.


Hint Constructors similar : similar.
Hint Resolve similarm_refl similarp_refl similarcn_refl similarkd_refl similar_symm similar_trans : similar.
Hint Resolve similarctx_cons similarctx_append similarctx_refl similarctx_symm : similar.
