
Require Import Tactics.
Require Import Sequence.
Require Import Syntax.
Require Import Cut.
Require Import Self.
Require Import Static.
Require Import Similar.


(* This could be a function instead of a relation now. *)

Inductive size : forall {z}, exp z -> nat -> Prop :=
| size_tuse {c s}
    : size (tuse c s) 1

| size_up {A N}
    : size A N
      -> size (up A) (S N)

| size_pi {p A B N1 N2}
    : size A N1
      -> size B N2
      -> size (pi p A B) (1 + N1 + N2)

| size_wth {A B N1 N2}
    : size A N1
      -> size B N2
      -> size (wth A B) (1 + N1 + N2)

| size_sigma {A B N1 N2}
    : size A N1
      -> size B N2
      -> size (sigma A B) (1 + N1 + N2)

| size_down {A N}
    : size A N
      -> size (down A) (S N)

| size_one
    : size one 1

| size_tens {A B N1 N2}
    : size A N1
      -> size B N2
      -> size (tens A B)  (1 + N1 + N2)

| size_ex {p A B N1 N2}
    : size A N1
      -> size B N2
      -> size (ex p A B)  (1 + N1 + N2)
.


Lemma shift_size :
  forall {z} i j (A : exp z) N,
    size A N
    -> size (sh i j A) N.
Proof.
intros z i j A N H.
revert i j.
#5 induct H; try (intros; simpl; eauto using @size; done).

(* pi *)
intros.
simpl.
apply size_pi; eauto; done.

(* wth *)
intros.
simpl.
apply size_wth; eauto; done.

(* sigma *)
intros.
simpl.
apply size_sigma; eauto; done.

(* tens *)
intros.
simpl.
apply size_tens; eauto; done.

(* ex *)
intros.
simpl.
apply size_ex; eauto; done.
Qed.


Lemma rsub_size :
  forall {z} f i (A A' : exp z) N,
    rsub f i A A'
    -> size A N
    -> size A' N.
Proof.
intros z f i A A' N Hsub Hsize.
revert N Hsize.
induct Hsub;
intros; invert Hsize; intros; subst; eauto using @size.
Qed.


Lemma cutp_size :
  forall {z} v p i (A A' : exp z) N,
    cutp v p i A A'
    -> size A N
    -> size A' N.
Proof.
intros z v p i A A' N HsubA Hsize.
revert N Hsize.
induct HsubA; intros; eauto using rsub_size.
Qed.


Lemma tp_size :
  (forall (A : ntp), exists N, size A N)
  /\
  (forall (A : ptp), exists N, size A N).
Proof.
cut (forall z (e : exp z),
       (forall (h : z = nt), exists N, size (eq_rect _ _ e _ h) N)
       /\ (forall (h : z = pt), exists N, size (eq_rect _ _ e _ h) N)).
  intros Hind.
  split.
    intros A; exact (Hind _ A andel (eq_refl _)).

    intros A; exact (Hind _ A ander (eq_refl _)).
intros z e.
induct e;
intros; split; intros; try discriminate;
repeat
  match goal with
  | H : ?x = ?x |- _ => substrefl H
  | H : _ /\ _ |- _ => destruct H
  | H : forall (_ : ?x = ?x), _ |- _ => so (H (eq_refl _)); clear H
  | H : exists _, _ |- _ => destruct H
  end;
eexists; eauto using @size.
Qed.


Lemma ntp_size :
  forall (A : ntp), exists N, size A N.
Proof.
exact (tp_size andel).
Qed.


Lemma ptp_size :
  forall (A : ptp), exists N, size A N.
Proof.
exact (tp_size ander).
Qed.


Lemma similar_size :
  forall z (A A' : exp z) N,
    similar A A'
    -> size A N
    -> size A' N.
Proof.
intros z A A' N Hsim Hsize.
revert A' Hsim.
induct Hsize; intros; invertc Hsim; intros; subst; eauto using @size.
Qed.


Lemma self_size :
  forall r A B N,
    self r A B
    -> size A N
    -> size B N.
Proof.
intros m A B N Hself Hsize.
eapply similar_size; eauto using self_similar.
Qed.
