
Require Import Tactics.
Require Import Sequence.
Require Import Syntax.
Require Import Cut.
Require Import Self.


Parameter tconstof : tconstant -> kind -> Prop.
Parameter constof : constant -> ntp -> Prop.


Inductive pof : pattern -> ptp -> context -> Prop :=
| pof_shiftm {A}
    : pof shiftm (down A) (nil ; A)

| pof_unitpat
    : pof unitpat one nil

| pof_pairpat {p p' A A' G G'}
    : pof p A G
      -> pof p' A' G'
      -> pof (pairpat p p') (tens A A') (G ;; shctx 0 (length G) G')

| pof_pairpatx {p p' A B G G'}
    : pof p A G
      -> pof p' B G'
      -> pof (pairpat p p') (ex p A B) (G ;; G')
.


Inductive lookup : context -> head -> ntp -> Prop :=
| lookup_const {G c A}
    : constof c A
      -> lookup G (const c) (sh 0 (length G) A)

| lookup_var {G i A}
    : index i G A
      -> lookup G (var i) (sh 0 (S i) A)
.


Inductive hof : context -> head -> ntp -> Prop :=
| hof_id {G h A}
    : lookup G h A
      -> hof G h A

| hof_self {G h m A A'}
    : lookup G h A
      -> expandm A (use h nl) m
      -> self m A A'
      -> hof G h A'
.


Inductive rof : context -> res -> con -> Prop :=
| rof_ret {G v A}
    : vof G v A
      -> rof G (ret v) (trp A)

| rof_use {G h s U A}
    : hof G h A
      -> sof G s A U
      -> rof G (use h s) U

with vof : context -> val -> ptp -> Prop :=
| vof_thunk {G m A}
    : tof G m A
      -> vof G (thunk m) (down A)

| vof_unitp {G}
    : vof G unitp one

| vof_pairp {G v w A B}
    : vof G v A
      -> vof G w B
      -> vof G (pairp v w) (tens A B)

| vof_pairpx {G v w p A B B'}
    : vof G v A
      -> cutp v p 0 B B'
      -> vof G w B'
      -> vof G (pairp v w) (ex p A B)

with sof : context -> spine -> ntp -> con -> Prop :=
| sof_nl {G A}
    : sof G nl A (susp A)

| sof_pm {G p r A B G'}
    : pof p A G'
      -> rof (G ;; G') r (trp (sh 0 (length G') B))
      -> sof G (pm p r) (up A) (trp B)

| sof_app {G v s p A B U B'}
    : vof G v A
      -> cutp v p 0 B B'
      -> sof G s B' U
      -> sof G (app v s) (pi p A B) U

| sof_pi1 {G s A B U}
    : sof G s A U
      -> sof G (pi1 s) (wth A B) U

| sof_pi2 {G s A B U}
    : sof G s B U
      -> sof G (pi2 s) (wth A B) U


with tof : context -> term -> ntp -> Prop :=
| tof_etam {G r h s}
      : rof G r (susp (tuse h s))
      -> tof G (etam r) (tuse h s)

| tof_shiftp {G r A}
    : rof G r (trp A)
      -> tof G (shiftp r) (up A)

| tof_lam {G p m A B G'}
    : pof p A G'
      -> tof (G ;; G') m B
      -> tof G (lam p m) (pi p A B)

| tof_pairm {G m n A B}
    : tof G m A
      -> tof G n B
      -> tof G (pairm m n) (wth A B)

| tof_pairms {G m n A B B'}
    : tof G m A
      -> rsub m 0 B B'
      -> tof G n B'
      -> tof G (pairm m n) (sigma A B)
.


Inductive tsof : context -> tspine -> kind -> kind -> Prop :=
| tsof_nl {G K}
    : tsof G tnl K K

| tsof_app {G v s p A K U K'}
    : vof G v A
      -> cutp v p 0 K K'
      -> tsof G s K' U
      -> tsof G (tapp v s) (tpi p A K) U
.


Inductive ntof : context -> ntp -> Prop :=
| ntof_tuse {G c s K}
    : tconstof c K
      -> tsof G s (sh 0 (length G) K) ty
      -> ntof G (tuse c s)

| ntof_up {G A}
    : ptof G A
      -> ntof G (up A)

| ntof_pi {G p A B G'}
    : ptof G A
      -> pof p A G'
      -> ntof (G;; G') B
      -> ntof G (pi p A B)

| ntof_wth {G A B}
    : ntof G A
      -> ntof G B
      -> ntof G (wth A B)

| ntof_sigma {G A B}
    : ntof G A
      -> ntof (G; A) B
      -> ntof G (sigma A B)

with ptof : context -> ptp -> Prop :=
| ptof_down {G A}
    : ntof G A
      -> ptof G (down A)

| ptof_one {G}
    : ptof G one

| ptof_tens {G A B}
    : ptof G A
      -> ptof G B
      -> ptof G (tens A B)

| ptof_ex {G p A B G'}
    : ptof G A
      -> pof p A G'
      -> ptof (G;; G') B
      -> ptof G (ex p A B)
.


Inductive kof : context -> kind -> Prop :=
| kof_ty {G}
    : kof G ty

| kof_tpi {G p A K G'}
    : ptof G A
      -> pof p A G'
      -> kof (G;; G') K
      -> kof G (tpi p A K)
.


Scheme rof_mut_ind := Minimality for rof Sort Prop
with   vof_mut_ind := Minimality for vof Sort Prop
with   sof_mut_ind := Minimality for sof Sort Prop
with   tof_mut_ind := Minimality for tof Sort Prop.
Combined Scheme static_term_ind from rof_mut_ind, vof_mut_ind, sof_mut_ind, tof_mut_ind.

Scheme ntof_mut_ind := Minimality for ntof Sort Prop
with   ptof_mut_ind := Minimality for ptof Sort Prop.
Combined Scheme static_tp_ind from ntof_mut_ind, ptof_mut_ind.


Inductive ok : context -> con -> Prop :=
| ok_trp {G A}
    : ptof G A
      -> ok G (trp A)

| ok_susp {G A}
    : ntof G A
      -> ok G (susp A)
.


Inductive wf : context -> context -> Prop :=
| wf_nil {G}
    : wf G nil

| wf_cons {G G' A}
    : wf G G'
      -> ntof (G;; G') A
      -> wf G (G'; A)
.


Inductive suspnormal : con -> Prop :=
| suspnormal_trp {A}
    : suspnormal (trp A)

| suspnormal_susp {c s}
    : suspnormal (susp (tuse c s))
.


Inductive gof : forall {z}, context -> exp z -> judgement -> Prop :=
| gof_r {G r U}
    : rof G r U
      -> suspnormal U
      -> gof G r (judr U)

| gof_v {G v A}
    : vof G v A
      -> gof G v (judv A)

| gof_s {G s A U}
    : sof G s A U
      -> suspnormal U
      -> gof G s (juds A U)

| gof_t {G m A}
    : tof G m A
      -> gof G m (judt A)

| gof_ts {G s K K'}
    : tsof G s K K'
      -> gof G s (judts K K')

| gof_nt {G A}
    : ntof G A
      -> gof G A judnt

| gof_pt {G A}
    : ptof G A
      -> gof G A judpt

| gof_k {G A}
    : kof G A
      -> gof G A judk
.


Axiom constof_ntof :
  forall c A,
    constof c A
    -> ntof nil A.


Axiom tconstof_kof :
  forall c K,
    tconstof c K
    -> kof nil K.
