
Require Import Compare_dec.
Require Import Sequence.
Require Import Tactics.
Require Import Syntax.
Require Import Cut.
Require Import Self.
Require Import Static.
Require Import StaticWeak.
Require Import Substitution.
Require Import Expand.



Lemma pof_bound_length :
  forall p A G,
    pof p A G
    -> length G = bound p.
Proof.
intros p A G H.
#2 induct H; try (intros; reflexivity).

(* pairpat *)
intros p p' A A' G G' _ IH1 _ IH2.
simpl.
rewrite -> app_length; [].
rewrite -> shctx_length; [].
omega.

(* pairpatx *)
intros p p' A B G G' _ IH1 _ IH2.
simpl.
rewrite -> app_length; [].
omega.
Qed.


Lemma weaken_pof :
  forall i j p A G,
    pof p A G
    -> pof p (sh i j A) (shctx i j G).
Proof.
intros i j p A G H.
revert i j.
induct H.

(* shiftm *)
intros A i j.
simpl.
repl (i + 0) with i by omega.
apply pof_shiftm; done.

(* unitpat *)
intros i j.
simpl.
apply pof_unitpat; done.

(* pairpat *)
intros p p' A A' G G' _ IH1 _ IH2 i j.
simpl.
rewrite -> (shctx_append _#5 (eq_refl _)); [].
force_exact (pof_pairpat (IH1 i j) (IH2 i j)).
f_equal; [].
f_equal; [].
rewrite -> shctx_length; [].
repl i with (0+i) at 1 by omega.
rewrite <- shctx_commute; [].
f_equal; omega.

(* pairpatx *)
intros p p' A A' G G' Hofp IH1 _ IH2 i j.
simpl.
rewrite -> (shctx_append _#5 (eq_refl _)); [].
force_exact (pof_pairpatx (IH1 i j) (IH2 (bound p + i) j)).
f_equal; [].  
rewrite -> (pof_bound_length _#3 Hofp); [].
repl (bound p + i) with (i + bound p) by omega.
reflexivity.
Qed.


Lemma weaken_lookup :
  forall G1 G2 G3 i j h A,
    length G2 = j
    -> length G3 = i
    -> lookup (G1;; G3) h A
    -> lookup (G1;; G2;; shctx 0 j G3) (sh_head i j h) (sh i j A).
Proof.
intros G1 G2 G3 i j h A HlenG2 HlenG3 Hofh.
#2 invert Hofh; clear Hofh.
  intros c B Hofc <- <-.
  simpl.
  rewrite -> app_length; [].
  rewrite -> sh_sum_middle; try omega; [].
  #2 repl (length G3 + length G1 + j) with (length (G1;; G2;; shctx 0 j G3)).
    rewrite -> !app_length; [].
    rewrite -> shctx_length; [].
    omega.
  apply lookup_const; auto; done.

  intros k B Hindex <- <-.
  unfold sh_head, sh_var.
  forget (lt_dec k i) as b; destruct b as [Hlt | Hnlt].
    repl i with (0 + S k + (i - S k)) by omega.
    rewrite -> sh_commute; [].
    apply lookup_var; [].
    sor (refine (index_app_lt _#5 _ Hindex)) as Hindex'; [omega |].
    so (shctx_index 0 j _#3 Hindex') as Hindex''.
    force_exact (index_app_left _ _ (G1;; G2) _ _ Hindex'').
    f_equal; [].
    f_equal; omega.

    rewrite -> sh_sum_middle; try omega; [].
    repl (S k + j) with (S (j+k)) by omega.
    apply lookup_var; [].
    sor (refine (index_app_ge _#5 _ Hindex)) as Hindex'; [omega |].
    so (index_app_right _ G2 _#3 Hindex') as Hindex''.
    so (index_app_right _ (shctx 0 j G3) _#3 Hindex'') as Hindex'''.
    rewrite -> shctx_length in Hindex'''; [].
    force_exact Hindex'''.
    f_equal; omega.
Qed.


Lemma sh_self :
  forall i j r A B,
    self r A B
    -> self (sh i j r) (sh i j A) (sh i j B).
Proof.
intros i j r A B H.
revert i.
induct H; eauto using self; [].
(* sigma *)
intros m n A B C B' D _ IH1 HsubB _ IH2 i.
simpl.
eapply self_sigma; eauto; [].
eapply rsub_sh_commute_left; eauto; done.
Qed.


Lemma weaken_hof :
  forall G1 G2 G3 i j h A,
    length G2 = j
    -> length G3 = i
    -> hof (G1;; G3) h A
    -> hof (G1;; G2;; shctx 0 j G3) (sh_head i j h) (sh i j A).
Proof.
intros G1 G2 G3 i j h A HlenG2 HlenG3 Hofh.
#2 invert Hofh.
  intro Hlookup.
  eauto using hof_id, weaken_lookup; done.

  intros m A' Hlookup Hexpand Hself.
  eapply hof_self; eauto using weaken_lookup, sh_self; [].
  so (sh_expandm_sh i j _#3 Hexpand) as Hexpand'.
  simpl in Hexpand'.
  exact Hexpand'.
Qed.


Lemma weakening :
  (forall G1 G2 G3 i j r U,
     length G2 = j
     -> length G3 = i
     -> rof (G1;; G3) r U
     -> rof (G1;; G2;; shctx 0 j G3) (sh i j r) (sh i j U))
  /\
  (forall G1 G2 G3 i j v A,
     length G2 = j
     -> length G3 = i
     -> vof (G1;; G3) v A
     -> vof (G1;; G2;; shctx 0 j G3) (sh i j v) (sh i j A))
  /\
  (forall G1 G2 G3 i j s A U,
     length G2 = j
     -> length G3 = i
     -> sof (G1;; G3) s A U
     -> sof (G1;; G2;; shctx 0 j G3) (sh i j s) (sh i j A) (sh i j U))
  /\
  (forall G1 G2 G3 i j m A,
     length G2 = j
     -> length G3 = i
     -> tof (G1;; G3) m A
     -> tof (G1;; G2;; shctx 0 j G3) (sh i j m) (sh i j A)).
Proof.
#8 exploit
  (static_term_ind
     (fun G r U =>
        forall G1 G2 G3 i j,
          G = G1;; G3
          -> length G2 = j
          -> length G3 = i
          -> rof (G1;; G2;; shctx 0 j G3) (sh i j r) (sh i j U))
     (fun G v A =>
        forall G1 G2 G3 i j,
          G = G1;; G3
          -> length G2 = j
          -> length G3 = i
          -> vof (G1;; G2;; shctx 0 j G3) (sh i j v) (sh i j A))
     (fun G s A U =>
        forall G1 G2 G3 i j,
          G = G1;; G3
          -> length G2 = j
          -> length G3 = i
          -> sof (G1;; G2;; shctx 0 j G3) (sh i j s) (sh i j A) (sh i j U))
     (fun G m A =>
        forall G1 G2 G3 i j,
          G = G1;; G3
          -> length G2 = j
          -> length G3 = i
          -> tof (G1;; G2;; shctx 0 j G3) (sh i j m) (sh i j A)))
  as Hind;
try (intros; subst; simpl; eauto using rof, vof, sof, tof; done).

(* use *)
intros G h s U A Hofh _ IH G1 G2 G3 i j -> HlenG2 HlenG3.
eapply rof_use; eauto; [].
apply weaken_hof; auto; done.

(* pairpx *)
intros G v w p A B B' _ IH1 Hcutp _ IH2 G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
eapply vof_pairpx; eauto; [].
repl (bound p + i) with (0 + bound p + i) by omega.
repl i with (0 + i) at 3 by omega.
apply cutp_sh_commute_left; eauto; done.

(* pm *)
intros ? p r A U G4 Hofp _ IH G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
eapply sof_pm; eauto using weaken_pof; [].
repl (G1;; G2;; shctx 0 j G3;; shctx i j G4) with (G1;; G2;; shctx 0 j (G3;; G4)).
  erewrite -> shctx_append; eauto; [].
  autorewrite with canonlist; reflexivity.
rewrite -> shctx_length; [].
so (pof_bound_length _#3 Hofp) as HlenG4.
repl i with (0 + i) at 2 by omega.
rewrite <- sh_commute; [].
repl (0 + length G4 + i) with (bound p + i) by omega.
#2 apply IH; auto.
  rewrite -> app_assoc; reflexivity.

  rewrite -> app_length; omega.

(* app *)
intros ? v s p A B U B' _ IH1 Hcutp _ IH2 G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
eapply sof_app; eauto using weaken_pof; [].
repl (bound p + i) with (0 + bound p + i) by omega.
repl i with (0 + i) at 3 by omega.
eapply cutp_sh_commute_left; eauto; done.

(* etam *)
intros ? r h s _ IH G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
apply tof_etam; [].
exact (IH _#5 (eq_refl _) HlenG2 HlenG3).

(* lam *)
intros ? p m A B G4 Hofp _ IH G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
eapply tof_lam; eauto using weaken_pof; [].
repl (G1;; G2;; shctx 0 j G3;; shctx i j G4) with (G1;; G2;; shctx 0 j (G3;; G4)).
  erewrite -> shctx_append; eauto; [].
  autorewrite with canonlist; reflexivity.
#2 apply IH; auto.
  rewrite -> app_assoc; reflexivity.

  so (pof_bound_length _#3 Hofp).
  rewrite -> app_length; omega.

(* pairms *)
intros ? m n A B B' _ IH1 HsubB _ IH2 G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
eapply tof_pairms; eauto; [].
eapply rsub_sh_commute_left; eauto; done.

(* wrapup *)
destruct_all Hind.
do2 3 split; intros; eauto; done.
Qed.
 

Lemma weaken_rof :
  forall G1 G2 G3 i j r U,
    length G2 = j
    -> length G3 = i
    -> rof (G1;; G3) r U
    -> rof (G1;; G2;; shctx 0 j G3) (sh i j r) (sh i j U).
Proof.
exact (weakening andel).
Qed.


Lemma weakenn_rof :
  forall G G' i r U,
    rof G r U
    -> length G' = i
    -> rof (G;; G') (sh 0 i r) (sh 0 i U).
Proof.
intros G G' i r U Hofr <-.
repl (G;; G') with (G;; G';; shctx 0 (length G') nil) by reflexivity.
apply weaken_rof; auto.
Qed.


Lemma weaken_vof :
  forall G1 G2 G3 i j v A,
    length G2 = j
    -> length G3 = i
    -> vof (G1;; G3) v A
    -> vof (G1;; G2;; shctx 0 j G3) (sh i j v) (sh i j A).
Proof.
exact (weakening anderl).
Qed.


Lemma weakenn_vof :
  forall G G' i v A,
    vof G v A
    -> length G' = i
    -> vof (G;; G') (sh 0 i v) (sh 0 i A).
Proof.
intros G G' i v A Hof <-.
repl (G;; G') with (G;; G';; shctx 0 (length G') nil) by reflexivity.
apply weaken_vof; auto.
Qed.


Lemma weaken_sof :
  forall G1 G2 G3 i j s A U,
    length G2 = j
    -> length G3 = i
    -> sof (G1;; G3) s A U
    -> sof (G1;; G2;; shctx 0 j G3) (sh i j s) (sh i j A) (sh i j U).
Proof.
exact (weakening anderrl).
Qed.


Lemma weaken_tof :
  forall G1 G2 G3 i j m A,
    length G2 = j
    -> length G3 = i
    -> tof (G1;; G3) m A
    -> tof (G1;; G2;; shctx 0 j G3) (sh i j m) (sh i j A).
Proof.
exact (weakening anderrr).
Qed.


Lemma weakenn_tof :
  forall G G' i m A,
    tof G m A
    -> length G' = i
    -> tof (G;; G') (sh 0 i m) (sh 0 i A).
Proof.
intros G G' i m A Hof <-.
repl (G;; G') with (G;; G';; shctx 0 (length G') nil) by reflexivity.
apply weaken_tof; auto.
Qed.


Lemma weaken_tsof :
  forall G1 G2 G3 i j s K L,
    length G2 = j
    -> length G3 = i
    -> tsof (G1;; G3) s K L
    -> tsof (G1;; G2;; shctx 0 j G3) (sh i j s) (sh i j K) (sh i j L).
Proof.
intros G1 G2 G3 i j s K L HlenG2 HlenG3 H.
remember (G1;; G3) as G eqn:HeqG.
revert G1 G2 G3 i j HeqG HlenG2 HlenG3.
induct H.

(* nl *)
intros G K G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
apply tsof_nl; done.

(* app *)
intros G v s p A K U K' Hofv Hcutp Hofs IH G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
eapply tsof_app; eauto using weaken_pof, weaken_vof; [].
repl (bound p + i) with (0 + bound p + i) by omega.
repl i with (0 + i) at 3 by omega.
eapply cutp_sh_commute_left; eauto; done.
Qed.


Lemma weaken_tpof :
  (forall G1 G2 G3 i j A,
     length G2 = j
     -> length G3 = i
     -> ntof (G1;; G3) A
     -> ntof (G1;; G2;; shctx 0 j G3) (sh i j A))
  /\
  (forall G1 G2 G3 i j A,
     length G2 = j
     -> length G3 = i
     -> ptof (G1;; G3) A
     -> ptof (G1;; G2;; shctx 0 j G3) (sh i j A)).
Proof.
#5 exploit
  (static_tp_ind
     (fun G A =>
        forall G1 G2 G3 i j,
          G = G1;; G3
          -> length G2 = j
          -> length G3 = i
          -> ntof (G1;; G2;; shctx 0 j G3) (sh i j A))
     (fun G A =>
        forall G1 G2 G3 i j,
          G = G1;; G3
          -> length G2 = j
          -> length G3 = i
          -> ptof (G1;; G2;; shctx 0 j G3) (sh i j A)))
  as Hind;
try (intros; subst; simpl; eauto using ntof, ptof; done).

(* use *)
intros G c s K Hofc Hofs G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
eapply ntof_tuse; eauto; [].
force_exact (weaken_tsof _#8 HlenG2 HlenG3 Hofs).
f_equal; [].
rewrite -> !app_length; [].
rewrite -> shctx_length; [].
rewrite -> sh_sum_middle; try omega; [].
f_equal; omega.

(* pi *)
intros ? p A B G4 _ IH1 Hofp _ IH2 G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
eapply ntof_pi; eauto using weaken_pof; [].
repl (G1;; G2;; shctx 0 j G3;; shctx i j G4) with (G1;; G2;; shctx 0 j (G3;; G4)).
  erewrite -> shctx_append; eauto; [].
  autorewrite with canonlist; reflexivity.
#2 apply IH2; auto.
  rewrite -> app_assoc; reflexivity.

  so (pof_bound_length _#3 Hofp).
  rewrite -> app_length; omega.

(* sigma *)
intros ? A B _ IH1 _ IH2 G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
eapply ntof_sigma; eauto; [].
unfold ntp.
repl (G1;; G2;; shctx 0 j G3; @sh nt i j A) with (G1;; G2;; shctx 0 j (G3; A)).
  simpl.
  rewrite -> HlenG3; auto; done.
apply IH2; eauto; [].
simpl; omega.

(* ex *)
intros ? p A B G4 _ IH1 Hofp _ IH2 G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
eapply ptof_ex; eauto using weaken_pof; [].
repl (G1;; G2;; shctx 0 j G3;; shctx i j G4) with (G1;; G2;; shctx 0 j (G3;; G4)).
  erewrite -> shctx_append; eauto; [].
  autorewrite with canonlist; reflexivity.
#2 apply IH2; auto.
  rewrite -> app_assoc; reflexivity.

  so (pof_bound_length _#3 Hofp).
  rewrite -> app_length; omega.

(* wrapup *)
destruct Hind.
split; intros; eauto; done.
Qed.


Lemma weaken_ntof :
  forall G1 G2 G3 i j A,
    length G2 = j
    -> length G3 = i
    -> ntof (G1;; G3) A
    -> ntof (G1;; G2;; shctx 0 j G3) (sh i j A).
Proof.
exact (weaken_tpof andel).
Qed.


Lemma weaken_ptof :
  forall G1 G2 G3 i j A,
    length G2 = j
    -> length G3 = i
    -> ptof (G1;; G3) A
    -> ptof (G1;; G2;; shctx 0 j G3) (sh i j A).
Proof.
exact (weaken_tpof ander).
Qed.


Lemma weakenn_ntof :
  forall G G' i B,
    ntof G B
    -> length G' = i
    -> ntof (G;; G') (sh 0 i B).
Proof.
intros G G' i B Hof <-.
repl (G;; G') with (G;; G';; shctx 0 (length G') nil) by reflexivity.
apply weaken_ntof; auto.
Qed.


Lemma weaken1_ntof :
  forall G A B,
    ntof G B
    -> ntof (G; A) (sh 0 1 B).
Proof.
intros G A B H.
repl (G; A) with (G;; (nil; A)) by reflexivity.
apply weakenn_ntof; auto.
Qed.


Lemma weaken_wf :
  forall G1 G2 G3 i j G,
    length G2 = j
    -> length G3 = i
    -> wf (G1;; G3) G
    -> wf (G1;; G2;; shctx 0 j G3) (shctx i j G).
Proof.
intros G1 G2 G3 i j G HlenG2 HlenG3 Hwf.
remember (G1;; G3) as GG eqn:HeqG.
revert G1 G2 G3 i j HeqG HlenG2 HlenG3.
induct Hwf.

(* nil *)
intros; simpl; apply wf_nil; done.

(* cons *)
intros GG G A _ IH HofA G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
apply wf_cons; auto; [].
repl (G1;; G2;; shctx 0 j G3;; shctx i j G) with (G1;; G2;; shctx 0 j (G3;; G)).
  erewrite -> shctx_append; eauto; [].
  autorewrite with canonlist.
  reflexivity.
#2 eapply weaken_ntof; eauto.
  rewrite -> app_length; omega.

  autorewrite with canonlist; assumption.
Qed.


Lemma weakenn_wf :
  forall G1 G2 i G,
    length G2 = i
    -> wf G1 G
    -> wf (G1;; G2) (shctx 0 i G).
Proof.
intros G1 G2 i G <- Hwf.
repl (G1;; G2) with (G1;; G2;; shctx 0 (length G2) nil) by reflexivity.
apply weaken_wf; auto.
Qed.


(* Weakening for weak typing *)

Lemma wpof_bound_length :
  forall p A G,
    wpof p A G
    -> length G = bound p.
Proof.
intros p A G H.
#2 induct H; try (intros; reflexivity).

(* pairpat *)
intros p p' A A' G G' _ IH1 _ IH2.
simpl.
rewrite -> app_length; [].
rewrite -> shctx_length; [].
omega.

(* pairpatx *)
intros p p' p'' A B G G' _ IH1 _ IH2.
simpl.
rewrite -> app_length; [].
omega.
Qed.


Lemma weakening_weak :
  (forall G1 G2 G3 i j r U,
     length G2 = j
     -> length G3 = i
     -> wrof (G1;; G3) r U
     -> wrof (G1;; G2;; G3) (sh i j r) U)
  /\
  (forall G1 G2 G3 i j v A,
     length G2 = j
     -> length G3 = i
     -> wvof (G1;; G3) v A
     -> wvof (G1;; G2;; G3) (sh i j v) A)
  /\
  (forall G1 G2 G3 i j s A U,
     length G2 = j
     -> length G3 = i
     -> wsof (G1;; G3) s A U
     -> wsof (G1;; G2;; G3) (sh i j s) A U)
  /\
  (forall G1 G2 G3 i j m A,
     length G2 = j
     -> length G3 = i
     -> wtof (G1;; G3) m A
     -> wtof (G1;; G2;; G3) (sh i j m) A).
Proof.
#4 exploit
  (weak_static_term_ind
     (fun G r U =>
        forall G1 G2 G3 i j,
          G = G1;; G3
          -> length G2 = j
          -> length G3 = i
          -> wrof (G1;; G2;; G3) (sh i j r) U)
     (fun G v A =>
        forall G1 G2 G3 i j,
          G = G1;; G3
          -> length G2 = j
          -> length G3 = i
          -> wvof (G1;; G2;; G3) (sh i j v) A)
     (fun G s A U =>
        forall G1 G2 G3 i j,
          G = G1;; G3
          -> length G2 = j
          -> length G3 = i
          -> wsof (G1;; G2;; G3) (sh i j s) A U)
     (fun G m A =>
        forall G1 G2 G3 i j,
          G = G1;; G3
          -> length G2 = j
          -> length G3 = i
          -> wtof (G1;; G2;; G3) (sh i j m) A))
  as Hind;
try (intros; subst; simpl; eauto using wrof, wvof, wsof, wtof; done).

Require Import Similar.

(* use *)
intros G h s A A' U Hlookup Hsim _ IH G1 G2 G3 i j -> HlenG2 HlenG3.
so (weaken_lookup _#7 HlenG2 HlenG3 Hlookup) as Hlookup'.
#2 sor (refine (lookup_similarctx _ (G1;; G2;; G3) _ _ Hlookup' _)) as H.
  apply similarctx_append; eauto using shctx_similarctx with similar; done.
destruct H as (A'' & Hsim' & Hlookup'').
eapply wrof_use; eauto; [].
eauto using sh_similarm with similar; done.

(* pm *)
intros ? p r A U G4 Hofp _ IH G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
eapply wsof_pm; eauto; [].
repl (G1;; G2;; G3;; G4) with (G1;; G2;; (G3;; G4)) by (autorewrite with canonlist; reflexivity).
#2 apply IH; auto.
  rewrite -> app_assoc; reflexivity.

  so (wpof_bound_length _#3 Hofp).
  rewrite -> app_length; omega.

(* lam *)
intros ? p m p' A B G4 Hofp _ IH G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
eapply wtof_lam; eauto; [].
repl (G1;; G2;; G3;; G4) with (G1;; G2;; (G3;; G4)) by (autorewrite with canonlist; reflexivity).
#2 apply IH; auto.
  rewrite -> app_assoc; reflexivity.

  so (wpof_bound_length _#3 Hofp).
  rewrite -> app_length; omega.

(* wrapup *)
destruct_all Hind.
do2 3 split; intros; eauto; done.
Qed.


Lemma weaken_wrof :
  forall G1 G2 G3 i j r U,
    length G2 = j
    -> length G3 = i
    -> wrof (G1;; G3) r U
    -> wrof (G1;; G2;; G3) (sh i j r) U.
Proof.
exact (weakening_weak andel).
Qed.


Lemma weaken_wvof :
  forall G1 G2 G3 i j v A,
    length G2 = j
    -> length G3 = i
    -> wvof (G1;; G3) v A
    -> wvof (G1;; G2;; G3) (sh i j v) A.
Proof.
exact (weakening_weak anderl).
Qed.


Lemma weaken_wtof :
  forall G1 G2 G3 i j m A,
    length G2 = j
    -> length G3 = i
    -> wtof (G1;; G3) m A
    -> wtof (G1;; G2;; G3) (sh i j m) A.
Proof.
exact (weakening_weak anderrr).
Qed.


Lemma weaken_wtsof :
  forall G1 G2 G3 i j s K L,
    length G2 = j
    -> length G3 = i
    -> wtsof (G1;; G3) s K L
    -> wtsof (G1;; G2;; G3) (sh i j s) K L.
Proof.
intros G1 G2 G3 i j s K L HlenG2 HlenG3 H.
remember (G1;; G3) as G eqn:HeqG.
revert G1 G2 G3 i j HeqG HlenG2 HlenG3.
induct H.

(* nl *)
intros G K G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
apply wtsof_nl; done.

(* app *)
intros G v s p A K U Hofv Hofs IH G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
eapply wtsof_app; eauto using weaken_wvof; done.
Qed.


Lemma weaken_wtpof :
  (forall G1 G2 G3 i j A,
     length G2 = j
     -> length G3 = i
     -> wntof (G1;; G3) A
     -> wntof (G1;; G2;; G3) (sh i j A))
  /\
  (forall G1 G2 G3 i j A,
     length G2 = j
     -> length G3 = i
     -> wptof (G1;; G3) A
     -> wptof (G1;; G2;; G3) (sh i j A)).
Proof.
#5 exploit
  (weak_static_tp_ind
     (fun G A =>
        forall G1 G2 G3 i j,
          G = G1;; G3
          -> length G2 = j
          -> length G3 = i
          -> wntof (G1;; G2;; G3) (sh i j A))
     (fun G A =>
        forall G1 G2 G3 i j,
          G = G1;; G3
          -> length G2 = j
          -> length G3 = i
          -> wptof (G1;; G2;; G3) (sh i j A)))
  as Hind;
try (intros; subst; simpl; eauto using wntof, wptof; done).

(* use *)
intros G c s K K' Hofc Hsim Hofs G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
eapply wntof_tuse; eauto; [].
exact (weaken_wtsof _#8 HlenG2 HlenG3 Hofs).

(* pi *)
intros ? p A B G4 _ IH1 Hofp _ IH2 G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
so (wpof_similar _ _ (sh i j A) _ Hofp (sh_similarp _#3)) as (G4' & HsimG4 & Hofp').
eapply wntof_pi; eauto; [].
#3 exploit (IH2 G1 G2 (G3;; G4) (bound p + i) j) as HofB; auto.
  rewrite -> app_assoc; reflexivity.
  
  so (wpof_bound_length _#3 Hofp).
  rewrite -> app_length; omega.
rewrite <- app_assoc in HofB; [].
eapply wntof_similarctx; eauto with similar; done.

(* sigma *)
intros ? A B _ IH1 _ IH2 G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
eapply wntof_sigma; eauto; [].
#2 exploit (IH2 G1 G2 (G3; A) (1 + i) j) as HofB; eauto.
  simpl.
  omega.
simpl in HofB.
eapply wntof_similarctx; eauto using sh_similarm with similar; done.

(* ex *)
intros ? p A B G4 _ IH1 Hofp _ IH2 G1 G2 G3 i j -> HlenG2 HlenG3.
simpl.
so (wpof_similar _ _ (sh i j A) _ Hofp (sh_similarp _#3)) as (G4' & HsimG4 & Hofp').
eapply wptof_ex; eauto; [].
#3 exploit (IH2 G1 G2 (G3;; G4) (bound p + i) j) as HofB; auto.
  rewrite -> app_assoc; reflexivity.
  
  so (wpof_bound_length _#3 Hofp).
  rewrite -> app_length; omega.
rewrite <- app_assoc in HofB; [].
eapply wptof_similarctx; eauto with similar; done.

(* wrapup *)
destruct Hind.
split; intros; eauto; done.
Qed.


Lemma weaken_wntof :
  forall G1 G2 G3 i j A,
    length G2 = j
    -> length G3 = i
    -> wntof (G1;; G3) A
    -> wntof (G1;; G2;; G3) (sh i j A).
Proof.
exact (weaken_wtpof andel).
Qed.


Lemma weakenn_wntof :
  forall G G' i A,
    wntof G A
    -> length G' = i
    -> wntof (G;; G') (sh 0 i A).
Proof.
intros G G' i A Hof Hlen.
repl (G;; G') with (G;; G';; nil) by reflexivity.
apply weaken_wntof; auto.
Qed.


Lemma weaken1_wntof :
  forall G A B,
    wntof G B
    -> wntof (G; A) (sh 0 1 B).
Proof.
intros G A B H.
repl (G; A) with (G;; (nil; A)) by reflexivity.
apply weakenn_wntof; auto.
Qed.


Lemma weaken_wptof :
  forall G1 G2 G3 i j A,
    length G2 = j
    -> length G3 = i
    -> wptof (G1;; G3) A
    -> wptof (G1;; G2;; G3) (sh i j A).
Proof.
exact (weaken_wtpof ander).
Qed.


Lemma weaken_wwf :
  forall G1 G2 G3 i j G,
    length G2 = j
    -> length G3 = i
    -> wwf (G1;; G3) G
    -> wwf (G1;; G2;; G3) (shctx i j G).
Proof.
intros G1 G2 G3 i j G HlenG2 HlenG3 Hwf.
revert G1 G2 G3 i j HlenG2 HlenG3 Hwf.
induct G.

(* nil *)
intros G1 G2 G3 i j HlenG2 HlenG3 Hwf.
simpl.
exact wwf_nil.

(* cons *)
intros A G IH G1 G2 G3 i j HlenG2 HlenG3 Hwf.
invertc Hwf; [].
intros HwfG HofA.
simpl.
apply wwf_cons; eauto; [].
repl (G1;; G2;; G3;; shctx i j G) with (G1;; G2;; (G3;; shctx i j G)) by (autorewrite with canonlist; reflexivity).
#2 eapply weaken_wntof; eauto.
  rewrite -> app_length; [].
  rewrite -> shctx_length; [].
  omega.

  autorewrite with canonlist.
  eapply wntof_similarctx; eauto using shctx_similarctx with similar; done.
Qed.
