
Require Import Arith.
Require Import Compare.
Require Import Tactics.
Require Import Sequence.
Require Import Syntax.
Require Import Cut.
Require Import Self.
Require Import Static.
Require Import Substitution.
Require Import Compose.
Require Import Size.
Require Import Similar.



Lemma expand_similar :
  (forall A A' r m,
     similar A A'
     -> expandm A r m
     -> expandm A' r m)
  /\
  (forall A A' p v,
     similar A A'
     -> expandp A p v
     -> expandp A' p v).
Proof.
exploit
  (expand_mut_ind
     (fun A r m => forall A' (Hsim : similar A A'), expandm A' r m)
     (fun A p v => forall A' (Hsim : similar A A'), expandp A' p v))
   as Hind;
try (intros;
     invertc Hsim;
     intros;
     subst;
     eauto using expandm, expandp;
     done); [].
destruct Hind.
split; intros; eauto.
Qed.


Lemma expandm_similar :
  forall A A' r m,
    similar A A'
    -> expandm A r m
    -> expandm A' r m.
Proof.
exact (expand_similar andel).
Qed.


Lemma expandp_similar :
  forall A A' p v,
    similar A A'
    -> expandp A p v
    -> expandp A' p v.
Proof.
exact (expand_similar ander).
Qed.


Lemma expand_fun :
  (forall A r m m',
     expandm A r m
     -> expandm A r m'
     -> m = m')
  /\
  (forall A p p' v v',
     expandp A p v
     -> expandp A p' v'     
     -> p = p' /\ v = v').
Proof.
exploit
  (expand_mut_ind
     (fun A r m =>
        forall m' (Hexpand : expandm A r m'),
          m = m')
     (fun A p v =>
        forall p' v' (Hexpand : expandp A p' v'),
          p = p' /\ v = v'))
  as Hind.

(* atom *)
intros c s r m' Hexpand.
invert Hexpand; intros; subst; eauto; done.

(* up *)
intros A r p v r' _ IH Hcompose m' Hexpand.
invertc Hexpand; [].
intros p' v' r'' HexpandA Hcompose' <-.
destruct (IH _ _ HexpandA); subst p' v'.
so (compose_fun _#4 Hcompose Hcompose'); subst r''.
reflexivity.

(* pi *)
intros p' A B r p v r' m _ IH1 Hcompose _ IH2 mm Hexpand.
invertc Hexpand; [].
intros p'' v' r'' m' HexpandA Hcompose' HexpandB <-.
destruct (IH1 _ _ HexpandA); subst p'' v'.
so (compose_fun _#4 Hcompose Hcompose'); subst r''.
f_equal; eauto; done.

(* wth *)
intros A B r r1 r2 m1 m2 Hcompose1 Hcompose2 _ IH1 _ IH2 mm Hexpand.
invertc Hexpand; [].
intros r1' r2' m1' m2' Hcompose1' Hcompose2' HexpandA HexpandB <-.
so (compose_fun _#4 Hcompose1 Hcompose1'); subst r1'.
so (compose_fun _#4 Hcompose2 Hcompose2'); subst r2'.
f_equal; eauto; done.

(* sigma *)
intros A B r r1 r2 m1 m2 Hcompose1 Hcompose2 _ IH1 _ IH2 mm Hexpand.
invertc Hexpand; [].
intros r1' r2' m1' m2' Hcompose1' Hcompose2' HexpandA HexpandB <-.
so (compose_fun _#4 Hcompose1 Hcompose1'); subst r1'.
so (compose_fun _#4 Hcompose2 Hcompose2'); subst r2'.
f_equal; eauto; done.

(* down *)
intros A m _ IH p' v' Hexpand.
invertc Hexpand; [].
intros m' HexpandA <- <-.
split; f_equal; eauto; done.

(* one *)
intros p' v' Hexpand.
invertc Hexpand; [].
intros <- <-.
eauto; done.

(* tens *)
intros A B p q v w _ IH1 _ IH2 pp vv Hexpand.
invertc Hexpand; [].
intros p' q' v' w' HexpandA HexpandB <- <-.
destruct (IH1 _ _ HexpandA).
destruct (IH2 _ _ HexpandB).
subst; split; f_equal; eauto; done.

(* ex *)
intros p'' A B p q v w _ IH1 _ IH2 pp vv Hexpand.
invertc Hexpand; [].
intros p' q' v' w' HexpandA HexpandB <- <-.
destruct (IH1 _ _ HexpandA).
destruct (IH2 _ _ HexpandB).
subst; split; f_equal; eauto; done.

(* wrapup *)
destruct Hind; split; intros; eauto; done.
Qed.


Lemma expandm_fun :
  forall A r m m',
    expandm A r m
    -> expandm A r m'
    -> m = m'.
Proof.
exact (expand_fun andel).
Qed.


Lemma expandm_similar_fun :
  forall A A' r m m',
    similar A A'
    -> expandm A r m
    -> expandm A' r m'
    -> m = m'.
Proof.
intros A A' r m m' Hsim Hexpand Hexpand'.
exact (expandm_fun _#4 (expandm_similar _#4 Hsim Hexpand) Hexpand').
Qed.


Lemma sh_expand :
  (forall i j A r m,
     expandm A r m
     -> expandm A (sh i j r) (sh i j m))
  /\
  (forall i j A p v,
     expandp A p v
     -> expandp A p (sh (i + bound p) j v)).
Proof.
exploit
  (expand_mut_ind
     (fun A r m =>
        forall i j,
          expandm A (sh i j r) (sh i j m))
     (fun A p v =>
        forall i j,
          expandp A p (sh (i + bound p) j v)))
  as Hind.

(* atom *)
intros c s r i j.
simpl.
apply expandm_atom; done.

(* up *)
intros A r p v s _ IH Hcompose i j.
simpl.
so (IH i j) as Hexpand.
eapply expandm_up; eauto; [].
so (sh_compose i j _#3 Hcompose) as Hcompose'.
simpl in Hcompose'.
repl (i + bound p) with (bound p + i) by omega.
exact Hcompose'.

(* pi *)
intros p' A B r p v s m HexpandA IH1 Hcompose HexpandB IH2 i j.
simpl.
so (IH1 i j) as HexpandA'.
so (IH2 (i + bound p) j) as HexpandB'.
so (sh_compose (i + bound p) j _#3 Hcompose) as Hcompose'.
#3 eapply expandm_pi.
  exact HexpandA'.

  simpl in Hcompose'.
  repl (i + bound p) with (0 + bound p + i) in Hcompose' at 1 by omega.
  rewrite -> sh_commute in Hcompose'; [].
  repl (0 + i) with i in Hcompose' by omega.
  exact Hcompose'.

  simpl in HexpandB'.
  repl (bound p + i) with (i + bound p) by omega.
  exact HexpandB'.

(* wth *)
intros A B r r1 r2 m1 m2 Hcompose1 Hcompose2 _ IH1 _ IH2 i j.
simpl.
so (sh_compose i j _#3 Hcompose1) as Hcompose1'.
so (sh_compose i j _#3 Hcompose2) as Hcompose2'.
so (IH1 i j) as HexpandA'.
so (IH2 i j) as HexpandB'.
simpl in Hcompose1'.
simpl in Hcompose2'.
simpl in HexpandA'.
simpl in HexpandB'.
eapply expandm_wth; eauto; done.

(* sigma *)
intros A B r r1 r2 m1 m2 Hcompose1 Hcompose2 _ IH1 _ IH2 i j.
simpl.
so (sh_compose i j _#3 Hcompose1) as Hcompose1'.
so (sh_compose i j _#3 Hcompose2) as Hcompose2'.
so (IH1 i j) as HexpandA'.
so (IH2 i j) as HexpandB'.
simpl in Hcompose1'.
simpl in Hcompose2'.
eapply expandm_sigma; eauto; done.

(* down *)
intros A m _ IH i j.
simpl.
apply expandp_down; [].
so (IH (i+1) j) as Hexpand.
simpl in Hexpand.
force_exact Hexpand; [].
f_equal; [].
f_equal; f_equal; [].
unfold sh_var.
forget (lt_dec 0 (i + 1)) as b; destruct b; omega.

(* one *)
intros i j.
simpl.
apply expandp_one; done.

(* tens *)
intros A B p q v w _ IH1 _ IH2 i j.
simpl.
repl (i + (bound p + bound q)) with (0 + bound q + (i + bound p)) at 1 by omega.
rewrite -> sh_commute; [].
#2 apply expandp_tens.
  exact (IH1 i j).

  force_exact (IH2 (bound p + i) j).
  f_equal; [].
  f_equal; [].
  omega.

(* ex *)
intros p' A B p q v w _ IH1 _ IH2 i j.
simpl.
repl (i + (bound p + bound q)) with (0 + bound q + (i + bound p)) at 1 by omega.
rewrite -> sh_commute; [].
#2 apply expandp_ex.
  exact (IH1 i j).

  force_exact (IH2 (bound p + i) j).
  f_equal; [].
  f_equal; [].
  omega.

(* wrapup *)
destruct Hind; split; intros; eauto; done.  
Qed.


Lemma sh_expandm :
  forall i j A r m,
    expandm A r m
    -> expandm A (sh i j r) (sh i j m).
Proof.
exact (sh_expand andel).
Qed.


Lemma sh_expandp :
  forall i j A p v,
    expandp A p v
    -> expandp A p (sh (i + bound p) j v).
Proof.
exact (sh_expand ander).
Qed.


Lemma sh_expandm_sh :
  forall i j A r m,
    expandm A r m
    -> expandm (sh i j A) (sh i j r) (sh i j m).
Proof.
intros i j A r m H.
eapply expandm_similar; eauto using sh_similarm; [].
apply sh_expandm; auto.
Qed.


Lemma sh_expandp_sh :
  forall i j A p v,
    expandp A p v
    -> expandp (sh i j A) p (sh (i + bound p) j v).
Proof.
intros i j A p v H.
eapply expandp_similar; eauto using sh_similarp; [].
apply sh_expandp; auto.
Qed.


Lemma expandp_closed :
  forall i j A p v,
    expandp A p v
    -> sh (bound p + i) j v = v.
Proof.
intros i j A p v Hexpand.
revert i j.
induct Hexpand.

(* down *)
intros A m Hexpand i j.
simpl.
f_equal; [].
so (sh_expandm (S i) j _#3 Hexpand) as Hexpand'.
simpl in Hexpand'.
repl (sh_var (S i) j 0) with 0 in Hexpand'.
  unfold sh_var.
  forget (lt_dec 0 (S i)) as b; destruct b; omega.
so (expandm_fun _#4 Hexpand Hexpand').
auto; done.

(* one *)
intros i j.
reflexivity.

(* tens *)
intros A B p q v w _ IH1 _ IH2 i j.
simpl.
#2 f_equal.
  repl (bound p + bound q + i) with (0 + bound q + (bound p + i)) by omega.
  rewrite -> sh_commute; [].
  f_equal; [].
  apply IH1; done.

  repl (bound p + bound q + i) with (bound q + (bound p + i)) by omega.
  apply IH2; done.

(* ex *)
intros p' A B p q v w _ IH1 _ IH2 i j.
simpl.
#2 f_equal.
  repl (bound p + bound q + i) with (0 + bound q + (bound p + i)) by omega.
  rewrite -> sh_commute; [].
  f_equal; [].
  apply IH1; done.

  repl (bound p + bound q + i) with (bound q + (bound p + i)) by omega.
  apply IH2; done.
Qed.


Lemma expandp_closed_rsub :
  forall f i A p v,
    expandp A p v
    -> rsub f (bound p + i) v v.
Proof.
intros f i A p v Hexpand.
exploit (rsub_sh_cancel f (bound p + i) (bound p + i) 1 v) as Hsub; try omega; [].
repl (1 - 1) with 0 in Hsub by omega.
rewrite -> sh_zero in Hsub; [].
rewrite -> (expandp_closed _#5 Hexpand) in Hsub; auto.
Qed.


Lemma expandp_closed_cutp :
  forall w q i A p v,
    patmatch w q
    -> expandp A p v
    -> cutp w q (bound p + i) v v.
Proof.
intros w q i A p v Hmatch Hexpand.
revert i.
induct Hmatch.

(* down *)
intros m i.
apply cutp_down; eauto using expandp_closed_rsub; done.

(* one *)
intro i.
apply cutp_one; done.

(* tens *)
intros w1 w2 q1 q2 _ IH1 _ IH2 i.
eapply cutp_tens; eauto; [].
repl (bound q2 + (bound p + i)) with (bound p + (bound q2 + i)) by omega.
apply IH1; done.
Qed.


Lemma expandp_patmatch :
  forall A p v,
    expandp A p v
    -> patmatch v p.
Proof.
intros A p v H.
induct H; intros; eauto using patmatch, sh_patmatch.
Qed.


Lemma inst_into_expandm :
  forall A r m f i r' m',
    nres r
    -> expandm A r m
    -> inst f i r r'
    -> rsub f i m m'
    -> expandm A r' m'.
Proof.
intros A r m f i r' m' His Hexpand Hsubr Hsubm.
revert f i r' m' His Hsubr Hsubm.
induct Hexpand.

(* atom *)
intros c s r f i r' m' _ Hsubr Hsub.
invert Hsub; clear Hsub; [].
intros r'' Hsubr' <-.
so (rsub_fun _#5 (inst_impl_rsub _#4 Hsubr) Hsubr'); subst r''.
apply expandm_atom; done.

(* up *)
intros A r p v r' HexpandA Hcompose f i r1 mm His Hsubr1 Hsub.
invert His; clear His; [].
intros h s His <-.
invert Hcompose; clear Hcompose; [].
intros s' Hcompose <-.
invert Hsub; clear Hsub; [].
intros r2 Hsubr2 <-.
invertc Hsubr1; [].
intros h' s1 Hsubh Hsubs1 <-.
#2 invert Hsubr2; clear Hsubr2.
  intros h'' s2 Hsubh' Hsubs2 <-.
  so (rsub_head_fun _#4 Hsubh Hsubh'); subst h''.
  eapply expandm_up; eauto; [].
  so (expandp_closed_rsub f i _#3 HexpandA) as Hsubv.
  so (rsub_pm (rsub_ret Hsubv)) as Hsubpm.
  so (sub_into_compose _#7 His Hcompose Hsubs1 Hsubpm) as (s2' & Hsubs2' & Hcompose').    
  so (rsub_fun _#5 Hsubs2 Hsubs2'); subst s2'.
  apply compose_use; [].
  exact Hcompose'.

  intros m _ _ <-.
  invert Hsubh; intros; omega.

(* pi *)
intros p' A B r p v r1 m HexpandA Hcompose HexpandB IH f i r' mm His Hsubr Hsubmm.
invert Hsubmm; clear Hsubmm; [].
intros m' Hsubm <-.
sor (refine (inst_sh_commute_right _ _ 0 (bound p) _ _ _ Hsubr)) as Hsubr'; [omega |].
repl (i + bound p) with (bound p + i) in Hsubr' by omega.
so (expandp_closed_rsub f i _#3 HexpandA) as Hsubv.
so (inst_into_compose _#7 (sh_nres _#3 His) Hcompose Hsubr' (rsub_app Hsubv rsub_nl)) as (r1' & Hsubr1 & Hcompose').
so (compose_nres_nspine _#3 Hcompose (sh_nres _#3 His) (nspine_app nspine_nl)) as His_r1.
so (IH _#4 His_r1 Hsubr1 Hsubm) as Hexpand.
eapply expandm_pi; eauto; done.

(* wth *)
intros A B r r1 r2 m1 m2 Hcompose1 Hcompose2 _ IH1 _ IH2 f i r' mm His Hsubr Hsubmm.
invert Hsubmm; clear Hsubmm; [].
intros m1' m2' Hsubm1 Hsubm2 <-.
so (inst_into_compose _#7 His Hcompose1 Hsubr (rsub_pi1 rsub_nl)) as (r1' & Hsubr1 & Hcompose1').
so (inst_into_compose _#7 His Hcompose2 Hsubr (rsub_pi2 rsub_nl)) as (r2' & Hsubr2 & Hcompose2').
so (compose_nres_nspine _#3 Hcompose1 His (nspine_pi1 nspine_nl)) as His_r1.
so (compose_nres_nspine _#3 Hcompose2 His (nspine_pi2 nspine_nl)) as His_r2.
so (IH1 _#4 His_r1 Hsubr1 Hsubm1) as Hexpand1.
so (IH2 _#4 His_r2 Hsubr2 Hsubm2) as Hexpand2.
eapply expandm_wth; eauto; done.

(* sigma *)
intros A B r r1 r2 m1 m2 Hcompose1 Hcompose2 _ IH1 _ IH2 f i r' mm His Hsubr Hsubmm.
invert Hsubmm; clear Hsubmm; [].
intros m1' m2' Hsubm1 Hsubm2 <-.
so (inst_into_compose _#7 His Hcompose1 Hsubr (rsub_pi1 rsub_nl)) as (r1' & Hsubr1 & Hcompose1').
so (inst_into_compose _#7 His Hcompose2 Hsubr (rsub_pi2 rsub_nl)) as (r2' & Hsubr2 & Hcompose2').
so (compose_nres_nspine _#3 Hcompose1 His (nspine_pi1 nspine_nl)) as His_r1.
so (compose_nres_nspine _#3 Hcompose2 His (nspine_pi2 nspine_nl)) as His_r2.
so (IH1 _#4 His_r1 Hsubr1 Hsubm1) as Hexpand1.
so (IH2 _#4 His_r2 Hsubr2 Hsubm2) as Hexpand2.
eapply expandm_sigma; eauto; done.
Qed.


Lemma cutp_into_expandm :
  forall A r m v p i r' m',
    nres r
    -> rigid i (bound p) r
    -> expandm A r m
    -> cutp v p i r r'
    -> cutp v p i m m'
    -> expandm A r' m'.
Proof.
intros A r m v p i r' m' Hisr Hrigid HexpandA Hsubr Hsubm.
revert A r m v i r' m' Hisr Hrigid HexpandA Hsubr Hsubm.
induct p.

(* down *)
intros A r m v i r' m' Hisr Hrigid HexpandA Hcutr Hcutm.
invertc Hcutr; [].
intros f Hsubr <-.
invertc Hcutm; [].
intros Hsubm.
simpl in Hrigid.
sor (refine (rsub_rigid_invert _#6 Hsubr Hrigid _ _)) as Hinst; try omega; [].
eapply inst_into_expandm; eauto; done.

(* one *)
intros A r m v i r' m' Hisr Hrigid HexpandA Hsubr Hsubm.
invert Hsubr; [].
intros <- <-.
invert Hsubm; [].
intros <-.
exact HexpandA.

(* tens *)
intros p IH1 q IH2 A r m vv i r' m' Hisr Hrigid HexpandA Hsubr Hsubm.
simpl in Hrigid.
invertc Hsubr; [].
intros v w ra Hsubra Hsubrb <-.
invertc Hsubm; [].
intros ma Hsubma Hsubmb.
assert (rigid (bound q + i) (bound p) r) as Hrigid'.
  refine (rigid_adjust _#5 _ _ Hrigid); omega.
so (IH1 _#7 Hisr Hrigid' HexpandA Hsubra Hsubma) as Hexpand'.
so (cutp_nres _#5 Hrigid' Hsubra Hisr) as Hisra.
sor (refine (cutp_rigid _#7 _ _ Hrigid Hsubra)) as Hrigida; try omega; [].
repl (bound p + bound q - bound p) with (bound q) in Hrigida by omega.
exact (IH2 _#7 Hisra Hrigida Hexpand' Hsubrb Hsubmb).
Qed.


Lemma compose_expand_cutp_second :
  forall r s r' A m v p i m',
    nres r
    -> nspine s
    -> rigid i (bound p) r
    -> compose r s r'
    -> expandm A r' m
    -> cutp v p i m m'
    -> exists s', cutp v p i s s'.
Proof.
intros r s r' A m v p i m' Hisr Hiss Hrigid Hcompose Hexpand Hsub.
revert r s v p i m' Hisr Hiss Hrigid Hcompose Hsub.
induct Hexpand.

(* atom *)
intros _ _ r r1 s2 v p i mm Hisr1 Hiss2 Hrigid Hcompose Hsub.
#2 exploit (cutp_invert1 etam) as Hinvert.
  intros f i0 d e H; invert H; intros; eauto; done.
so (Hinvert _#5 Hsub) as (r' & -> & Hsubr).
eapply cutp_into_rcompose_reverse; eauto; done.

(* up *)
intros A r12 p v r123 Hexpand Hcompose12_3 r1 s2 w q i m' Hisr1 Hiss2 Hrigid Hcompose12 Hsub.
#2 exploit (cutp_invert1 shiftp) as Hinvert.
  intros f i0 d e H; invert H; intros; eauto; done.
so (Hinvert _#5 Hsub) as (r123' & -> & Hsubr123).
so (compose_assoc _#5 Hcompose12 Hcompose12_3) as (s23 & Hcompose23 & Hcompose1_23).
so (cutp_into_rcompose_reverse _#7 Hisr1 Hrigid Hcompose1_23 Hsubr123) as (s23' & Hsubs23).
exact (cutp_into_compose_reverse _#7 Hiss2 Hcompose23 Hsubs23 andel).

(* pi *)
intros p' A B r12 p v r' m HexpandA Hcompose12_3 HexpandB IH r1 s2 w q i mm Hisr1 Hiss2 Hrigid Hcompose12 Hsub.
#2 exploit (cutp_invert1_sh (lam p) (bound p)) as Hinvert.
  intros f i0 d e H; invert H; eauto; done.
so (Hinvert _#5 Hsub) as (m' & -> & Hsubm).
so (sh_compose 0 (bound p) _#3 Hcompose12) as Hcompose12'.
so (compose_assoc _#5 Hcompose12' Hcompose12_3) as (s23 & Hcompose23 & Hcompose1_23).
so (IH _#6 (sh_nres _#3 Hisr1) (compose_nspine _#3 Hcompose23 (sh_nspine _#3 Hiss2) (nspine_app nspine_nl)) (sh_rigid _#4 Hrigid) Hcompose1_23 Hsubm) as (s23' & Hsubs23).
so (cutp_into_compose_reverse _#7 (sh_nspine _#3 Hiss2) Hcompose23 Hsubs23) as ((s2' & Hsubs2) & _).
repl (bound p + i) with (i + 0 + bound p) in Hsubs2 by omega.
so (cutp_of_sh _#7 Hsubs2) as (s2'' & Hsubs2' & ->).
repl (i + 0) with i in Hsubs2' by omega.  
eauto; done.

(* wth *)
intros A B r12 r123 _ m1 m2 Hcompose12_3 _ _ IH _ _ r1 s2 w q i mm Hisr1 Hiss2 Hrigid Hcompose12 Hsub.
#2 exploit (cutp_invert2 pairm) as Hinvert.
  intros f i0 d1 d2 e H; invert H; eauto; done.
so (Hinvert _#6 Hsub) as (m1' & _ & _ & Hsubm1 & _).
so (compose_assoc _#5 Hcompose12 Hcompose12_3) as (s23 & Hcompose23 & Hcompose1_23).
so (IH _#6 Hisr1 (compose_nspine _#3 Hcompose23 Hiss2 (nspine_pi1 nspine_nl)) Hrigid Hcompose1_23 Hsubm1) as (s23' & Hsubs23).
so (cutp_into_compose_reverse _#7 Hiss2 Hcompose23 Hsubs23) as ((s2' & Hsubs2) & _).
eauto; done.

(* sigma *)
intros A B r12 r123 _ m1 m2 Hcompose12_3 _ _ IH _ _ r1 s2 w q i mm Hisr1 Hiss2 Hrigid Hcompose12 Hsub.
#2 exploit (cutp_invert2 pairm) as Hinvert.
  intros f i0 d1 d2 e H; invert H; eauto; done.
so (Hinvert _#6 Hsub) as (m1' & _ & _ & Hsubm1 & _).
so (compose_assoc _#5 Hcompose12 Hcompose12_3) as (s23 & Hcompose23 & Hcompose1_23).
so (IH _#6 Hisr1 (compose_nspine _#3 Hcompose23 Hiss2 (nspine_pi1 nspine_nl)) Hrigid Hcompose1_23 Hsubm1) as (s23' & Hsubs23).
so (cutp_into_compose_reverse _#7 Hiss2 Hcompose23 Hsubs23) as ((s2' & Hsubs2) & _).
eauto; done.
Qed.


Inductive pcut : term -> spine -> term -> Prop :=
| pcut_nl {m}
    : pcut m nl m

| pcut_app {p m v s m' n}
    : cutp v p 0 m m'
      -> pcut m' s n
      -> pcut (lam p m) (app v s) n

| pcut_pi1 {m n s l}
    : pcut m s l
      -> pcut (pairm m n) (pi1 s) l

| pcut_pi2 {m n s l}
    : pcut n s l
      -> pcut (pairm m n) (pi2 s) l
.


Lemma split_cut :
  forall s1 s2 s m r,
    nspine s1
    -> compose s1 s2 s
    -> cutm m s r
    -> exists m',
         pcut m s1 m'
         /\ cutm m' s2 r.
Proof.
intros s1 s2 s m r His Hcompose Hcut.
revert s2 s m r Hcompose Hcut.
induct His.

(* nl *)
intros s2 s m r Hcompose Hcut.
invertc Hcompose; [].
intros <-.
eauto using pcut; done.

(* app *)
intros v s1 _ IH s2 s m r Hcompose Hcut.
invertc Hcompose; [].
intros s3 Hcompose' <-.
invertc Hcut; [].
intros p n n' Hcutp Hcutm <-.
so (IH _#4 Hcompose' Hcutm) as (l & Hpcut & Hcutm').
exists l.
split; auto; [].
eapply pcut_app; eauto; done.

(* pi1 *)
intros s1 _ IH s2 s m r Hcompose Hcut.
invertc Hcompose; [].
intros s3 Hcompose' <-.
invertc Hcut; [].
intros n1 n2 Hcutm <-.
so (IH _#4 Hcompose' Hcutm) as (l & Hpcut & Hcutm').
exists l.
split; auto; [].
eapply pcut_pi1; eauto; done.

(* pi2 *)
intros s1 _ IH s2 s m r Hcompose Hcut.
invertc Hcompose; [].
intros s3 Hcompose' <-.
invertc Hcut; [].
intros n1 n2 Hcutm <-.
so (IH _#4 Hcompose' Hcutm) as (l & Hpcut & Hcutm').
exists l.
split; auto; [].
eapply pcut_pi2; eauto; done.
Qed.


Lemma split_pcut :
  forall s1 s2 s m m',
    compose s1 s2 s
    -> pcut m s m'
    -> exists m'',
         pcut m s1 m''
         /\ pcut m'' s2 m'.
Proof.
intros s1 s2 s m m' Hcompose.
revert m m'.
pattern s1, s.
match goal with
| |- ?Q _ _ =>
    cut (forall z (E : z = sp) (s1 s : exp z),
           compose s1 s2 s
           -> Q (eq_rect _ _ s1 _ E) (eq_rect _ _ s _ E))
end. 
  intros Hind.
  so (Hind sp (eq_refl _) s1 s).
  intros; eauto; done.
clear Hcompose s1 s.
intros z E s1 s Hcompose m m' Hcut.
revert E m m' Hcut.
induct Hcompose; try (revertnew (intros; substrefls)).

(* nl *)
intros s m m' Hcut.
eauto using pcut; done.

(* pm *)
intros p r1 s2 r _ _ m m' Hcut.
invertc Hcut; done.

(* app *)
intros v s1 s2 s _ IH m m' Hcut.
invertc Hcut; [].
intros p n n' Hcutp Hcut' <-.
so (IH (eq_refl _) _ _ Hcut') as (n'' & Hcut1 & Hcut2).
exists n''.
split; eauto using pcut; done.

(* pi1 *)
intros s1 s2 s _ IH m m' Hcut.
invertc Hcut; [].
intros n1 n2 Hcut <-.
so (IH (eq_refl _) _ _ Hcut) as (n & Hcut1 & Hcut2).
exists n.
split; eauto using pcut; done.

(* pi2 *)
intros s1 s2 s _ IH m m' Hcut.
invertc Hcut; [].
intros n1 n2 Hcut <-.
so (IH (eq_refl _) _ _ Hcut) as (n & Hcut1 & Hcut2).
exists n.
split; eauto using pcut; done.

(* use *)
intros h s1 s2 s _ _ E m m' Hcut.
discriminate E.
Qed.


(* You can use split_cut, but this is actually shorter. *)
Lemma nspine_cut :
  forall m s r,
    nspine s
    -> cutm m s r
    -> pcut m s (etam r).
Proof.
intros m s r His Hcut.
revert m r Hcut.
induct His; intros; invert Hcut; intros; subst; eauto using pcut.
Qed.


Lemma pcut_fun :
  forall m s m1 m2,
    pcut m s m1
    -> pcut m s m2
    -> m1 = m2.
Proof.
intros m s m1 m2 H Hcut.
revert m2 Hcut.
induct H; try (intros; invert Hcut; intros; subst; eauto; f_equal; eauto; done); [].
(* app *)
intros p m v s m' n Hcutp _ IH m2 Hcut.
invertc Hcut; [].
intros m'' Hcutp' Hpcut.
so (cutp_fun _#6 Hcutp Hcutp'); subst m''.
apply IH; auto; done.
Qed.


Lemma sh_pcut :
  forall m s m' i j,
    pcut m s m'
    -> pcut (sh i j m) (sh i j s) (sh i j m').
Proof.
intros m s m' i j Hcut.
revert i j.
induct Hcut;
try (intros; eauto using pcut; done); [].
(* app *)
intros p m v s m' n Hcutp _ IH i j.
simpl.
eapply pcut_app; eauto; [].
repl (bound p + i) with (0 + bound p + i) by omega.
repl i with (0 + i) at 3 by omega.
apply cutp_sh_commute_left; eauto.
Qed.


Lemma sub_into_pcut :
  forall f i m1 m1' s s' m2,
    pcut m1 s m2
    -> rsub f i m1 m1'
    -> rsub f i s s'
    -> exists m2',
         rsub f i m2 m2'
         /\ pcut m1' s' m2'.
Proof.
intros f i m1 m1' s s' m2 Hcut Hsubm1 Hsubs.
revert m1' s' Hsubm1 Hsubs.
induct Hcut.

(* nl *)
intros m m1' s' Hsubm1 Hsubs.
invertc Hsubs; [].
intros <-.
exists m1'; eauto using pcut; done.

(* app *)
intros p m v s m2 n Hcutp _ IH mm ss Hsublam Hsubapp.
invertc Hsublam; [].
intros m1 Hsubm1 <-.
invertc Hsubapp; [].
intros v' s' Hsubv Hsubs <-.
repl i with (bound p + i - 0 - bound p) in Hsubv by omega.
sor (refine (commute_cutp_rsub _#10 _ Hsubm1 Hcutp Hsubv)) as (m' & Hcutp' & Hsubm2); [omega |].
repl (bound p + i - bound p) with i in Hsubm2 by omega.
so (IH _ _ Hsubm2 Hsubs) as (n' & Hsubn & Hcutm).
exists n'.
split; eauto using pcut; done.

(* pi1 *)
intros m n s l _ IH mm ss Hsubmm Hsubss.
invertc Hsubmm; [].
intros m' n' Hsubm Hsubn <-.
invertc Hsubss; [].
intros s' Hsubs <-.
so (IH _ _ Hsubm Hsubs) as (l' & Hsubl & Hcut).
exists l'; eauto using pcut; done.

(* pi2 *)
intros m n s l _ IH mm ss Hsubmm Hsubss.
invertc Hsubmm; [].
intros m' n' Hsubm Hsubn <-.
invertc Hsubss; [].
intros s' Hsubs <-.
so (IH _ _ Hsubn Hsubs) as (l' & Hsubl & Hcut).
exists l'; eauto using pcut; done.
Qed.


Lemma cutp_into_pcut :
  forall v p i m1 m1' s s' m2,
    pcut m1 s m2
    -> cutp v p i m1 m1'
    -> cutp v p i s s'
    -> exists m2',
         cutp v p i m2 m2'
         /\ pcut m1' s' m2'.
Proof.
intros v p i m1 m1' s s' m2 Hpcut Hsubm1 Hsubs.
revert v i m1 m1' s s' m2 Hpcut Hsubm1 Hsubs.
induct p.

(* down *)
intros v i m1 m1' s s' m2 Hpcut Hcutm1 Hcuts.
invertc Hcutm1; [].
intros m Hsubm1 <-.
invertc Hcuts; [].
intros Hsubs.
so (sub_into_pcut _#7 Hpcut Hsubm1 Hsubs) as (m2' & Hsubm2 & Hpcut').
exists m2'.
split; eauto using @cutp; done.

(* one *)
intros v i m1 m1' s s' m2 Hpcut Hsubm1 Hsubs.
invertc Hsubm1; [].
intros <- <-.
invertc Hsubs; [].
intros <-.
exists m2.
split; eauto using @cutp; done.

(* tens *)
intros p IH1 q IH2 vv i m1 m1'' s s'' m2 Hpcut Hsubm1 Hsubs.
invertc Hsubm1; [].
intros v w m1' Hsubm1' Hsubm1'' <-.
invertc Hsubs; [].
intros s' Hsubs' Hsubs''.
so (IH1 _#7 Hpcut Hsubm1' Hsubs') as (m2' & Hsubm2 & Hpcut1).
so (IH2 _#7 Hpcut1 Hsubm1'' Hsubs'') as (m2'' & Hsubsm2' & Hpcut2).
exists m2''.
split; eauto using @cutp; done.
Qed.


Lemma pcut_of_sh :
  forall i j m1 s m2,
    pcut (sh i j m1) (sh i j s) m2
    -> exists m2',
         pcut m1 s m2'
         /\ m2 = sh i j m2'.
Proof.
intros i j m1 s m2 Hcut.
so (dummy_sub j) as (v & p & Hmatch & Hlen & _).
so (cutp_sh_cancel_all v p i m1 Hmatch) as Hsubm1.
rewrite -> Hlen in Hsubm1; [].
so (cutp_sh_cancel_all v p i s Hmatch) as Hsubs.
rewrite -> Hlen in Hsubs; [].
so (cutp_into_pcut _#8 Hcut Hsubm1 Hsubs) as (m2' & Hsubm2 & Hcut').
exists m2'.
split; auto; [].
so (sh_pcut _#3 i j Hcut') as Hcut''.
so (pcut_fun _#4 Hcut Hcut''); subst m2.
reflexivity.
Qed.


Definition cutm_expand_prop (N : nat) :=
  forall A r m s r',
    size A N
    -> nres r
    -> expandm A r m
    -> cutm m s r'
    -> compose r s r'.

Definition cutp_expand_prop (N : nat) :=
  forall z A p v i (e e' : exp z),
    size A N
    -> expandp A p v
    -> cutp v p i (sh (i + bound p) (bound p) e) e'
    -> e = e'.

Definition rsub_expand_prop (N : nat) :=
  forall z A f i (e e' : exp z),
    size A N
    -> expandm A (use (var 0) nl) f
    -> rsub f i (sh (i+1) 1 e) e'
    -> e = e'.

Definition lsub_expand_prop (N : nat) :=
  forall z A p v (e : exp z) q r e' r',
    size A N
    -> expandp A p v
    -> lsub e q r e'
    -> compose r' (pm p (ret v)) e
    -> compose r' (pm q r) e'.

Definition sub_into_expandm_prop (N : nat) :=
  forall A i s m n m' s',
    size A N
    -> nspine s
    -> expandm A (use (var i) s) m
    -> rsub n i m m'
    -> rsub n i s s'
    -> pcut (sh 0 i n) s' m'.

Definition sub_into_expandp_prop (N : nat) :=
  forall A p v w v',
    size A N
    -> expandp A p v
    -> cutp w p 0 v v'
    -> v' = w.


Lemma magnificent_expansion_lemma :
  forall N,
    cutm_expand_prop N
    /\ cutp_expand_prop N
    /\ rsub_expand_prop N
    /\ lsub_expand_prop N
    /\ sub_into_expandm_prop N
    /\ sub_into_expandp_prop N.
Proof.
intro N.
wfinduct N using lt_wf.
intros N IH.
so (fun N' Hlt => IH N' Hlt andel) as IHcutm.
so (fun N' Hlt => IH N' Hlt anderl) as IHcutp.
so (fun N' Hlt => IH N' Hlt anderrl) as IHrsub.
so (fun N' Hlt => IH N' Hlt anderrrl) as IHlsub.
so (fun N' Hlt => IH N' Hlt anderrrrl) as IHexpandm.
so (fun N' Hlt => IH N' Hlt anderrrr ander) as IHexpandp.
clear IH.




(* cutm_expand *)
assert (cutm_expand_prop N) as Hind.

clear IHexpandm.
unfold cutm_expand_prop.
intros A r m s r' Hsize His Hexpand Hcutm.
revert A r Hsize His Hexpand.
cases Hcutm.

(* atom *)
intros r s r' _ His Hexp.
invert Hexp; [].
intros c s' <- ->.
apply nres_compose_nl; auto; done.

(* up A *)
intros r1 p r2 r Hlsub AA r' Hsize His Hexpand.
invert Hexpand; clear Hexpand; [].
intros A p' v HexpandA Hcompose <-.
invertc Hsize; [].
intros N' Hsize' <-.
eapply IHlsub; eauto; done.

(* pi *)
intros p m w s r' m' Hcutp Hcutm AA r Hsize His Hexpand.
invert Hexpand; [].
intros p' A B v rv HexpandA Hcompose_rv HexpandB <-.
invertc Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
assert (exists v', cutp w p 0 v v') as (v' & Hsubv).
  sor (refine (sh_implies_rigid 0 (bound p) (bound p) r _ His)) as Hrigid; [omega |].
  so (compose_expand_cutp_second _#9 (sh_nres _#3 His) (nspine_app nspine_nl) Hrigid Hcompose_rv HexpandB Hcutp) as (ss & Hcuts).
  #2 exploit (cutp_invert2 app) as Hinvert.
    intros f i d1 d2 e H.
    invert H; intros; subst; eauto; done.
  so (Hinvert _#6 Hcuts) as (v' & _ & _ & Hcutv & _).
  eauto; done.
unfold sub_into_expandp_prop in IHexpandp.
sor (refine (IHexpandp N1 _ _#5 HsizeA HexpandA Hsubv)) as ?; [omega |]; subst v'.
exploit (cutp_sh_cancel w p 0 0 (bound p) r) as Hunsh; eauto using cutp_patmatch; try omega; [].
repl (bound p - bound p) with 0 in Hunsh by omega.
rewrite -> sh_zero in Hunsh; [].
exploit (cutp_compat2 w p 0 (fun s (_ : exp vl) => app s nl) v v w w) as Hsubvn; eauto using @rsub; [].
simpl in Hsubvn.  
sor (refine (sh_implies_rigid 0 (bound p) (bound p) _ _ His)) as Hrigid; [omega |].
so (cutp_into_rcompose _#8 (sh_nres _#3 His) Hcompose_rv Hrigid Hunsh Hsubvn) as (rv' & Hsubrv & Hcompose_rw).
so (compose_nres_nspine _#3 Hcompose_rv (sh_nres _#3 His) (nspine_app nspine_nl)) as Hisrv.
so (compose_nres_nspine _#3 Hcompose_rw His (nspine_app nspine_nl)) as Hisrv'.
so (cutp_into_expandm _#8 Hisrv (rigid_compose _#5 Hcompose_rv Hrigid) HexpandB Hsubrv Hcutp) as HexpandB'.
sor (refine (IHcutm N2 _ _#5 HsizeB Hisrv' HexpandB' Hcutm)) as Hcompose_rv'_s; [omega |].
so (compose_assoc _#5 Hcompose_rw Hcompose_rv'_s) as (ws & Hcompose_w_s & Hcompose_r_ws).
invert Hcompose_w_s; clear Hcompose_w_s; [].
intros s' Hcompose_nl_s <-.
invert Hcompose_nl_s; clear Hcompose_nl_s; [].
intros <-.
exact Hcompose_r_ws.

Lemma invert_expand_pair :
  forall A r m n N,
    expandm A r (pairm m n)
    -> size A N
    -> exists B C r1 r2 N1 N2,
         compose r (pi1 nl) r1
         /\ compose r (pi2 nl) r2
         /\ expandm B r1 m
         /\ expandm C r2 n
         /\ size B N1
         /\ size C N2
         /\ 1 + N1 + N2 = N.
Proof.
intros A r m n N Hexpand Hsize.
invert Hexpand; intros; subst; invert Hsize; intros; do 6 eexists; eauto 8.
Qed.

(* wth1 *)
intros m n s r Hcutm A r' Hsize His Hexpand.
so (invert_expand_pair _#5 Hexpand Hsize) as (B & C & r1 & _ & N1 & N2 & Hcompose1 & _ & HexpandB & _ & HsizeB & _ & <-).
so (compose_nres_nspine _#3 Hcompose1 His (nspine_pi1 nspine_nl)) as His'.
sor (refine (IHcutm _ _ _#5 HsizeB His' HexpandB Hcutm)) as Hcompose_r1_s_r; [omega |].
so (compose_assoc _#5 Hcompose1 Hcompose_r1_s_r) as (s1' & Hcompose_pi & Hcompose_s'_s1').
invert Hcompose_pi; [].
intros ? H <-.
invert H; [].
intros <-.
exact Hcompose_s'_s1'.

(* wth2 *)
intros m n s r Hcutm A r' Hsize His Hexpand.
so (invert_expand_pair _#5 Hexpand Hsize) as (C & B & _ & r1 & N2 & N1 & _ & Hcompose1 & _ & HexpandB & _ & HsizeB & <-).
so (compose_nres_nspine _#3 Hcompose1 His (nspine_pi2 nspine_nl)) as His'.
sor (refine (IHcutm _ _ _#5 HsizeB His' HexpandB Hcutm)) as Hcompose_r1_s_r; [omega |].
so (compose_assoc _#5 Hcompose1 Hcompose_r1_s_r) as (s1' & Hcompose_pi & Hcompose_s'_s1').
invert Hcompose_pi; [].
intros ? H <-.
invert H; [].
intros <-.
exact Hcompose_s'_s1'.

(* wrapup *)
split; auto; [].
assert (forall N', N' <= N -> cutm_expand_prop N') as H.
  intros N' Hleq.
  destruct (le_decide _ _ Hleq) as [Hlt | Heq].
    apply IHcutm; auto; done.

    subst N'; exact Hind.
renameover H into IHcutm.
clear Hind.




(* cutp_expand_prop *)
assert (cutp_expand_prop N) as IHcutpeq.
  clear IHlsub IHexpandm IHexpandp IHcutm.
  unfold cutp_expand_prop.
  intros z A p v i e e' Hsize Hexpand Hcutp.
  revert z i e e' Hsize Hcutp.
  cases Hexpand.
    (* down *)
    intros A m HexpandA z i e e' Hsize Hcutp.
    simpl in Hcutp.
    invertc Hsize; [].
    intros N' HsizeA <-.
    invertc Hcutp; [].
    intros Hsubm.
    unfold rsub_expand_prop in IHrsub.
    sor (refine (IHrsub N' _ _#6 HsizeA HexpandA Hsubm)) as ?; [omega |].
    subst; reflexivity.

    (* one *)
    intros z i e e' Hsize Hcutp.
    invert Hcutp; [].
    intros <-.
    rewrite -> sh_zero; [].
    reflexivity.

    (* tens *)
    intros A B p q v w HexpandA HexpandB z i e e' Hsize Hcutp.
    simpl in Hcutp.
    invertc Hsize; [].
    intros N1 N2 HsizeA HsizeB <-.
    invertc Hcutp; [].
    intros e2 Hcut1 Hcut2.
    unfold cutp_expand_prop in IHcutp.
    repl (i + (bound p + bound q)) with (bound q + i + 0 + bound p) in Hcut1 by omega.
    repl (bound p + bound q) with (bound q + bound p) in Hcut1 by omega.
    rewrite <- sh_sum in Hcut1; [].
    so (cutp_of_sh_left _#7 Hcut1) as (e2' & Hcut1' & ->).
    clear Hcut1.
    repl (bound q + i + 0 + bound p) with (bound q + i + bound p) in Hcut1' by omega.
    sor (refine (IHcutp N1 _ _#7 HsizeA HexpandA Hcut1')) as ?; [omega |].
    subst e2'.
    repl (bound q + i + 0) with (i + bound q) in Hcut2 by omega.
    sor (refine (IHcutp N2 _ _#7 HsizeB HexpandB Hcut2)) as ?; [omega |].
    subst; reflexivity.

    (* ex *)
    intros p' A B p q v w HexpandA HexpandB z i e e' Hsize Hcutp.
    simpl in Hcutp.
    invertc Hsize; [].
    intros N1 N2 HsizeA HsizeB <-.
    invertc Hcutp; [].
    intros e2 Hcut1 Hcut2.
    unfold cutp_expand_prop in IHcutp.
    repl (i + (bound p + bound q)) with (bound q + i + 0 + bound p) in Hcut1 by omega.
    repl (bound p + bound q) with (bound q + bound p) in Hcut1 by omega.
    rewrite <- sh_sum in Hcut1; [].
    so (cutp_of_sh_left _#7 Hcut1) as (e2' & Hcut1' & ->).
    clear Hcut1.
    repl (bound q + i + 0 + bound p) with (bound q + i + bound p) in Hcut1' by omega.
    sor (refine (IHcutp N1 _ _#7 HsizeA HexpandA Hcut1')) as ?; [omega |].
    subst e2'.
    repl (bound q + i + 0) with (i + bound q) in Hcut2 by omega.
    sor (refine (IHcutp N2 _ _#7 HsizeB HexpandB Hcut2)) as ?; [omega |].
    subst; reflexivity.
split; [exact IHcutpeq |].




(* rsub_expand *)
split.
clear IHrsub IHlsub IHexpandm IHexpandp IHcutpeq.
unfold rsub_expand_prop.
intros z A f i d e' Hsize HexpandG Hrsub.
remember (sh (i+1) 1 d) as e eqn:Heqd.
revert A d Hsize Heqd HexpandG.
#7 induct Hrsub;
try (intros;
     revert Heqd;
     cases d; try (intros; discriminate Heqd);
     intros;
     try (injection Heqd);
     intros;
     subst;
     f_equal; eauto;
     done).

(* inst *)
intros f i hh ss h' s' Hsubh Hsubs IH A e Hsize Heq HexpandA.
revert Heq.
cases e; try (intros; simpl; discriminate Heq); [].
intros h s Heq.
injection Heq; clear Heq; intros; subst hh ss.
so (IH _#2 Hsize (eq_refl _) HexpandA); subst s'.
f_equal; [].
destruct h as [c | k].
  invert Hsubh; auto; done.

  #2 invert Hsubh.
    intros Hlt <-.
    f_equal; [].
    unfold sh_var; unfold sh_var in Hlt.
    forget (lt_dec k (i+1)) as b.
    destruct b; omega.

    intros Hgeq <-.
    f_equal; [].
    unfold sh_var; unfold sh_var in Hgeq.
    forget (lt_dec k (i+1)) as b.
    destruct b; omega.

(* reduce *)
intros f i ss r s' Hsubs IH1 Hcutm A e Hsize Heq HexpandA.
revert Heq; cases e; try (intros; simpl; discriminate Heq); [].
intros h s Heq.
injection Heq; clear Heq; intros -> Heqh.
so (IH1 _#2 Hsize (eq_refl _) HexpandA); subst s'.
destruct h as [c | k].
  simpl in Heqh.
  discriminate Heqh.
simpl in Heqh.
injection Heqh; clear Heqh; [].
intros Heqk.
assert (i = k).
  unfold sh_var in Heqk.
  forget (lt_dec k (i+1)) as b; destruct b; try omega.
subst k.
so (sh_expandm_sh 0 i _#3 HexpandA) as HexpandA'.
simpl in HexpandA'.
repl (sh_var 0 i 0) with i in HexpandA'.
  unfold sh_var.
  forget (lt_dec 0 0) as b.
  destruct b; omega.
unfold cutm_expand_prop in IHcutm.
so (IHcutm _ (le_refl _) _#5 (shift_size _#4 Hsize) (nres_use nspine_nl) HexpandA' Hcutm) as Hcompose.
invert Hcompose; [].
intros s' Hcompose' <-.
invert Hcompose'; [].
intros <-.
reflexivity.

(* pm *)
intros f i p r r' _ IH G d Hsize Heqd HexpandG.
revert Heqd.
cases d; try (intros; discriminate Heqd); [].
intros q  r'' Heq.
injection Heq.
intros -> <-.
f_equal; [].
eapply IH; eauto; [].
f_equal; omega.

(* lam *)
intros f i p n n' _ IH G d Hsize Heqd HexpandG.
revert Heqd.
cases d; try (intros; discriminate Heqd); [].
intros q n'' Heq.
injection Heq.
intros -> <-.
f_equal; [].
eapply IH; eauto; [].
f_equal; omega.

(* pi *)
intros f i p A B A' B' _ IH1 _ IH2 G d Hsize Heqd HexpandG.
revert Heqd.
cases d; try (intros; discriminate Heqd); [].
intros q A'' B'' Heq.
injection Heq.
intros -> -> <-.
f_equal; eauto; [].
eapply IH2; eauto; [].
f_equal; omega.

(* ex *)
intros f i p A B A' B' _ IH1 _ IH2 G d Hsize Heqd HexpandG.
revert Heqd.
cases d; try (intros; discriminate Heqd); [].
intros q A'' B'' Heq.
injection Heq.
intros -> -> <-.
f_equal; eauto; [].
eapply IH2; eauto; [].
f_equal; omega.

(* tpi *)
intros f i p A B A' B' _ IH1 _ IH2 G d Hsize Heqd HexpandG.
revert Heqd.
cases d; try (intros; discriminate Heqd); [].
intros q A'' B'' Heq.
injection Heq.
intros -> -> <-.
f_equal; eauto; [].
eapply IH2; eauto; [].
f_equal; omega.




(* lsub_expand *)
split.
clear IHlsub IHexpandm IHexpandp.
unfold lsub_expand_prop.
intros z A p v e q r e' r' Hsize Hexpandp Hlsub Hcompose.
revert A p v r' Hsize Hexpandp Hcompose.
induct Hlsub.

(* ret *)
intros v p r r' Hcutp A p' v' r'' Hsize Hexpandp Hcompose.
invert Hcompose; done.

(* use *)
intros h s p r s' Hsub IH A p' v r' Hsize Hexpand Hcompose.
invert Hcompose; clear Hcompose; [].
intros s'' Hcompose' <-.
apply compose_use; [].
eapply IH; eauto; done.

(* pm *)
intros p1 r1 p2 r2 r Hsub IH A p v r' Hsize Hexpandp Hcompose.
#2 invert Hcompose; clear Hcompose.
  intros <- <- <-.
  invertc Hsub; [].
  intros Hcutp.
  so (patmatch_fun _#3 (cutp_patmatch _#5 Hcutp) (expandp_patmatch _#3 Hexpandp)); subst p2.
  unfold cutp_expand_prop in IHcutpeq.
  repl (bound p) with (0 + bound p) in Hcutp at 2 by omega.
  so (IHcutpeq _#7 Hsize Hexpandp Hcutp); subst r.
  apply compose_nl; done.

  intros r'' Hcompose' <-.
  apply compose_pm; [].
  simpl.
  repl (bound p2 + 0) with (bound p2) by omega.
  simpl in Hcompose'.
  repl (bound p + 0) with (bound p) in Hcompose' by omega.
  so (sh_expandp_sh 0 (bound p1) _#3 Hexpandp) as Hexpandp'.
  exact (IH _#4 (shift_size _#4 Hsize) Hexpandp' Hcompose').

(* app *)
intros v' s p r s' _ IH A p' v r' Hsize HexpandA Hcompose.
invert Hcompose; clear Hcompose; [].
intros s'' Hcompose' <-.
apply compose_app; [].
eapply IH; eauto; done.

(* pi1 *)
intros s p r s' _ IH A p' v r' Hsize Hexpand Hcompose.
invert Hcompose; clear Hcompose; [].
intros s'' Hcompose' <-.
apply compose_pi1; [].
eapply IH; eauto; done.

(* pi2 *)
intros s p r s' _ IH A p' v r' Hsize Hexpand Hcompose.
invert Hcompose; clear Hcompose; [].
intros s'' Hcompose' <-.
apply compose_pi2; [].
eapply IH; eauto; done.




(* sub_into_expandm *)
split.
clear IHlsub IHcutm IHcutpeq.
unfold sub_into_expandm_prop.
intros A i s m n m' s' Hsize His Hexpandm Hsubm Hsubs.
remember (use (var i) s) as r eqn:Heqr.
revert i s n m' s' Hsize Heqr His Hsubm Hsubs.
cases Hexpandm.

(* atom *)
intros c t r i s n mm s' Hsize -> His Hsubmm Hsubs.
invert Hsubmm; clear Hsubmm; [].
intros r Hsub <-.
#2 invert Hsub; clear Hsub.
  intros h' _ Hsubh _ _.
  invert Hsubh; simpl; intros; omega.

  intros s'' Hsubs' Hcut.
  so (rsub_fun _#5 Hsubs Hsubs'); subst s''; clear Hsubs'.
  apply nspine_cut; eauto using rsub_nspine; done.

(* up *)
intros A rr p v r Hexpand Hcompose i s n mm s' Hsize -> His Hsubmm Hsubs.
clear IHexpandm IHrsub.
invert Hsubmm; clear Hsubmm; [].
intros r' Hsubr <-.
invert Hcompose; clear Hcompose; [].
intros s2 Hcompose' <-.
#2 invertc Hsubr.
  intros h' _ Hsubh _ _.
  simpl in Hsubh.
  exfalso; eapply rsub_head_contra; eauto; omega.
intros s2' Hsubs2 Hcut.
so (expandp_closed_rsub n i _#3 Hexpand) as Hsubv.
so (sub_into_compose _#7 His Hcompose' Hsubs (rsub_pm (rsub_ret Hsubv))) as (s2'' & Hsubs2' & Hcompose'').
so (rsub_fun _#5 Hsubs2 Hsubs2'); subst s2''; clear Hsubs2'.
so (split_cut _#5 (rsub_nspine _#4 Hsubs His) Hcompose'' Hcut) as (n' & Hpcut & Hcut').
invertc Hcut'; [].
intros r2 Hlsub <-.
cut (r2 = r').
  intros ->.
  exact Hpcut.
clear i s n s' His Hsubs s2 Hcompose' s2' Hsubs2 Hcut Hsubv Hcompose'' Hpcut.
remember p as q eqn:Heqq in Hlsub.
remember (ret v) as rv eqn:Heqrv.
invertc Hsize; [].
intros N' HsizeA <-.
revert Heqq Heqrv.
cut (forall z (E : z = re) (r2 r' : exp z),
       lsub r2 q rv r'
       -> q = p -> rv = ret v -> r2 = r').
  intro Hind.
  so (Hind _ (eq_refl _) r2 r').
  intros; eauto; done.
clear Hlsub r2 r'.
intros z E r2 r' Hlsub.
#2 induct Hlsub; try (intros; subst; f_equal; auto; done).
  (* ret *)
  intros w p' rr rv Hcut -> ->.
  unfold sub_into_expandp_prop in IHexpandp.
  #2 exploit (cutp_invert1 ret) as Hinvert.
    intros f i d e Hsub.
    invert Hsub; [].
    intros; eauto; done.
  so (Hinvert _#5 Hcut) as (v' & -> & Hcutv).
  f_equal; [].
  symmetry.
  eapply IHexpandp; eauto; done.

  (* pm *)
  intros p1 r1 p2 r2 r _ IIH -> ->.
  f_equal; [].
  apply IIH; auto; [].
  simpl.
  f_equal; [].
  so (expandp_closed 0 (bound p1) _#3 Hexpand) as Heq.
  repl (bound p + 0) with (bound p) in Heq by omega.
  exact Heq.

(* pi *)
intros p' A B rr p v r' m HexpandA Hcompose HexpandB i s n mm s' Hsize -> His Hsubmm Hsubs.
invertc Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
invert Hsubmm; clear Hsubmm; [].
intros m' Hsubm <-.
simpl in Hcompose.
rewrite -> sh_var_at_zero in Hcompose; [].
invert Hcompose; [].
intros sv Hcomposesv <-.
so (rsub_sh_commute_right _ _ 0 (bound p) _ _ (le_0_n _) Hsubs) as Hsubs'.
repl (i + bound p) with (bound p + i) in Hsubs' by omega.
so (expandp_closed_rsub n i _#3 HexpandA) as Hsubv.
so (sub_into_compose _#7 (sh_nspine _#3 His) Hcomposesv Hsubs' (rsub_app Hsubv rsub_nl)) as (sv' & Hsubsv & Hcomposesv').
sor (refine (IHexpandm N2 _ _#7 HsizeB (compose_nspine _#3 Hcomposesv (sh_nspine _#3 His) (nspine_app nspine_nl)) HexpandB Hsubm Hsubsv)) as Hpcut; [omega |].
so (split_pcut _#5 Hcomposesv' Hpcut) as (ll & Hpcut1 & Hpcut2).
invert Hpcut2; clear Hpcut2; [].
intros p'' l l' Hcutp Hpcut2' <-.
invert Hpcut2'; clear Hpcut2'; [].
intros ->.
so (cutp_patmatch _#5 Hcutp) as Hmatch.
so (expandp_patmatch _#3 HexpandA) as Hmatch'.
so (patmatch_fun _#3 Hmatch Hmatch'); subst p''; clear Hmatch Hmatch'.
unfold cutp_expand_prop in IHcutp.
rewrite <- sh_sum in Hpcut1; [].
so (pcut_of_sh _#5 Hpcut1) as (ll & Hpcut1' & Heq).
clear Hpcut1.
revert Hcutp Hpcut1' Heq.
casesdep ll; try (intros; discriminate Heq); [].
intros p'' l' Hsubl Hpcut1' Heq.
injection Heq; clear Heq.
intros -> <-.
repl (bound p + 0) with (bound p) in Hsubl by omega.
sor (refine (IHcutp N1 _ _#7 HsizeA HexpandA Hsubl)) as ?; [omega |]; subst l'.
exact Hpcut1'.

(* wth *)
intros A B r r1 r2 m1 m2 Hcompose1 Hcompose2 HexpandA HexpandB i s n mm s' Hsize -> His Hsubmm Hsubs.
invertc Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
invertc Hsubmm; [].
intros m1' m2' Hsubm1 Hsubm2 <-.
invertc Hcompose1; [].
intros s1 Hcompose1' <-.
invertc Hcompose2; [].
intros s2 Hcompose2' <-.
so (sub_into_compose _#7 His Hcompose1' Hsubs (rsub_pi1 rsub_nl)) as (s1' & Hsubs1 & Hcompose1'').
so (sub_into_compose _#7 His Hcompose2' Hsubs (rsub_pi2 rsub_nl)) as (s2' & Hsubs2 & Hcompose2'').
sor (refine (IHexpandm N1 _ _#7 HsizeA (compose_nspine _#3 Hcompose1' His (nspine_pi1 nspine_nl)) HexpandA Hsubm1 Hsubs1)) as Hpcut1; [omega |].
sor (refine (IHexpandm N2 _ _#7 HsizeB (compose_nspine _#3 Hcompose2' His (nspine_pi2 nspine_nl)) HexpandB Hsubm2 Hsubs2)) as Hpcut2; [omega |].
so (split_pcut _#5 Hcompose1'' Hpcut1) as (l & Hpcut1a & Hpcut1b).
so (split_pcut _#5 Hcompose2'' Hpcut2) as (l' & Hpcut2a & Hpcut2b).
so (pcut_fun _#4 Hpcut1a Hpcut2a); subst l'.
invertc Hpcut1b; [].
intros l1 l2 H <-.
invertc H; [].
intros ->.
invertc Hpcut2b; [].
intros H.
invertc H; [].
intros ->.
exact Hpcut1a.

(* sigma *)
intros A B r r1 r2 m1 m2 Hcompose1 Hcompose2 HexpandA HexpandB i s n mm s' Hsize -> His Hsubmm Hsubs.
invertc Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
invertc Hsubmm; [].
intros m1' m2' Hsubm1 Hsubm2 <-.
invertc Hcompose1; [].
intros s1 Hcompose1' <-.
invertc Hcompose2; [].
intros s2 Hcompose2' <-.
so (sub_into_compose _#7 His Hcompose1' Hsubs (rsub_pi1 rsub_nl)) as (s1' & Hsubs1 & Hcompose1'').
so (sub_into_compose _#7 His Hcompose2' Hsubs (rsub_pi2 rsub_nl)) as (s2' & Hsubs2 & Hcompose2'').
sor (refine (IHexpandm N1 _ _#7 HsizeA (compose_nspine _#3 Hcompose1' His (nspine_pi1 nspine_nl)) HexpandA Hsubm1 Hsubs1)) as Hpcut1; [omega |].
sor (refine (IHexpandm N2 _ _#7 HsizeB (compose_nspine _#3 Hcompose2' His (nspine_pi2 nspine_nl)) HexpandB Hsubm2 Hsubs2)) as Hpcut2; [omega |].
so (split_pcut _#5 Hcompose1'' Hpcut1) as (l & Hpcut1a & Hpcut1b).
so (split_pcut _#5 Hcompose2'' Hpcut2) as (l' & Hpcut2a & Hpcut2b).
so (pcut_fun _#4 Hpcut1a Hpcut2a); subst l'.
invertc Hpcut1b; [].
intros l1 l2 H <-.
invertc H; [].
intros ->.
invertc Hpcut2b; [].
intros H.
invertc H; [].
intros ->.
exact Hpcut1a.




(* sub_into_expandp *)
clear IHrsub IHlsub IHcutm IHcutpeq.
unfold sub_into_expandp_prop.
intros A p v w v' Hsize Hexpandp Hcutp.
revert w v' Hsize Hcutp.
cases Hexpandp.

(* down *)
intros A m Hexpand w v' Hsize Hcut.
invertc Hsize; [].
intros N' HsizeA <-.
invert Hcut; clear Hcut; [].
intros n Hsubn <-.
invertc Hsubn; [].
intros m' Hsubm <-.
f_equal; [].
so (IHexpandm N' (lt_n_Sn _) _#7 HsizeA nspine_nl Hexpand Hsubm rsub_nl) as Hpcut.
rewrite -> sh_zero in Hpcut; [].
invert Hpcut; [].
auto; done.

(* one *)
intros w v' _ Hcut.
invert Hcut; clear Hcut; [].
intros <- <-.
reflexivity.

(* tens *)
intros A B p q v w HexpandA HexpandB ww vv2 Hsize Hcut.
invertc Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
invert Hcut; clear Hcut; [].
intros w1 w2 vv1 Hcut1 Hcut2 <-.
repl (bound q + 0) with (bound q) in Hcut1 by omega.
#2 exploit (cutp_invert2 pairp) as Hinvert.
  intros f i d1 d2 e H.
  invert H; intros; subst; eauto; done.
so (Hinvert _#6 Hcut1) as (v1 & v2 & -> & Hcutv1 & Hcutv2).
so (Hinvert _#6 Hcut2) as (v1' & v2' & Heq & Hcutv1' & Hcutv2').
subst vv2.
clear Hcut1 Hcut2 Hinvert.
repl (bound q) with (0 + 0 + bound q) in Hcutv1 at 1 by omega.
so (cutp_of_sh _#7 Hcutv1) as (v1s & Hcutv1s & ->).
simpl in Hcutv1s.
clear Hcutv1.
so (expandp_closed_cutp _ _ 0 _#3 (cutp_patmatch _#5 Hcutv2) HexpandB) as Hcutv2s.
repl (bound q + 0) with (bound q) in Hcutv2s by omega.
so (cutp_fun _#6 Hcutv2 Hcutv2s); subst v2.
clear Hcutv2 Hcutv2s.
so (cutp_sh_cancel_all _ _ 0 v1s (cutp_patmatch _#5 Hcutv2')) as Hcutv1s'.
so (cutp_fun _#6 Hcutv1' Hcutv1s'); subst v1s.
clear Hcutv1' Hcutv1s'.
unfold sub_into_expandp_prop in IHexpandp.
sor (refine (IHexpandp N1 _ _#5 HsizeA HexpandA Hcutv1s)) as ?; [omega |].
sor (refine (IHexpandp N2 _ _#5 HsizeB HexpandB Hcutv2')) as ?; [omega |].
subst; reflexivity.

(* ex *)
intros p' A B p q v w HexpandA HexpandB ww vv2 Hsize Hcut.
invertc Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
invert Hcut; clear Hcut; [].
intros w1 w2 vv1 Hcut1 Hcut2 <-.
repl (bound q + 0) with (bound q) in Hcut1 by omega.
#2 exploit (cutp_invert2 pairp) as Hinvert.
  intros f i d1 d2 e H.
  invert H; intros; subst; eauto; done.
so (Hinvert _#6 Hcut1) as (v1 & v2 & -> & Hcutv1 & Hcutv2).
so (Hinvert _#6 Hcut2) as (v1' & v2' & Heq & Hcutv1' & Hcutv2').
subst vv2.
clear Hcut1 Hcut2 Hinvert.
repl (bound q) with (0 + 0 + bound q) in Hcutv1 at 1 by omega.
so (cutp_of_sh _#7 Hcutv1) as (v1s & Hcutv1s & ->).
simpl in Hcutv1s.
clear Hcutv1.
so (expandp_closed_cutp _ _ 0 _#3 (cutp_patmatch _#5 Hcutv2) HexpandB) as Hcutv2s.
repl (bound q + 0) with (bound q) in Hcutv2s by omega.
so (cutp_fun _#6 Hcutv2 Hcutv2s); subst v2.
clear Hcutv2 Hcutv2s.
so (cutp_sh_cancel_all _ _ 0 v1s (cutp_patmatch _#5 Hcutv2')) as Hcutv1s'.
so (cutp_fun _#6 Hcutv1' Hcutv1s'); subst v1s.
clear Hcutv1' Hcutv1s'.
unfold sub_into_expandp_prop in IHexpandp.
sor (refine (IHexpandp N1 _ _#5 HsizeA HexpandA Hcutv1s)) as ?; [omega |].
sor (refine (IHexpandp N2 _ _#5 HsizeB HexpandB Hcutv2')) as ?; [omega |].
subst; reflexivity.
Qed.


Lemma rsub_expand :
  forall {z} A f i (e e' : exp z),
    expandm A (use (var 0) nl) f
    -> rsub f i (sh (i+1) 1 e) e'
    -> e = e'.
Proof.
intros z A f i e e'.
so (ntp_size A) as (N & Hsize).
revert z A f i e e' Hsize.
exact (magnificent_expansion_lemma N anderrl).
Qed.


Lemma cutp_expand :
  forall {z} A p v i (e e' : exp z),
    expandp A p v
    -> cutp v p i (sh (i + bound p) (bound p) e) e'
    -> e = e'.
Proof.
intros z A p v i e e'.
so (ptp_size A) as (N & Hsize).
revert z A p v i e e' Hsize.
exact (magnificent_expansion_lemma N anderl).
Qed.


Lemma sub_into_expandm :
  forall A i s m n m' s',
    nspine s
    -> expandm A (use (var i) s) m
    -> rsub n i m m'
    -> rsub n i s s'
    -> pcut (sh 0 i n) s' m'.
Proof.
intros A i s m n m' s'.
so (ntp_size A) as (N & Hsize).
revert A i s m n m' s' Hsize.
exact (magnificent_expansion_lemma N anderrrrl).
Qed.


Lemma sub_into_expandp :
  forall A p v w v',
    expandp A p v
    -> cutp w p 0 v v'
    -> v' = w.
Proof.
intros A p v w v'.
so (ptp_size A) as (N & Hsize).
revert A p v w v' Hsize.
exact (magnificent_expansion_lemma N anderrrrr). 
Qed.
