
Require Import Arith.
Require Import Tactics.
Require Import Sequence.
Require Import Syntax.
Require Import Cut.
Require Import Static.
Require Import Substitution.
Require Import Weakening.
Require Import Compose.
Require Import Expand.
Require Import Size.
Require Import CutEliminationWeak.
Require Import Self.
Require Import Similar.
Require Import StaticWeak.
Require Import IdentityWeak.
Require Import Selfish.
Require Import Regularity.


(* Could be replaced by call to wpof_fun. *)
Lemma pof_fun :
  forall A p p' G G',
    pof p A G
    -> pof p' A G'
    -> p = p' /\ G = G'.
Proof.
intros A p p' G G' H Hofp'.
revert p' G' Hofp'.
induct H.

(* down *)
intros A p' G' Hofp'.
invert Hofp'; [].
intros; subst; auto; done.

(* one *)
intros p' G' Hofp'.
invert Hofp'; [].
intros; subst; auto; done.

(* tens *)
intros p1 p2 A B G1 G2 _ IH1 _ IH2 pp GG Hofpp.
invert Hofpp; clear Hofpp; [].
intros p1' p2' G1' G2' Hofp1' Hofp2' <- <-.
so (IH1 _ _ Hofp1') as (<- & <-).
so (IH2 _ _ Hofp2') as (<- & <-).
auto; done.

(* ex *)
intros p1 p2 A B G1 G2 Hofp1 IH1 Hofp2 IH2 pp GG Hofpp.
invertc Hofpp; [].
intros p2' G1' G2' Hofp1' Hofp2' <- <-.
so (IH1 _ _ Hofp1') as (_ & <-).
so (IH2 _ _ Hofp2') as (<- & <-).
auto.
Qed.


Lemma can_pof_expandp :
  forall G A p v,
    wptof G A
    -> expandp A p v
    -> exists G', pof p A G'.
Proof.
intros G A p v Hofty Hexpand.
revert G Hofty.
induct Hexpand;
try (intros; invertc Hofty; intros; subst;
     repeat
       (match goal with
        | H : forall _, _ -> _ |- _ => eexploit H; clear H; [eassumption |]
        end);
     destruct_everywhere;
     eexists; eauto using pof;
     done); [].
(* ex *)
intros p' A B p q v w _ IH1 _ IH2 G Hofty.
invertc Hofty; [].
intros G' HofA Hofp' HofB.
so (IH1 _ HofA) as (G1 & Hofp).
so (IH2 _ HofB) as (G2 & Hofq).
so (wpof_fun _#5 (pof_weak _#3 Hofp) Hofp') as (<- & _).
eauto using pof; done.
Qed.


Theorem identity_expansion :
  (forall G r A m,
     wwf nil G
     -> rof G r (susp A)
     -> expandm A r m
     -> tof G m A)
  /\
  (forall G A G' p v,
     wwf nil G
     -> wptof G A
     -> expandp A p v
     -> pof p A G'
     -> vof (G;; G') v (sh 0 (length G') A)).
Proof.
cut (forall N,
       (forall G r A m,
          size A N
          -> wwf nil G
          -> rof G r (susp A)
          -> expandm A r m
          -> tof G m A)
       /\
       (forall G A G' p v,
          size A N
          -> wwf nil G
          -> wptof G A
          -> expandp A p v
          -> pof p A G'
          -> vof (G;; G') v (sh 0 (length G') A))).
  intro Hind.
  split.
    intros G r A m.
    so (ntp_size A) as (N & Hsize).
    intros; eapply (Hind _ andel); eauto; done.

    intros G A.
    so (ptp_size A) as (N & Hsize).
    intros; eapply (Hind _ ander); eauto; done.
intro N.
wfinduct N using lt_wf.
intros N IH.
so (fun N Hlt => IH N Hlt andel) as IHn.
so (fun N Hlt => IH N Hlt ander) as IHp.
clear IH.
split.



(* negative types *)
intros G r A m Hsize Hwf Hofr Hexpand.
revert G Hsize Hwf Hofr.
cases Hexpand.

(* atom *)
intros c s r G Hsize Hofr.
apply tof_etam; auto; done.

(* up *)
intros A r p v r' HexpandA Hcompose G Hsize Hwf Hofr.
invertc Hsize; [].
intros N' HsizeA <-.
invert (rof_reg_weak _#3 Hwf Hofr); [].
intros HofA.
so (can_pof_expandp _#4 HofA HexpandA) as (G' & Hofp).
so (IHp _ (lt_n_Sn _) _#5 HsizeA Hwf HofA HexpandA Hofp) as Hofv.
#2 sor (refine (rcompose_formation _ _ (trp A) _#3 Hcompose Hofr _)) as Hofr'.
  eapply sof_pm; eauto; [].
  simpl.
  apply rof_ret; auto; done.
apply tof_shiftp; auto; done.

(* pi *)
intros p' A B r p v r' m HexpandA Hcompose HexpandB G Hsize Hwf Hofr.
invertc Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
invert (rof_reg_weak _#3 Hwf Hofr); [].
intros G' HofA Hofp' HofB.
so (can_pof_expandp _#4 HofA HexpandA) as (G'' & Hofp).
so (wpof_fun _#5 (pof_weak _#3 Hofp) Hofp') as (<- & ->).
so (IHp N1 (le_lt_n_Sm _ _ (le_plus_l _ _)) _#5 HsizeA Hwf HofA HexpandA Hofp) as Hofv.
so (wpof_similar _ _ (sh 0 (length G') A) _ Hofp' (sh_similarp _#3)) as (G'' & HsimG' & Hofp'').
#3 exploit (can_cutp_nt_weak (G;; G') (sh 0 (length G') A) G'' nil 0 v p (sh (length G') (length G') B)) as (B' & HsubB & _); auto.
  apply vof_weak; auto; done.

  simpl.
  #2 apply (wntof_similarctx (G;; G';; shctx 0 (length G') G')).
    apply weaken_wntof; auto using shctx_length.
    eapply wntof_similarctx; eauto using shctx_similarctx with similar; done.

    apply similarctx_append; eauto with similar.
    apply (similarctx_trans _ G'); eauto using shctx_similarctx with similar; done.
rewrite -> (pof_bound_length _#3 Hofp) in HsubB; [].
so (cutp_expand _#6 HexpandA HsubB); subst B'.
so (pof_reg_weak _#4 HofA Hofp) as HwfG'.
so (wwf_append_closed _ _ Hwf HwfG') as Hwf'.
so (weakenn_rof _ G' _ _ _ Hofr (eq_refl _)) as Hofrw.
simpl in Hofrw.
repl (bound p + 0) with (bound p) in Hofrw by omega.
rewrite -> (pof_bound_length _#3 Hofp) in Hofrw, Hofv; [].
so (rcompose_formation _#6 Hcompose Hofrw (sof_app Hofv HsubB sof_nl)) as Hofr'.
so (IHn N2 (le_lt_n_Sm _ _ (le_plus_r _ _)) _#4 HsizeB Hwf' Hofr' HexpandB) as Hofm.
eapply tof_lam; eauto; done.

(* wth *)
intros A B r r1 r2 m1 m2 Hcompose1 Hcompose2 HexpandA HexpandB G Hsize Hwf Hofr.
clear IHp.
invertc Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
so (rcompose_formation _#6 Hcompose1 Hofr (sof_pi1 sof_nl)) as Hofr1.
so (rcompose_formation _#6 Hcompose2 Hofr (sof_pi2 sof_nl)) as Hofr2.
so (IHn N1 (le_lt_n_Sm _ _ (le_plus_l _ _)) _#4 HsizeA Hwf Hofr1 HexpandA) as Hofm1.
so (IHn N2 (le_lt_n_Sm _ _ (le_plus_r _ _)) _#4 HsizeB Hwf Hofr2 HexpandB) as Hofm2.
eapply tof_pairm; eauto; done.

(* sigma *)
intros A B r r1 r2 m1 m2 Hcompose1 Hcompose2 HexpandA HexpandB G Hsize Hwf Hofr.
clear IHp.
invertc Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
so (selfify _#4 Hwf Hofr (expandm_sigma Hcompose1 Hcompose2 HexpandA HexpandB)) as (CD & Hself & HofC & Hofr').
invertc Hself; [].
intros C B' D HselfA HsubB HselfB <-.
so (rcompose_formation _#6 Hcompose1 Hofr' (sof_pi1 sof_nl)) as Hofr1.
so (rcompose_formation _#6 Hcompose2 Hofr' (sof_pi2 sof_nl)) as Hofr2.
so (IHn N1 (le_lt_n_Sm _ _ (le_plus_l _ _)) _#4 (self_size _#4 HselfA HsizeA) Hwf Hofr1 (expandm_similar _#4 (self_similar _#3 HselfA) HexpandA)) as Hofm1.
so (IHn N2 (le_lt_n_Sm _ _ (le_plus_r _ _)) _#4 (self_size _#4 HselfB (rsub_size _#5 HsubB HsizeB)) Hwf Hofr2 (expandm_similar _#4 (similar_trans _#3 (rsub_similarm _#4 HsubB) (self_similar _#3 HselfB)) HexpandB)) as Hofm2.
eapply self_member_rev; eauto using self; [].
apply tof_pairm; auto; done.



intros G A G' p v Hsize Hwf HofA Hexpand Hofp.
revert G G' Hsize Hwf HofA Hofp.
cases Hexpand.

(* down *)
intros A m HexpandA G G' Hsize Hwf Hofty Hofpat.
invertc Hofpat; [].
intros <-.
invertc Hofty; [].
intros HofA.
invert Hsize; clear Hsize; [].
intros N' HsizeA <-.
#4 exploit (IHn _ (lt_n_Sn _) (G; A) (use (var 0) nl) (sh 0 1 A) m) as Hofm; eauto using shift_size.
  apply wwf_cons; auto; [].
  autorewrite with canonlist; assumption.

  eapply rof_use; eauto using sof; [].
  apply hof_id; eauto using lookup_var, index_0; done.

  eapply expandm_similar; eauto using sh_similarm; done.
simpl.
apply vof_thunk; auto; done.

(* one *)
intros G G' Hsize Hwf Hofty Hofpat.
invertc Hofpat; [].
intros <-.
simpl.
apply vof_unitp; done.

(* tens *)
intros A B p q v w HexpandA HexpandB G1 GG Hsize Hwf Hofty Hofpat.
invertc Hofpat; [].
intros G2 G3 Hofp Hofq <-.
invertc Hofty; [].
intros HofA HofB.
invert Hsize; clear Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
so (IHp N1 (le_lt_n_Sm _ _ (le_plus_l _ _)) _#5 HsizeA Hwf HofA HexpandA Hofp) as Hofv.
so (IHp N2 (le_lt_n_Sm _ _ (le_plus_r _ _)) _#5 HsizeB Hwf HofB HexpandB Hofq) as Hofw.
autorewrite with canonlist.
simpl.
#2 apply vof_pairp.
  #2 sor (refine (weakenn_vof _ (shctx 0 (length G2) G3) (bound q) _ _ Hofv _)) as Hofv'.
    rewrite -> shctx_length; eapply pof_bound_length; eauto; done.
  rewrite -> sh_sum in Hofv'; [].    
  rewrite -> app_length; [].
  rewrite -> shctx_length; [].
  rewrite -> (pof_bound_length _#3 Hofq); [].
  exact Hofv'.

  so (weaken_vof _ G2 _#5 (eq_refl _) (eq_refl _) Hofw) as Hofw'.
  rewrite -> (pof_bound_length _#3 Hofq) in Hofw' at 1; [].
  repl (bound q) with (bound q + 0) in Hofw' by omega.
  rewrite -> (expandp_closed _#5 HexpandB) in Hofw'; [].
  rewrite -> app_length; [].
  rewrite -> shctx_length; [].
  repl (length G3) with (0 + length G3) in Hofw' at 1 by omega.
  rewrite -> sh_sum_left in Hofw'; [].
  repl (length G3 + length G2) with (length G2 + length G3) by omega.
  exact Hofw'.

(* ex *)
intros p' A B p q v w HexpandA HexpandB G1 GG Hsize Hwf Hofty Hofpat.
clear IHn.
invertc Hofpat; [].
intros G2 G3 Hofp Hofq <- <-.
invertc Hofty; [].
intros G2' HofA Hofp' HofB.
so (wpof_fun _#5 (pof_weak _#3 Hofp) Hofp') as (_ & <-).
invert Hsize; clear Hsize; [].
intros N1 N2 HsizeA HsizeB <-.
so (IHp N1 (le_lt_n_Sm _ _ (le_plus_l _ _)) _#5 HsizeA Hwf HofA HexpandA Hofp) as Hofv.
so (pof_reg_weak _#4 HofA Hofp) as HwfG2.
so (wwf_append_closed _ _ Hwf HwfG2) as Hwf'.
so (IHp N2 (le_lt_n_Sm _ _ (le_plus_r _ _)) _#5 HsizeB Hwf' HofB HexpandB Hofq) as Hofw.
so (weaken_pof 0 (length G2) _#3 Hofp) as Hofpw.
#2 exploit (can_cutp_pt_weak (G1;; G2) (sh 0 (length G2) A) (shctx 0 (length G2) G2) nil 0 v p (sh (length G2) (length G2) B)) as H; eauto using vof_weak, pof_weak.
  autorewrite with canonlist.
  so (weaken_wptof G1 G2 G2 _ _ B (eq_refl _) (eq_refl _) HofB) as HofBw.
  eapply wptof_similarctx; eauto using shctx_similarctx with similar; done.
destruct H as (B' & HsubB & _).
so HsubB as HsubB'.
rewrite -> (pof_bound_length _#3 Hofp) in HsubB'; [].
repl (bound p) with (0 + bound p) in HsubB' at 1 by omega.
so (cutp_expand _#6 HexpandA HsubB'); subst B'.
clear HsubB'.
so (cutp_sh_commute_left _#4 0 (bound q) _ _ HsubB) as HsubBw.
simpl in HsubBw.
rewrite <- (pof_bound_length _#3 Hofp) in HsubBw; [].
repl (length G2 + 0) with (length G2) in HsubBw by omega.
rewrite -> sh_sum in HsubBw; [].
autorewrite with canonlist.
simpl.
#3 eapply vof_pairpx.
  so (weakenn_vof _ G3 (bound q) _ _ Hofv (pof_bound_length _#3 Hofq)) as Hofv'.
  rewrite -> sh_sum in Hofv'; [].    
  rewrite -> app_length; [].
  rewrite -> (pof_bound_length _#3 Hofq); [].
  exact Hofv'.

  force_exact HsubBw.
  f_equal; [].
  #2 f_equal.
    so (pof_bound_length _#3 Hofp); omega.

    rewrite -> app_length; [].
    so (pof_bound_length _#3 Hofq); omega.

  rewrite <- (pof_bound_length _#3 Hofq); [].
  exact Hofw.
Qed.
